<?php
namespace Service;

use Doctrine\ORM\EntityManager;
use Service\Calendar\Precursor\Strategy\PrecursorFactory;
use Service\Calendar\Restriction\Strategy\RestrictionFactory;
use Util\InstanceProvider;
use Util\Messages;
use Util\Utilities;
use Zeedhi\DevOps\Util\Utilities as DevOpsUtilities;
use Util\Mail;
use Zeedhi\Framework\ORM\DateTime;
use Zeedhi\DevOps\Model\Projectmanagement\Entities;
use Zeedhi\DevOps\Model\Projectmanagement\Repositories;
use Zeedhi\DevOps\Model\Devops\Entities as EntitiesDevOps;

/**
 * Class Task
 *
 * @package Service
 */
class Task {

	/** @var EntityManager $entityManager */
	protected $entityManager;
	/** @var \Util\Utilities $environment */
	protected $environment;

    const ACTIVE = 'A';
    const IS_ISSUE = 'Y';
    const IS_TASK = 'Y';
    const IS_PROJECT = 'Y';

    const RED = '#D85252';
    const GREEN = '#85AB2C';
    const DARK_GREEN = '#547704';
    const GREY = '#d3d3d3';

    const ALLOCATION_ASSIGNER = 'A';
    const ALLOCATION_REPORTER = 'R';
    const ALLOCATION_SPONSOR  = 'S';

    const TASK_PREFIX = 'T';
    const PROJECT_PREFIX = 'P';

    const RELEVANT = Entities\Allocation::RELEVANT;
    const NOT_RELEVANT = Entities\Allocation::NOT_RELEVANT;

    const TASK_STATUS_CONTROL_NOT_INITIALIZED = Entities\TaskStatus::TASK_STATUS_CONTROL_NOT_INITIALIZED;
    const TASK_STATUS_CONTROL_IN_DEVELOPING = Entities\TaskStatus::TASK_STATUS_CONTROL_IN_DEVELOPING;
    const TASK_STATUS_CONTROL_PAUSED  = Entities\TaskStatus::TASK_STATUS_CONTROL_PAUSED;
    const TASK_STATUS_CONTROL_CONCLUDED  = Entities\TaskStatus::TASK_STATUS_CONTROL_CONCLUDED;
    const TASK_STATUS_CONTROL_BACKLOG  = Entities\TaskStatus::TASK_STATUS_CONTROL_BACKLOG;

    const ADD_ALLOCATION = 'ADD';
    const REPLACE_ALLOCATION = 'REPLACE';

    const FIRST_COMMENTS = 'Abertura de issue.';
    const CLOSE_COMMENTS = 'Demanda encerrada pelo solicitante.';
    const MAXIMUM_CHARACTER_LIMIT_MESSAGE = ' ... limite máximo de caracteres excedidos. Vá ate a tela de Acompanhamento de Tarefas para visualizar as respostas das terafas geradoras.';

    public function __construct(){

        $this->environment = InstanceProvider::getEnvironment();
        $this->entityManager = InstanceProvider::getEntityManager();
        //InstanceProvider::getService()->getCalendarTakCalendar() = $taskCalendar;
        //InstanceProvider::getMailIssueMail() = $issueCommentMail;

        //InstanceProvider::getService()->getIssue() = $issueService;
        //InstanceProvider::getService()->getResource() = $resourceService;
        //InstanceProvider::getService()->getStructure() = $structureService;
        //InstanceProvider::getService()->getTaskPrecursor()ce = $taskPrecursor;
        //InstanceProvider::getService()->getAllocation() = $allocationService;
        //InstanceProvider::getService()->getAppropriation() = $appropriationService;
        //InstanceProvider::getService()->getAttachment() = $attachmentService;
        //InstanceProvider::getService()->getTaskTypeClassification() = $taskTypeClassificationService;

        //InstanceProvider::getFactory()->getTask() = $taskFactory;
        //InstanceProvider::getFactory()->getLabel() = $labelFactory;
        //InstanceProvider::getFactory()->getComments() = $commentsFactory;
        //InstanceProvider::getFactory()->getTaskType() = $taskTypeFactory;
        //InstanceProvider::getFactory()->getTaskLabel() = $taskLabelFactory;
        //InstanceProvider::getFactory()->getTaskStatus() = $taskStatusFactory;
        //InstanceProvider::getFactory()->getChangedFiles() = $changedFilesFactory;
        //InstanceProvider::getFactory()->getClassification() = $classificationFactory;
        //InstanceProvider::getFactory()->getResourceProject() = $resourceProjectFactory;
        //InstanceProvider::getFactory()->getTaskRelationship() = $taskRelationshipFactory;

        //InstanceProvider::getRepository()->getTask() = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Task');
        //InstanceProvider::getRepository()->getLabel() = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Label');
        //InstanceProvider::getRepository()->getProject() = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Project');
        //InstanceProvider::getRepository()->getResource() = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Resource');
        //$this->taskTypeRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'TaskType');
        //$this->commentsRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Comments');
        //$this->taskLabelRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'TaskLabel');
        //$this->organizationRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Organization');
        //$this->allocationRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Allocation');
        //$this->taskStatusRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'TaskStatus');
        //$this->workedTimeRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'WorkedTime');
        //$this->repositoryRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Repository');
        //$this->taskFeaturesRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'TaskFeatures');
        //$this->changedFilesRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'ChangedFiles');
        //$this->classificationRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Classification');
        //$this->taskPropertiesRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'TaskProperties');
        //$this->resourceProjectRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'ResourceProject');
        //$this->taskRelationshipRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'TaskRelationship');
        //$this->commentTypeRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'CommentType');
        //$this->taskTypeClassificationRepository = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'TaskTypeClassification');


        \Util\InstanceProvider::getLog()->setLog('instanciando task service',false,2);
    }

    /**
     * Creates a new task.
     *
     * @param Entities\Project $project
     * @param $latency
     * @param $name
     * @param \DateTime $estimatedBeginDate
     * @param \DateTime $estimatedEndDate
     * @param $status
     * @param $description
     * @param $percentCompleted
     * @param $duration
     * @param Entities\Task $parent
     * @param $taskStatus
     * @param $dateExceptionType
     * @param $dateException
     * @param string $isParent
     * @param null $urlToTest
     * @param null $classification
     * @return Entities\Task
     */
	public function createTask(Entities\Project $project, $latency, $name, \DateTime $estimatedBeginDate,
							   \DateTime $estimatedEndDate, $status, $description, $percentCompleted, $duration,
							   Entities\Task $parent = null ,$taskStatus, $dateExceptionType,
                               $dateException, $isParent = 'N', $urlToTest = null, $classification = null,
                               Entities\TaskFeatures $taskFeatures = null, $errorClassification = null, $timePlanned = null, $versionDevId = null) {
        $isIssue        = 'N';
        $errorMessage   = '';
        $priority       = null;
        $endDate        = null;
        $productVersion = null;

        if($versionDevId){
            $productVersion =  InstanceProvider::getRepository()->getProductVersion()->find($versionDevId);
        }
        $entity = InstanceProvider::getFactory()->getTask()->createTask($project, $latency, $name, $estimatedBeginDate, $estimatedEndDate, $status,
            $description, $percentCompleted, $duration, $parent, $taskStatus, $dateExceptionType, $dateException, $isIssue,
            $isParent, $classification, $errorMessage, $priority, $urlToTest, $taskFeatures, $endDate, $errorClassification, $timePlanned, null, $productVersion);
		$this->entityManager->persist($entity);
		return $entity;
	}

	public function editTask(Entities\Task $task, Entities\Project $project, $latency, $name, \DateTime $estimatedBeginDate,
							   \DateTime $estimatedEndDate, $status, $description, $percentCompleted, $duration,
							   Entities\Task $parent = null ,$taskStatus, $dateExceptionType,
                               $dateException, $isParent = 'N', $urlToTest = null, $classification = null, $priority = null) {
        $isIssue        = 'N';
        $errorMessage   = '';
        $entity = InstanceProvider::getFactory()->getTask()->updateTask($task,$project, $latency, $name, $estimatedBeginDate, $estimatedEndDate, $status,
            $description, $percentCompleted, $duration, $parent, $taskStatus, $dateExceptionType, $dateException, $isIssue,
            $isParent, $classification, $errorMessage, $priority, $urlToTest);
		$this->entityManager->persist($entity);
		return $entity;
	}

    public function createIssue(\Zeedhi\DevOps\Model\Projectmanagement\Entities\Project $project, $latency, $name, \DateTime $estimatedBeginDate,
							   \DateTime $estimatedEndDate, $status, $description, $percentCompleted, $duration,
							   \Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $parent = null ,$taskStatus, $dateExceptionType,
                               $dateException, $isParent = 'N', $urlToTest = null, $classification = null, $priority = null) {
        $isIssue        = 'Y';
        $errorMessage   = '';
        $entity = InstanceProvider::getFactory()->getTask()->createTask($project, $latency, $name, $estimatedBeginDate, $estimatedEndDate, $status,
            $description, $percentCompleted, $duration, $parent, $taskStatus, $dateExceptionType, $dateException, $isIssue,
            $isParent, $classification, $errorMessage, $priority, $urlToTest);
		$this->entityManager->persist($entity);
		return $entity;
	}

	protected function getTaskTopMostParent($parent){
		$pParent = $parent->getParent();
		return empty($pParent) ? $parent : $this->getTaskTopMostParent($pParent);
	}

	protected function calculateTaskProgressRecursively($parent){
		$childrenTasks = InstanceProvider::getRepository()->getTask()->findBy(array('parent' => $parent->getId()));
		if(empty($childrenTasks)){
			$duration = $parent->getDuration();
			$progress = $parent->getPercentCompleted() * $duration / 100;
			return array(
				'progress'=> $progress,
				'duration' => $duration
			);
		}
		else{
			$maxProgress = 0;
			$progress = 0;
			foreach ($childrenTasks as $task) {
				$taskInfo = $this->calculateTaskProgressRecursively($task);
				$maxProgress += $taskInfo['duration'];
				$progress += $taskInfo['progress'];
			}
			InstanceProvider::getFactory()->getTask()->updatePercentCompleted($parent, ($progress / $maxProgress * 100));
			$this->entityManager->persist($parent);
			return array(
				'progress'=> $progress,
				'duration' => $maxProgress
			);
		}
	}

    public function createdTaskClientAccept(Entities\Task $issue,Entities\Project $project){
        $name = $issue->getName();
        $projectId = $project->getId();
        /** @var Entities\Classification $classification */
        $classification = InstanceProvider::getRepository()->getClassification()->findOneBy(array('control'=>Entities\Classification::CONTROL_CLIENT_ACCEPT));
        $taskClassificationId = $classification->getId();
        $parentId = null;
        $description = null;
        /** @var Entities\Allocation $allocationReporter */
        $allocationReporter = InstanceProvider::getRepository()->getAllocation()->findOneBy(array('task' => $issue,'type' =>Entities\Allocation::TYPE_REPORTER ));
        $resource = array($allocationReporter->getResource()->getId());
        $task = $this->createFastTask($name, $projectId, $taskClassificationId, $parentId, $description,$resource,array(),
            null,null,true);
        $task->setTimePlanned(0);
        $task->setDuration(0);
        $type =  Entities\TaskRelationship::RELATIONSHIP_ISSUE_TYPE;
        /** @var Entities\TaskRelationship $taskRelationshipIssue */
        $taskRelationshipIssue = InstanceProvider::getRepository()->getTaskRelationship()
            ->findOneBy(array('task' => $issue,'type' => $type,'taskRelationship'=>$task));
        if(empty($taskRelationshipIssue)){
            $taskRelationship = InstanceProvider::getFactory()->getTaskRelationship()->createTaskRelationship($issue, $task,$type);
            $this->entityManager->persist($taskRelationship);
        }
        /** @var  Entities\TaskStatus $taskStatusClientAccept */
        $taskStatusClientAccept = InstanceProvider::getRepository()->getTaskStatus()->findOneBy(array('control'=>Entities\TaskStatus::TASK_STATUS_CONTROL_CLIENT_ACCEPT));
        $issue->setTaskStatus($taskStatusClientAccept);
        $this->entityManager->persist($issue);
        $this->entityManager->persist($task);
        $this->entityManager->flush();
    }

	/**
     * Recalculates given task progress.
     *
	 * @param $taskId
	 * @param $progress
	 * @return null|object|\Zeedhi\DevOps\Model\Projectmanagement\Entities\Task
	 */
	public function recalculateTaskProgress($taskId, $progress){

		$taskEntity = $this->entityManager->find('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Task', $taskId);
		if($taskEntity){
		    	$taskEntity = InstanceProvider::getFactory()->getTask()->updatePercentCompleted($taskEntity, $progress);
		$this->entityManager->persist($taskEntity);
		$taskParent = $taskEntity->getParent();
		if(!empty($taskParent)){
			$taskParent = $this->getTaskTopMostParent($taskParent);
			$this->calculateTaskProgressRecursively($taskParent);
		}
		$this->entityManager->flush();
		return $taskEntity;
		}

	}

	public function allocateCurrentUserInTasks($rows){
	    $connection = $this->entityManager->getConnection();
	    $allocatedTasks = [];
        try {
            $connection->beginTransaction();
            $taskRepository = InstanceProvider::getRepository()->getTask();
    		foreach($rows as $row){
    		    if($row['TASK_ID']){
    		        $task = $taskRepository->find($row['TASK_ID']);
    		        $allocation = $this->setAllocationsInTask($task, [$this->environment->getCurrentUserId()], null );

    		        $row['ALLOCATION'] = $allocation->getId();
    		        $row['RESOURCE'] = $this->environment->getCurrentUserId();
    		        array_push($allocatedTasks, $row );
    		    }
    		}
            $connection->commit();
            return $allocatedTasks;

        }  catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }


	}

    /**
     * @param \Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $task
     * @param \Zeedhi\DevOps\Model\Projectmanagement\Entities\Project $project
     * @param $latency
     * @param $name
     * @param \DateTime $estimatedBeginDate
     * @param \DateTime $estimatedEndDate
     * @param $status
     * @param $description
     * @param $percentCompleted
     * @param $duration
     * @param \Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $parent
     * @param $taskStatus
     * @param $dateExceptionType
     * @param $dateException
     * @param $isParent
     * @param null $urlToTest
     * @param null $classification
     * @throws \Doctrine\ORM\ORMException
     * @throws \Doctrine\ORM\OptimisticLockException
     * @throws \Doctrine\ORM\TransactionRequiredException
     * @return \Zeedhi\DevOps\Model\Projectmanagement\Entities\Task
     */
	public function updateTask(\Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $task, \Zeedhi\DevOps\Model\Projectmanagement\Entities\Project $project, $latency, $name, \DateTime $estimatedBeginDate,
							   \DateTime $estimatedEndDate, $status, $description, $percentCompleted, $duration,
							   \Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $parent = null, $taskStatus,$dateExceptionType, $dateException, $isParent,
                               $urlToTest = null, $classification = null, $response = null, $taskFeatures = null, $versionDevId = null) {

        $currentStatus = $task->getStatus();
		$duration = InstanceProvider::getService()->getCalendarTakCalendar()->getDurationInMilliseconds($duration,$task);
        $productVersion = null;
        if($versionDevId){
            $productVersion = InstanceProvider::getRepository()->getProductVersion()->find($versionDevId);
        }
		$entity = InstanceProvider::getFactory()->getTask()->updateTask($task, $project, $latency, $name, $estimatedBeginDate, $estimatedEndDate,
			$status, $description, $percentCompleted, $duration, $parent, $taskStatus,
			$dateExceptionType, $dateException, $isParent, 'N', $classification, '', null, $urlToTest, $response, $taskFeatures, null, $productVersion);
		if(!empty($parent)){
			$parentEntity = $this->entityManager->find('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Task', $parent);
			InstanceProvider::getFactory()->getTask()->updateTaskParent($parentEntity);
		}
		$this->updateStatusRelatedTasks($task->getId(), $entity->getStatus(), $currentStatus);
		$this->entityManager->persist($entity);
		$this->entityManager->flush();
		return $entity;
	}

	public function getTasksByCurrentUser() {
		return $this->getTasksByUser($this->environment->getCurrentUserId());
	}

	public function findTaskBranch($taskId) {
		return InstanceProvider::getRepository()->getTask()->findTaskBranch($taskId);
	}

	public function validateAddTaskClassification($taskTypeId) {
		return InstanceProvider::getRepository()->getTask()->validateAddTaskClassification($taskTypeId);
	}

	public function validateAddProjectType($taskTypeId) {
		return InstanceProvider::getRepository()->getTask()->validateAddTaskClassification($taskTypeId);
    }

    public function findAllTaskByProject($projectId) {
		return InstanceProvider::getRepository()->getTask()->findAllTaskByProject($projectId);
	}

	public function getTasksByUser($userId) {
		return InstanceProvider::getRepository()->getTask()->getStartedTasksByUser($userId);
	}

    public function getTasksByPriority ($priorityId) {
		return InstanceProvider::getRepository()->getTask()->getStartedTaskPriority($priorityId);
	}

    public function findErrorTaskFromErrorMerge($task_id) {
        return InstanceProvider::getRepository()->getTask()->findErrorTaskFromErrorMerge($task_id);
    }

	public function getAllStatusTask($conditions = null) {
		# Checks if the project allows or does not allow his tasks to be ended
        $allowEndTasks = null;
        if(!empty($conditions)) {
            $allowEndTasks = $conditions[0]['value'];
        }
        $filterString = Utilities::getFilterTableColumn($conditions, 'LABEL');
        $filterString = DevOpsUtilities::createQueryFiltered($filterString);
		return InstanceProvider::getRepository()->getTask()->getAllStatusTask($this->environment->getCurrentOrganizationId(), $allowEndTasks, $filterString);
	}

	public function getAllTypeTask() {
		return InstanceProvider::getRepository()->getTask()->getAllTypeTask($this->environment->getCurrentOrganizationId());
	}

    /****************************************************************************
     * Gets an array and transforms it in a string.
     ****************************************************************************/
    private function transformArrayInQuery($array, $initialText = "", $finalText = ""){
        if(empty($array)){
            $array = array('0');
        }
        return $initialText . implode(", ", $array) . $finalText;
    }

	public function getParentTaskByProjectAndTask($projectId, $taskId, $selectedTasks, $isParent, $param = array()) {
        $filter = Utilities::createQueryFiltered($param);
        $selectedTasksNotInQuery = $this->transformArrayInQuery($selectedTasks, "NOT IN (", ")");
		return InstanceProvider::getRepository()->getTask()->getParentTaskByProjectAndTask($this->environment->getCurrentOrganizationId(),
            $projectId, $taskId, $selectedTasksNotInQuery,$isParent, $filter);
	}

	public function getPrecursorTaskByProjectAndTask($projectId, $taskId) {
		return InstanceProvider::getRepository()->getTask()->getPrecursorTaskByProjectAndTask($this->environment->getCurrentOrganizationId(), $projectId, $taskId);
	}

	public function getDateOfUserTasksFromToday($userId) {
		return InstanceProvider::getRepository()->getTask()->getDateOfUserTasksFromToday($userId);
	}

	public function getDateTasksByCurrentUser() {
		return $this->getDateOfUserTasksFromToday($this->environment->getCurrentUserId());
	}

	public function getTasksByUserFromDate($userId, $dateTask) {
		/** @var Entities\TaskProperties $taskPropertiesEntity */
		$taskPropertiesEntity = InstanceProvider::getRepository()->getTaskProperties()->findOneBy(array('organization'=>$this->environment->getCurrentOrganizationId()));
		$finalTaskStatusId = $taskPropertiesEntity->getFinalTaskStatus()->getId();
		return InstanceProvider::getRepository()->getTask()->getTasksByUserFromDate($userId, $dateTask, $finalTaskStatusId);
	}

	public function getUnatendedTasksByUserFromDate($dateTask) {
		/** @var Entities\TaskProperties $taskPropertiesEntity */
		$taskPropertiesEntity = InstanceProvider::getRepository()->getTaskProperties()->findOneBy(array('organization'=>$this->environment->getCurrentOrganizationId()));
		$finalTaskStatusId = $taskPropertiesEntity->getFinalTaskStatus()->getId();
		return InstanceProvider::getRepository()->getTask()->getUnatendedTasksByUserFromDate($dateTask, $finalTaskStatusId, $this->environment->getCurrentUserId());
	}


    public function setTaskTimePlannedIfNull(Entities\Task $task, $resource,$timePlanned)
    {
        if (!$timePlanned) {
//            /** @var Entities\Resource[] $resourcesObjects */
//            $resourcesObjects = InstanceProvider::getRepository()->getResource()->findBy(array('id' => $resource));
//            $totalRating = 0;
//            if (!$resourcesObjects) {
//                $totalRating = 1;
//            }
//            foreach ($resourcesObjects as $resourceObjects) {
//                if ($resourceObjects->getRating()) {
//                    $totalRating += $resourceObjects->getRating();
//                } else {
//                    $totalRating += 1;
//                }
//            }
//            $task->setTimePlanned($task->getDuration() * $totalRating);

            $task->setTimePlanned(0);
        }
    }

    /**
     * @param $taskCreated
     * @param $typeRelationship
     * @return mixed
     */
    public function setNotInFraming($taskCreated, $typeRelationship,$relatedTaskIds)
    {
        $task = $this->getGeneratorTask($taskCreated, $typeRelationship, $relatedTaskIds);

        if ($typeRelationship == Entities\TaskRelationship::ERROR_TYPE
            && $task->getClassification()->getControl()
        ) {
            $task->setUsedInFraming(Entities\Task::NOT_USED_IN_FRAMING);
            $this->entityManager->persist($task);
            $this->entityManager->flush();

        }
    }

    public function setColorGantt($task)
    {
        $dataRealEndCalculated = '';
        $dateEstimatedEndCalculated = '';
        if ($task['REAL_END']) {
            $dataRealEndCalculated = implode(array_reverse(explode('/', $task['REAL_END'])));
        }
        if ($task['ESTIMATED_END']) {
            $dateEstimatedEndCalculated = implode(array_reverse(explode('/', $task['ESTIMATED_END'])));
        }

        if ($task['CONTROL'] == Entities\TaskStatus::TASK_STATUS_CONTROL_BACKLOG) {
            $task['COLOR'] = self::GREY;
            $task['COLOR_REAL'] = self::GREY;
            $task['COLOR_ESTIMATED'] = self::GREY;
        } else if ($task['COMPLETED'] < 100 && $dataRealEndCalculated > $dateEstimatedEndCalculated) {
            $task['COLOR'] = self::RED;
            $task['COLOR_REAL'] = self::RED;
        } else if ($task['COMPLETED'] == 100) {
            $task['COLOR'] = self::DARK_GREEN;
            $task['COLOR_REAL'] = self::GREEN;
        }else if (
                    $task['COMPLETED'] < 100 &&
                    $dataRealEndCalculated &&
                    $dataRealEndCalculated <= $dateEstimatedEndCalculated
                 ) {
            $task['COLOR'] = self::GREEN;
            $task['COLOR_REAL'] = self::GREEN;
        }
        return $task;
    }

    private function joinTasks($tasks, $delayedTasks) {
        $tasks = array_merge($tasks, $delayedTasks);
        return $tasks;

    }

    public function getTasksByUserFromDateOnDemand($fromDate, $projectFilter){
        $today = date("d/m/Y");
        if(!empty($projectFilter)) {
            $project = str_replace("P", "", $projectFilter);
            if($fromDate == $today){
                $tasks = $this->getDelayedTasksByCurrentUserOnDemand($project);
            }else{
                $tasks = $this->getTasksByCurrentUserFromDateOnDemand($fromDate, $project);
            }
            $appropriationService = InstanceProvider::getService()->getAppropriation();
            foreach ($tasks as &$task) {
                $task['BRANCH_INFO'] = $this->getBranchInfoByTaskId($task['ID']);
                $task['PATHS'] = $appropriationService->findPaths($task['ID'], $task['RESOURCE_T']);
                $task['ID'] = self::TASK_PREFIX . $task['ID'];
                $task['PROJECT_ID'] = self::PROJECT_PREFIX . $task['PROJECT_ID'];
            }
            $dataSet = $tasks;
        }else{
            if($fromDate == $today){
                $tasks = $this->getDelayedTasksByCurrentUser();
            }else{
                $tasks = $this->getTasksByCurrentUserFromDate($fromDate);
            }
            $projects = $this->getProjectByTasks($tasks);
            foreach ($projects as &$project) {
                $project['ID'] = self::PROJECT_PREFIX . $project['ID'];
            }
            $dataSet = $projects;
        }
        return $dataSet;
    }

    public function getUnatendedTasksFullDataByUserOnDemand($fromDate, $projectFilter){
        $today = date("d/m/Y");

        if(!empty($projectFilter)) {
            $project = str_replace("P", "", $projectFilter);
            if($fromDate == $today){
                $tasks = $this->getUnatendedDelayedTasksByCurrentUserOnDemand($project);
            }else{
                $tasks = $this->getUnatendedTasksByCurrentUserFromDateOnDemand($fromDate, $project);
            }
            $appropriationService = InstanceProvider::getService()->getAppropriation();
            foreach ($tasks as &$task) {
                $task['BRANCH_INFO'] = $this->getBranchInfoByTaskId($task['ID']);
                $task['PATHS'] = $appropriationService->findPaths($task['ID'], $task['RESOURCE_T']);
                $task['ID'] = self::TASK_PREFIX . $task['ID'];
                $task['PROJECT_ID'] = self::PROJECT_PREFIX . $task['PROJECT_ID'];
            }
            $dataSet = $tasks;
        }else{
            if($fromDate == $today){
                $tasks = $this->getDelayedUnatendedTasks();
            }else{
                $tasks = $this->getUnatendedTasksByUserFromDate($fromDate);
            }
            $projects = $this->getProjectByTasks($tasks);
            foreach ($projects as &$project) {
                $project['ID'] = self::PROJECT_PREFIX . $project['ID'];
            }
            $dataSet = $projects;
        }
        return $dataSet;
    }

	public function getTasksByCurrentUserFromDateOnDemand($dateTask,$projectId) {
		return InstanceProvider::getRepository()->getTask()->getTasksByUserFromDateOnDemand($this->environment->getCurrentUserId(), $dateTask, $projectId);
	}

	public function getUnatendedTasksByCurrentUserFromDateOnDemand($dateTask,$projectId) {
		return InstanceProvider::getRepository()->getTask()->getUnatendedTasksByUserFromDateOnDemand($dateTask, $projectId);
	}

	public function getTasksByCurrentUserFromDate($dateTask) {
		return $this->getTasksByUserFromDate($this->environment->getCurrentUserId(), $dateTask);
	}

	public function getDelayedTasksByCurrentUserOnDemand($projectId) {
		$tasks = InstanceProvider::getRepository()->getTask()->getDelayedTasksByUserOnDemand($this->environment->getCurrentUserId(), $projectId);
		return $tasks;
	}

	public function getUnatendedDelayedTasksByCurrentUserOnDemand($projectId) {
		$tasks = InstanceProvider::getRepository()->getTask()->getUnatendedDelayedTasksByCurrentUserOnDemand( $projectId);
		return $tasks;
	}
	public function getDelayedTasksByUser() {
		$userId = $this->environment->getCurrentUserId();
		return InstanceProvider::getRepository()->getTask()->getDelayedTasksByUser($userId);
	}

	public function getDelayedUnatendedTasks() {
		return InstanceProvider::getRepository()->getTask()->getDelayedUnatendedTasks($this->environment->getCurrentUserId());
	}

	public function getDelayedTasksByCurrentUser() {
		return $this->getDelayedTasksByUser();
	}
	public function saveNewTaskStatus($label,$typeTask) {
		$name = str_replace(' ','_',strtoupper($label));
		$typeTask = $this->entityManager->find('\Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskType',$typeTask);
		$taskStatus = InstanceProvider::getFactory()->getTaskStatus()->createTaskStatus($label,$name,$typeTask);
		$this->entityManager->persist($taskStatus);
		$this->entityManager->flush();
	}

	public function saveNewTaskType($control, $label, $isDismissible = 'Y', $frameOrder = null, $orderToView = null)
	{
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            if (empty($control)) {
                $control = str_replace(' ', '_', strtoupper($label));
            }
            $taskType = InstanceProvider::getFactory()->getTaskType()->createTaskType($label, $control, null, 'N', $isDismissible, $frameOrder,
                $orderToView);
            $this->entityManager->persist($taskType);
            if ($this->uniqueValidationsTaskType($frameOrder, $orderToView, $taskType, $control)) {
                $this->entityManager->flush();
            }
            $connection->commit();

            return array("ID" =>  $taskType->getId());
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
	}


    public function removeTaskType($id){
       InstanceProvider::getService()->getIssue()->removeTaskType($id);

    }

    public function updateTaskType($id, $control, $label, $allowCloseIssue,$orderToView,$dayNumbers,
            $taskClassification, $allowPlayTask, $issueClassification,$comment,$targetStatusId) {
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            /** @var Entities\TaskType $taskType */
            $taskType = InstanceProvider::getRepository()->getTaskType()->findOneBy(array('id'=>$id));

            $taskType = InstanceProvider::getFactory()->getTaskType()->saveTaskTypeIssue($taskType, $control, $label, $allowCloseIssue,
                $orderToView);

            $this->entityManager->persist($taskType);
            $this->entityManager->flush();

            $this->uniqueValidationsTaskType($orderToView, $taskType, $control);
            $this->saveTaskTypeClassification($taskClassification, $taskType, Entities\TaskTypeClassification::TYPE_TASK);
            $this->saveTaskTypeClassification($issueClassification, $taskType, Entities\TaskTypeClassification::TYPE_ISSUE);
            $taskType->setTimeAutoChangeStatus($dayNumbers);
            /** @var Entities\TaskStatus $targetStatus */
            $targetStatus = InstanceProvider::getRepository()->getTaskStatus()->findOneBy(array('id' => $targetStatusId));
            $taskType->setTaskStatusAutoChangeStatus($targetStatus);
            $taskType->setAllowPlayTask($allowPlayTask);
            $taskType->setDescriptionAutoChangeStatus($comment);

            $this->entityManager->persist($taskType);
            $this->entityManager->flush();
            $connection->commit();
            return array("ID" =>  $taskType->getId());
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    /**
     * @param Entities\Task $task
     */
	protected function recursiveDeleteTask($task){
        /** @var Entities\Task[] $childrenTasks */
		$childrenTasks = InstanceProvider::getRepository()->getTask()->findBy(array('parent' => $task->getId()));
		if (!empty($childrenTasks)) {
			foreach ($childrenTasks as $childTask) {
                $this->deleteTaskRelationship($task);
				$this->recursiveDeleteTask($childTask);
			}
		}
		InstanceProvider::getFactory()->getTask()->logicalDeleteTask($task);
		$this->entityManager->persist($task);
        $this->entityManager->flush($task);
	}

    public function deleteTask($task){
        InstanceProvider::getFactory()->getTask()->logicalDeleteTask($task);
    }

    public function deleteTaskRelated(Entities\Task $task)
    {
        $relatedTasks = InstanceProvider::getRepository()->getTaskRelationship()->findRelatedTaskIds($task->getId());
        if (!empty($relatedTasks)) {
            $taskIds = DevOpsUtilities::setSelectMultiInQuery($relatedTasks['RELATED_TASKS']);
            $taskIds[] = $relatedTasks['MAIN_TASK'];
            foreach ($taskIds as $taskId) {
                if($taskId != $task->getId()){
                	if($this->taskHasAppropriations($taskId)){
						throw new \Exception(Messages::ERR_NOT_DELETE_TASK_WITH_APPROPRIATIONS);
					} else {
                    	$relatedTask = InstanceProvider::getRepository()->getTask()->find($taskId);
                    	$this->deletingTask($relatedTask);
					}
                }
            }

        }
    }

    /**
     * @param $taskId
     * @throws \Doctrine\DBAL\ConnectionException
     */
    public function deleteTaskById($taskId) {
	    $taskRepository =  InstanceProvider::getRepository()->getTask();
        $projectService = InstanceProvider::getService()->getProject();
		$connection = $this->entityManager->getConnection();
		$connection->beginTransaction();
		try {
			if($this->taskHasAppropriations($taskId)){
				throw new \Exception(Messages::ERR_NOT_DELETE_TASK_WITH_APPROPRIATIONS);
			} else {
		        /** @var Entities\Task $task */
				$task = $taskRepository->find($taskId);

                /** @var Entities\Project $project */
                $project = InstanceProvider::getRepository()->getProject()->find($task->getProject());

                /** @var Entities\Project $baselineProject */
                $baselineProject = $projectService->getBaselineProjectRelatedOrNull($task->getProject());
                $hasBaselineTaskRelated = $baselineProject ? $taskRepository->hasRelationshipWithBaselineTaskInThisProject($task->getId(), $baselineProject->getId()) : null;

                if (!empty($hasBaselineTaskRelated) && $baselineProject &&  !$projectService->isProjectInPlanning($project)){
                    $baselineTask = $taskRepository->findBaselineTaskOfTaskCurrentProject($task->getId(), $baselineProject->getId());
                    /** @var Entities\Task $baselineTask */
                    $baselineTask = $taskRepository->find($baselineTask['ID']);
                    $baselineTask->setBaselineOccurrence(Entities\Task::BASELINE_OCCURRENCE_DELETED);
                    $baselineTask->setBaselineOccurrenceAt(new \DateTime());
                    $this->entityManager->persist($baselineTask);
                }

                $this->deleteTaskRelated($task);
                $this->deletingTask($task);
		        $connection->commit();
			}
        } catch (\Exception $e) {
			$connection->rollBack();
			throw $e;
		}
	}

	private function taskHasAppropriations($taskId)
	{
		$appropriations = InstanceProvider::getService()->getAppropriation()->validateApropriationByTask($taskId);
		return count($appropriations) > 0 ? true : false;
	}

	public function deleteIssues($issue) {
		$connection = $this->entityManager->getConnection();
		$connection->beginTransaction();
        $taskRepository = InstanceProvider::getRepository()->getTask();
        $issueToDelete = $taskRepository->find($issue);
        $reporter = $issueToDelete->getCreatedBy();
        if($reporter && $reporter != $this->environment->getCurrentUserId()){
            throw new \Exception(Messages::ERR_NOT_ISSUE_REPORTER);
        }
        try {
			$taskRelationship = InstanceProvider::getRepository()->getTaskRelationship()->findOneBy(array('task'=>$issue));
			$comments = InstanceProvider::getRepository()->getComments()->findBy(array('task'=>$issue));
			if(!($taskRelationship || count($comments) > 1)){
				$task = $taskRepository->find($issue);
				$this->recursiveDeleteTask($task);
			}else{
				throw new \Exception('Esta Issue já possui tarefas ou comentários associados e não pode ser excluída.');
			}
			$this->entityManager->flush();
			$connection->commit();
		} catch (\Exception $e) {
			$connection->rollBack();
			throw $e;
		}
	}
    public function fixBugGanttSetChildrenInArray($tasks,$task,&$returnTasks,&$childrenNotAdd){
        if ($task['PARENT'] && DevOpsUtilities::searchForIndex($task['PARENT'], $returnTasks, 'TASK_ID') === false) {
            $childrenNotAdd[$task['PARENT']][] = $task;
        } else if(DevOpsUtilities::searchForIndex($task['TASK_ID'], $returnTasks, 'TASK_ID') === false){
            $returnTasks[] = $task;
            if(!empty($childrenNotAdd[$task['TASK_ID']])){
                $moreChildren = $childrenNotAdd[$task['TASK_ID']];
                foreach ($moreChildren as $moreChild) {
                    $this->fixBugGanttSetChildrenInArray($tasks, $moreChild, $returnTasks, $childrenNotAdd);
                }
            }
        }
        if (!empty($task['CHILDREN'])) {
            foreach ($task['CHILDREN'] as $childIndex) {
                $child = $tasks[$childIndex];
                $this->fixBugGanttSetChildrenInArray($tasks, $child, $returnTasks, $childrenNotAdd);
            }
        }
    }

    public function fixBugGantt($tasks){
        $returnTasks = array();
        foreach ($tasks as &$task) {
            if(!empty($task['CHILDREN'])){
                $childrenIndex = array();
                foreach ($task['CHILDREN'] as $child) {
                    $index = DevOpsUtilities::searchForIndex($child,$tasks,'TASK_ID');
                    if($index !== false){
                        $childrenIndex[] = $index;
                    }
                }
                sort($childrenIndex);
                $task['CHILDREN'] = $childrenIndex;
            }
        }
        $childrenNotAdd = [];
        foreach ($tasks as &$task) {
            if(empty($task['CHILDREN']) && !$task['PARENT']){
                $returnTasks[] = $task;
            }else if(!empty($task['CHILDREN'])){
                $this->fixBugGanttSetChildrenInArray($tasks,$task,$returnTasks,$childrenNotAdd);
            }
        }
        return $returnTasks;
    }
    private function getRowsByTaskId($ids,$tasks){
        $rows = array();
        foreach ($tasks as $task) {
            if(in_array($task['TASK_ID'],$ids)){
                $rows[] = $task;
            }
        }
        return $rows;

    }

    private function convertDateCalInDateSrtGantt($dateCal){
        $day= substr($dateCal, 6);
        $month= substr($dateCal, 4,-2);
        $year= substr($dateCal, 0,-4);
        return $day."/".$month."/".$year;
    }

    private function getDateRealBeginAndEndGantt(&$task,&$tasks,$dataRealBeginCalculated = null,
                                                 $dataRealEndCalculated = null){

        if ($task['REAL_END']) {
            $dataRealEndCalculatedNew = implode(array_reverse(explode('/', $task['REAL_END'])));
            if($dataRealEndCalculatedNew > $dataRealEndCalculated){
                $dataRealEndCalculated = $dataRealEndCalculatedNew;
            }
        }
        if ($task['REAL_BEGIN']) {
            $dataRealBeginCalculatedNew = implode(array_reverse(explode('/', $task['REAL_BEGIN'])));
            if(is_null($dataRealBeginCalculated)){
                $dataRealBeginCalculated = $dataRealBeginCalculatedNew;
            }else if(!is_null($task['REAL_BEGIN']) && $dataRealBeginCalculatedNew < $dataRealBeginCalculated){
                $dataRealBeginCalculated = $dataRealBeginCalculatedNew;
            }
        }

        $childrenIds = $task['CHILDREN'];
        if(is_string($childrenIds)){
            $childrenIds = DevOpsUtilities::setSelectMultiInQuery($task['CHILDREN']);
        }

        if(!empty($childrenIds)){
            $children = self::getRowsByTaskId($childrenIds,$tasks);
            foreach ($children as $child) {
                list($dataRealBeginCalculatedNew,$dataRealEndCalculatedNew) =
                $this->getDateRealBeginAndEndGantt($child,$tasks,$dataRealBeginCalculated,$dataRealEndCalculated);

                if ($dataRealEndCalculatedNew > $dataRealEndCalculated) {
                    $dataRealEndCalculated = $dataRealEndCalculatedNew;
                }

                if(is_null($dataRealBeginCalculated)){
                    $dataRealBeginCalculated = $dataRealBeginCalculatedNew;
                }else if ($dataRealBeginCalculatedNew < $dataRealBeginCalculated) {
                    $dataRealBeginCalculated = $dataRealBeginCalculatedNew;
                }
            }

            if($dataRealBeginCalculated){
                $task['REAL_BEGIN'] = $this->convertDateCalInDateSrtGantt($dataRealBeginCalculated);
            }
            if($dataRealEndCalculated){
                $task['REAL_END'] = $this->convertDateCalInDateSrtGantt($dataRealEndCalculated);

            }

        }
        return array($dataRealBeginCalculated,$dataRealEndCalculated);
    }

    public function setBacklogParent(&$taskParent,&$tasks){
        $controlChildren = false;
        $controlBacklog = Entities\TaskStatus::TASK_STATUS_CONTROL_BACKLOG;
        foreach ($tasks as &$task) {
            if ($taskParent['TASK_ID'] == $task['PARENT']) {
                if ($task['CONTROL'] == $controlBacklog || isset($task['childBacklog'])) {
                    $taskParent['childBacklog'] = true;
                    if ($taskParent['COMPLETED'] == '100') {
                        $taskParent['COMPLETED'] = '99.99';
                    }
                }
                if ($task['CONTROL'] == $controlBacklog ) {
                    if(!$controlChildren){
                        $controlChildren = $controlBacklog;
                    }

                } else {
                    $controlChildren = true;
                }
            }
        }
        if ($controlChildren === $controlBacklog) {
            $taskParent['CONTROL'] = $controlBacklog;
        }
    }

	public function getTaskByProjectGantt($projectId,$showCompletedTasks,$conditions = array()) {
        $organizationId = $this->environment->getOrganizationId();
        $conditionsString = DevOpsUtilities::createQueryFiltered($conditions);

        $taskRepository = InstanceProvider::getRepository()->getTask();
		$tasks = $taskRepository->getTaskByProjectGantt($projectId,$organizationId,$conditionsString);
        $ganttScale = $taskRepository->getScaleByProjectGantt($projectId,$organizationId,$conditionsString);

        $index = count($tasks);
        if($index){
            while($index) {
                $task = $tasks[--$index];
                $task['SCALE'] = $ganttScale;
                $task['CHILDREN'] = DevOpsUtilities::setSelectMultiInQuery($task['CHILDREN']);
                $task['NAME_GANTT'] = " ";
                $this->getDateRealBeginAndEndGantt($task,$tasks);
                $this->setBacklogParent($task,$tasks);
                $task = $this->setColorGantt($task);
                $tasks[$index] = $task;
            }
        }

        $tasks = $this->fixBugGantt($tasks);
        $tasks = Utilities::fixParentsForTreeGrid($tasks,'TASK_ID');
        $tasks = Utilities::fixParentsForTreeGrid($tasks,'TASK_ID','PREDECESSOR');
        return $tasks;
	}

    public function findMergeRequestByTask($taskIds) {
        $taskIds = implode(',', $taskIds);
		return InstanceProvider::getRepository()->getTask()->findMergeRequestByTask($taskIds);
	}

	public function getTaskByDate($date) {
		$tasks = InstanceProvider::getRepository()->getTask()->getTaskByDate($date, $this->environment->getCurrentUserId());

		foreach($tasks as $taskKey => $task) {
			$hours = floor($task['TOTAL']);
			$minutes = 60*($task['TOTAL'] - floor($task['TOTAL']));

			$tasks[$taskKey]['TOTAL'] = "";
			if($hours > 0)
				$tasks[$taskKey]['TOTAL'] .= $hours . "h";
			if($minutes > 0)
				$tasks[$taskKey]['TOTAL'] .= floor($minutes) . "min";

			if($task['PERCENT_COMPLETED'] != "100%")
				$tasks[$taskKey]['END_HOUR'] = "Não concluída";
		}

		return $tasks;
	}

	public function getPercentAppropriated($date) {
		$tasks = InstanceProvider::getRepository()->getTask()->getTaskByDate($date, $this->environment->getCurrentUserId());
		$totalHours = 0;
		foreach($tasks as $task) {
			$totalHours += $task['TOTAL'];
		}
		if(!empty($tasks)) {
			$tasks[0]['DAY_HOURS'] = $tasks[0]['DAY_HOURS'] == 0 ? 8 : $tasks[0]['DAY_HOURS'];
			$percentAppropriated = ($totalHours * 100)/$tasks[0]['DAY_HOURS'];
		}
		else
			$percentAppropriated = 0;

		$hours = floor($totalHours);
		$minutes = 60*($totalHours - floor($totalHours));

		$totalHours = "";
		if($hours > 0)
			$totalHours .= $hours . "h";
		if($minutes > 0)
			$totalHours .= floor($minutes) . "m";

		return array("valor" => $percentAppropriated, "hint" => $totalHours);
	}

	public function getUserTasksName($project) {
		return $this->getUserTasksNameById($this->environment->getCurrentUserId(), $project);
	}

	public function getUserTasks() {
		return $this->getUserTasksById($this->environment->getCurrentUserId());
	}
	public function getUserTasksById($userId) {
		return InstanceProvider::getRepository()->getTask()->getUserTasksById($userId);
	}

	public function getUserTasksNameById($userId, $project) {
		return InstanceProvider::getRepository()->getTask()->getUserTasksNameById($userId, $project);
	}

	public function getAllTaskStatus() {
		return InstanceProvider::getRepository()->getTask()->getAllTaskStatus($this->environment->getCurrentOrganizationId());
	}

    public function getAllTaskStatusByIssueStatus() {
		return InstanceProvider::getRepository()->getTask()->getAllTaskStatus($this->environment->getCurrentOrganizationId());
    }

    public function updateTaskStatusByControl($taskId, $taskStatusControl,
                                              $initTask = null, $allocation = null,
                                              $taskResponse = null, $resource = null){
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskStatus $taskStatus */
            $taskStatus = $this->getTaskStatusByControl($taskStatusControl);
            $taskIds = $this->getTaskIdsToUpdate($taskId);
            $appropriationService = InstanceProvider::getService()->getAppropriation();
            $allocationService = InstanceProvider::getService()->getAllocation();
            $taskRepository = InstanceProvider::getRepository()->getTask();
            $resourceRepository = InstanceProvider::getRepository()->getResource();
            $allocationRepository = InstanceProvider::getRepository()->getAllocation();
            foreach($taskIds as $taskId) {
                /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $task */
                $task = $taskRepository->find($taskId);
                /** @var Entities\Resource $resourceEntity */
                $resourceEntity = $resourceRepository->find($resource);
                /** @var Entities\Allocation $allocationEntity */
                $allocationEntity = $allocationRepository->findOneBy(array('task' => $task, 'resource' => $resourceEntity));

                if (($taskStatusControl == self::TASK_STATUS_CONTROL_NOT_INITIALIZED) ||
                    ($taskStatusControl == self::TASK_STATUS_CONTROL_IN_DEVELOPING)) {
                    $allocationService->setPlayDateTask($taskId, $this->environment->getCurrentUserId(), $initTask);
                }

                if ($taskStatusControl === self::TASK_STATUS_CONTROL_CONCLUDED && !$task->validateIsNotAllowEndTask()) {
                    $allocationService->updateTaskStatusAllocationByTask($task, $taskStatus);
                    $task->setPercentCompleted(100);
                    if (!empty($taskResponse)) {
                        if(strlen($taskResponse) > 3800){
                            $taskResponse =  substr($taskResponse, 0, 3801) . self::MAXIMUM_CHARACTER_LIMIT_MESSAGE;
                            $task->setResponse($taskResponse);
                        } else {
                            $task->setResponse($taskResponse);
                        }
                    }
                } else {
                    $allocationService->updateTaskStatusAllocation($allocationEntity, $taskStatus);
                }

                $task->setTaskStatus($taskStatus);
                $this->entityManager->persist($task);
                $this->entityManager->flush();
                $this->applyOrderToIssuesByTask($task, false);
            }
            $this->entityManager->flush();
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }

    }

	public function updateTaskStatus($taskId, $taskStatusId) {
		$connection = $this->entityManager->getConnection();
		try {
			$connection->beginTransaction();
			$task = InstanceProvider::getRepository()->getTask()->find($taskId);
			$taskStatus = InstanceProvider::getRepository()->getTaskStatus()->find($taskStatusId);
			$task->setTaskStatus($taskStatus);
			$this->entityManager->persist($task);
			$this->entityManager->flush();

            $this->applyOrderToIssuesByTask($task);

			$connection->commit();
		} catch (\Exception $e) {
			$connection->rollBack();
			throw $e;
		}
	}

    public function updateTaskClassification($taskId, $taskClassificationId) {
        $connection = $this->entityManager->getConnection();
        try{
            $connection->beginTransaction();
            $task = InstanceProvider::getRepository()->getTask()->find($taskId);
            $taskClassificationControl = $task->getClassification()->getControl();
            if(!$taskClassificationControl){
            	$taskClassification = InstanceProvider::getRepository()->getClassification()->find($taskClassificationId);
            	$task->setClassification($taskClassification);
            }
            $this->entityManager->persist($task);
            $this->entityManager->flush();
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

	public function getTaskData($taskId) {
		return InstanceProvider::getRepository()->getTask()->getTaskData($taskId);
	}

	public function getTaskAppropriations($taskId) {
		return InstanceProvider::getRepository()->getTask()
			->getTaskAppropriations($taskId, $this->environment->getCurrentUserId());
	}

	public function getProjectByTasks($tasks) {
		$tasksIds = '';
		foreach($tasks as $task){
			$tasksIds .= $task['ID'].',';
		}
		$tasksIds .= '0';

		return InstanceProvider::getRepository()->getTask()->getProjectByTasks($tasksIds);
	}

	public function findIssueStatus($filterName = false, $filterOperator = false, $filterValue = false) {
		return InstanceProvider::getRepository()->getTask()->findIssueStatus($this->environment->getCurrentOrganizationId(), $filterName, $filterOperator, $filterValue);
	}

	public function findIssueClassification($filter = '') {
		return InstanceProvider::getRepository()->getTask()->findIssueClassification($this->environment->getCurrentOrganizationId(),
            $filter);
	}

	public function findIssueType($boardId) {
		return InstanceProvider::getRepository()->getTask()->findIssueType($this->environment->getCurrentOrganizationId(), $boardId);
	}

	public function saveNewIssueStatus($label, $description, $issueType) {
		$issueStatus = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskStatus')
			->findOneBy(array("organization" => $this->environment->getCurrentOrganizationId(), "label" => $label));
		$issueType = $this->entityManager->find('\Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskType', $issueType);
		if(empty($issueStatus)) {
			$issueStatus = InstanceProvider::getFactory()->getTaskStatus()->createTaskStatus($label, null, $issueType, self::IS_ISSUE);
			$this->entityManager->persist($issueStatus);
			$this->entityManager->flush();
		}
		else {
			return false;
		}

		return true;
	}

	public function saveNewIssueType($label, $description, $board) {
		$board = InstanceProvider::getRepository()->getProject()->find($board);
		$issueType = InstanceProvider::getRepository()->getTaskType() ->findOneBy(array("organization" => $this->environment->getCurrentOrganizationId(), "label" => $label));

		if(empty($issueType)) {
			$issueType = InstanceProvider::getFactory()->getTaskType()->createTaskType($label, null, $board, self::IS_ISSUE);
			$this->entityManager->persist($issueType);
			$this->entityManager->flush();
		}
		else {
			return false;
		}

		return true;
	}

    private function getIssueStatus($issueStatus)
    {
        $taskStatusRepository = InstanceProvider::getRepository()->getTaskStatus();
        if($issueStatus){
            return $taskStatusRepository->find($issueStatus);
        } else {
            return $taskStatusRepository->findOneBy(array('CONTROL' => Entities\TaskStatus::TASK_STATUS_CONTROL_BACKLOG));
        }
    }

	public function saveNewIssue($name, $classification, $issueStatus, $reporter, $dueDate, $priority, $percentCompleted,
								 $description, $board, $followers, $labels = null, $sponsor = null, $endDate = null,
                                 $productVersionId = null, $version = null, $projectId = null, $versionDev = null,
                                 $productVersionDevId = null, $isBoardIssue = null, $cditemtrab = null, $documents = null) {
		$connection = $this->entityManager->getConnection();
		try {
            
			$connection->beginTransaction();
            $name = substr($name, 0, 200);
			$issueStatus 	= $this->getIssueStatus($issueStatus);
			$board 			= $this->entityManager->find(Utilities::ENTITY_DIR.'Project', $board);
            if($endDate){
                $endDate = \DateTime::createFromFormat('d/m/Y', $endDate);
            }
            if($dueDate) {
                $dueDate = \DateTime::createFromFormat('d/m/Y', $dueDate);
            }else {
                $dueDate = new \DateTime();
            }
            if($priority){
                $priority = $this->entityManager->find(Utilities::ENTITY_DIR . 'Priority', $priority);
            }
            /** @var Entities\Classification $classification */
			$classification = InstanceProvider::getRepository()->getClassification()->find($classification);
			$reporter 		= InstanceProvider::getRepository()->getResource()->find($reporter);

             if($isBoardIssue){
                $productVersion = $this->getProductVersion($productVersionId);
                $productVersionDev = $this->getProductVersion($productVersionDevId);
            } else {
                $productVersion = $this->getProductVersion($productVersionId);
                $productVersionDev = $this->getProductVersionDev($projectId, $version, $productVersionId);
            }

            $issue  = InstanceProvider::getFactory()->getTask()->createTask($board, null, $name, $this->environment->getCurrentTime(), $dueDate, self::ACTIVE, $description, $percentCompleted, null, null, $issueStatus, null, null, self::IS_ISSUE, 'N', $classification, null, $priority, null, null, $endDate, null, null, $productVersion, $productVersionDev);

            $this->entityManager->persist($issue);

            $allocationService = InstanceProvider::getService()->getAllocation();
			if(!empty($reporter)){
				$allocationService->createAllocation($issue, $reporter, Allocation::REPORTER);
			}

			if(!empty($sponsor)){
				$sponsorEntity = $this->entityManager->find(Utilities::ENTITY_DIR.'Resource', $sponsor);
				$allocationService->createAllocation($issue, $sponsorEntity, Allocation::SPONSOR);
			}

            $this->setFollowers($followers, $issue);

            if(is_array($labels)){
				$this->setLabels($labels,$issue);
			}

			$this->entityManager->flush();

            $commentType = InstanceProvider::getRepository()->getCommentType()->findOneBy(array('control' => Entities\CommentType::TYPE_OPEN_ISSUE));
			$fistComments = InstanceProvider::getFactory()->getComments()->createComments($issue,self::FIRST_COMMENTS, $commentType);
            $issue->setCurrentTaskStatusIn(Utilities::getEnvironment()->getCurrentTime());
            $this->entityManager->persist($issue);
            $this->entityManager->persist($fistComments);
            $this->viewCommentsSaveData($issue->getId());
            $this->entityManager->flush();

            $reporterAllocation = $allocationService->getAllocationByUserAndTask($reporter->getId(), $issue->getId());
            if(!empty($reporterAllocation)){
                $reporterAllocation->setLastPostView($this->environment->getCurrentTime());
                $this->entityManager->persist($reporterAllocation);
                $this->entityManager->flush();
            }

            InstanceProvider::getService()->getTaskWorkitem()->save($issue->getId(), $cditemtrab);
            if($documents){
                $attachmentService = InstanceProvider::getService()->getAttachment();
                $attachmentService->updateDocumentTask($documents, $issue, null);
                $this->entityManager->flush();
            }
			$connection->commit();
            \Service\Integration\IntegrationCP::createCall($name,$reporter,$description,
                $issue->getId(),$issue->getProject()->getId(),$classification->getId());
            $this->sendOpenIssueMail($issue);

            
			return $issue->getId();
		}
		catch(\Exception $e){
			$connection->rollBack();
			throw $e;
		}
	}

    public function getProductVersion($productVersionId)
    {
        if($productVersionId){
            return InstanceProvider::getRepository()->getProductVersion()->find($productVersionId);
        }
        return null;
    }

    public function getProductVersionDev($projectId, $version, $productVersionId)
    {
        if($productVersionId){
            $productVersionDevId = $this->getProductVersionDevFromProductVersion($version, $projectId);
            if($productVersionDevId){
                return InstanceProvider::getRepository()->getProductVersion()->find($productVersionDevId);
            } else {
                return null;
            }
        } else {
            return null;
        }
    }

    public function getProductVersionDevFromProductVersion($version, $projectId)
    {
        $product = InstanceProvider::getRepository()->getProduct()->findOneBy(array('project' => $projectId));
        $organizationId = InstanceProvider::getEnvironment()->getOrganizationId();
        $versions = explode('.', $version);
        $version = $versions[0];
        $revision = $versions[1];
        $productVersionsLts = InstanceProvider::getRepository()->getProductVersion()->findLtsCurrentProductVersionFromVersion($product->getId(), $organizationId, $version, $revision);
        return !empty($productVersionsLts) ? $productVersionsLts[0]['ID'] : null;
    }

    public function sendOpenIssueMail(Entities\Task $issue)
    {
        if($issue->getProject()->getNotifyIssueOpening() == Entities\Project::IS_NOTIFY_ISSUE){
            $commenterId = $this->environment->getCurrentUserId();
            $emails    = InstanceProvider::getRepository()->getTask()->getManagersEmailsByIssueId($issue->getId(), $commenterId);
            $issueName = $issue->getName();
            $reporter  = DevOpsUtilities::getResourceNameForSendingMail($issue->getId(), Entities\Allocation::TYPE_REPORTER);
            $sponsor   = DevOpsUtilities::getResourceNameForSendingMail($issue->getId(), Entities\Allocation::TYPE_SPONSOR);
            $organizationName = InstanceProvider::getRepository()->getOrganization()->find($issue->getOrganization())->getName();
            $commenterName = $this->environment->getCurrentUser()->getName();
            $date = $this->environment->getCurrentTime()->format(Utilities::FORMAT_DATE_DEFAULT_NO_SECS);
            $issueTitle  = $issue->getId() . " - " . $issueName;
            $productName = $issue->getProject()->getName();

            foreach ($emails as $email) {
                InstanceProvider::getMailIssueMail()->sendOpenIssueMail($email['EMAIL'], $issueTitle, $commenterName, $date, $issue->getDescription(),
                    $reporter, $sponsor, $issueName, $issue->getId(), $productName, $organizationName);
            }
        }
    }

    public function setLabelAccompanyingIssue($issueId,$labelIds, $labelIdsNotEditable){
        $issue = InstanceProvider::getRepository()->getTask()->find($issueId);
        $this->setLabelsIssue($labelIds, $labelIdsNotEditable, $issue);
    }

    public function saveIssueAndRelationship(
        $name, $classification, $issueStatus, $reporter, $dueDate, $priority, $percentCompleted, $description,
        $board, $followers, $labels, $sponsor, $issueDependentId, $endDate
    ){
        $issueId = $this->saveNewIssue($name, $classification, $issueStatus, $reporter, $dueDate, $priority,
            $percentCompleted, $description, $board, $followers, $labels, $sponsor, $endDate);

        $this->onlyCreateTaskRelationships($issueId, array($issueDependentId), Entities\TaskRelationship::ISSUE_WITH_ISSUE_TYPE);
    }

	public function saveIssue(
		$id, $name, $classification, $issueStatus, $reporter, $dueDate, $priority,
		$percentCompleted, $description, $followers, $endDate, $board = null,
        $labels = null, $sponsor = null, $labelsNotEditable = array(), $isRelevant = self::NOT_RELEVANT, $productVersionId = null,
        $version = null, $projectId = null, $isBoardIssue = null, $productVersionDevId = null, $issueResponse = null, $cditemtrab = []
    ){

        $connection = $this->entityManager->getConnection();
        try{
            $connection->beginTransaction();

            $name = substr($name, 0, 200);
            $resourceRepository = InstanceProvider::getRepository()->getResource();
            $issue          = InstanceProvider::getRepository()->getTask()->find($id);
            $issueStatus    = $this->entityManager->find(Utilities::ENTITY_DIR . 'TaskStatus', $issueStatus);
            $reporter       = empty($reporter) ? null : $resourceRepository->find($reporter);
            $dueDate        = empty($dueDate) ?  $this->environment->getCurrentTime() : \DateTime::createFromFormat('d/m/Y', $dueDate);
            $endDate        = empty($endDate) ?  null : \DateTime::createFromFormat('d/m/Y', $endDate);
            $classification = $this->entityManager->find(Utilities::ENTITY_DIR . 'Classification', $classification);

            if($priority){
                $priority = $this->entityManager->find(Utilities::ENTITY_DIR . 'Priority', $priority);
            }

            $frontReload = false;
            if($issue->getName() != $name || $issue->getDescription() != $description ||
               $issue->getTaskStatus()->getId() != $issueStatus->getId() ||
                $issue->getClassification()->getId() != $classification->getId()){
                $frontReload = true;
            }

            if($issue->getProductVersionDev()){
                if($issue->getProductVersionDev()->getId() != $productVersionDevId){
                    $frontReload = true;
                }
            } else if($productVersionDevId){
                $frontReload = true;
            }


            if($priority && $issue->getPriority()) {
                if($priority->getId() != $issue->getPriority()->getId()){
                    $frontReload = true;
                }
            } else if($priority && !$issue->getPriority()) {
                $frontReload = true;
            } else if(!$priority && $issue->getPriority()) {
                $frontReload = true;
            }

            if($board){
                $board = $this->entityManager->find(Utilities::ENTITY_DIR . 'Project', $board);
            }else{
                $board = $issue->getProject();
            }
            if($isBoardIssue){
                $productVersion = $this->getProductVersion($productVersionId);
                $productVersionDev = $this->getProductVersion($productVersionDevId);
            } else {
                $productVersion = $this->getProductVersion($productVersionId);
                $productVersionDev = $this->getProductVersion($productVersionDevId);
            }

            $issue = InstanceProvider::getFactory()->getTask()->updateTask($issue, $board, null, $name, $this->environment->getCurrentTime(), $dueDate, self::ACTIVE, $description, $percentCompleted, null, null, $issueStatus, null, null, 'N', self::IS_ISSUE, $classification,null,	$priority, null, null, null, $productVersion, $productVersionDev, $issueResponse);
            $issue->setEndDate($endDate);
			$this->entityManager->persist($issue);

            $this->setLabelsIssue($labels, $labelsNotEditable, $issue);

            $allocationService = InstanceProvider::getService()->getAllocation();
            $allocationRepository = InstanceProvider::getRepository()->getAllocation();
			if(!empty($reporter)){
                /** @var Entities\Allocation $allocation */
				$allocation = $allocationRepository->findOneBy(array('task' => $issue->getId(), 'type' => Allocation::REPORTER));
				if(empty($allocation)){
					$allocationService->createAllocation($issue, $reporter, Allocation::REPORTER, null, true);
				}
				else{
					$allocationService->updateAllocation($allocation, $issue, $reporter, Allocation::REPORTER, null, true);
				}
			}

			if(!empty($sponsor)){
				$sponsorEntity	= $this->entityManager->find(Utilities::ENTITY_DIR.'Resource', $sponsor);
				$allocation = $allocationRepository->findOneBy(array('task' => $issue->getId(), 'type' => Allocation::SPONSOR));
				if(empty($allocation)){
					$allocationService->createAllocation($issue, $sponsorEntity, Allocation::SPONSOR, null, true);
				} else {
					$allocationService->updateAllocation($allocation, $issue, $sponsorEntity, Allocation::SPONSOR, null, true);
				}
			}

            foreach ($followers as $eachResource) {
                if(!empty($eachResource)){
                    $resourceProject = $resourceRepository->find($eachResource);
                    $allocation = $allocationRepository->findBy(array("task" => $issue->getId(), "resource" => $resourceProject->getId(),
						'type' => Allocation::FOLLOWER));
                    if (empty($allocation)) {
                        $allocationService->createAllocation($issue, $resourceProject, Allocation::FOLLOWER, null, true);
                    }
                }
            }
            /** @var Entities\Allocation[] $allocation */
            $allocation = $allocationRepository->findBy(array("task" => $issue->getId(), 'type' => Allocation::FOLLOWER));

            foreach ($allocation as $eachAllocation) {
                if(!in_array($eachAllocation->getResource()->getId(), $followers)){
                    $this->entityManager->remove($eachAllocation);
                }
            }

            $user = $this->environment->getCurrentUserId();
            if(!empty($user)){
                $resourceEntity	= $this->entityManager->find(Utilities::ENTITY_DIR.'Resource', $user);
                $allocation = $allocationRepository->findOneBy(array('task' => $issue->getId(), 'type' => Allocation::FOLLOWER, 'resource' => $user));
                if($allocation){
                    $allocationService->updateAllocation($allocation, $issue, $resourceEntity, Allocation::FOLLOWER, null, true, $isRelevant);
                } else if($isRelevant == self::RELEVANT) {
                    $allocationService->createAllocation($issue, $resourceEntity, Allocation::FOLLOWER, null, true, $isRelevant);
                }
            }

            InstanceProvider::getService()->getTaskWorkitem()->save($issue->getId(), $cditemtrab);

            $this->entityManager->flush();
			$connection->commit();


            return array('ID' => $issue->getId(), 'RELOAD' => $frontReload);
        }
		catch(\Exception $e){
			$connection->rollBack();
			throw $e;
		}
	}

	public function saveEditedIssue($newIssueType, $issueId, $sendMail = false, $description = null) {
		$connection = $this->entityManager->getConnection();
		try{
			$connection->beginTransaction();
            /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $issue */
			$issue =InstanceProvider::getRepository()->getTask()->find($issueId);
			/** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskType $newIssueType */
			$newIssueType = $this->entityManager->find(Utilities::ENTITY_DIR . 'TaskType',$newIssueType);
			$initialStatus = $newIssueType->getInitialTaskStatus();
			if(empty($initialStatus)){
				$newIssueStatus = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'TaskStatus')
					->findOneBy(array("taskType" => $newIssueType));
			}else{
				$newIssueStatus = $initialStatus;
			}
			if ($issue->getTaskStatus() != $newIssueStatus){
				$issue->setCurrentTaskStatusIn((new \DateTime()));
			}
            $isbudget = $issue->getClassification()->getId() == 8;
			$issue->setTaskStatus($newIssueStatus);
			$this->entityManager->persist($issue);
			$this->entityManager->flush();
            if($newIssueStatus->getControl() == Entities\TaskStatus::TASK_STATUS_CONTROL_HOMOLOGATED){
                $localBudget = \Util\Instance\Repository::getLocalBudget();
                $organizationId = Utilities::getEnvironment()->getOrganizationId();
                if($isbudget &&
                    empty($localBudget->getIssueBudgetSMAccepted($issueId)) &&
                    !empty($localBudget->getIssueBudgetSMConcluded($issueId)))
                {
                    Integration\IntegrationSM::integrationSMCancel($issueId,$organizationId);
                }
                if($sendMail){
                    Integration\IntegrationSM::integrationSMApproveContract($issueId,$organizationId);
                    $this->sendMailOnAvailableIssue($issueId, $description);
                }
            }
            $connection->commit();
        }
		catch(\Exception $e){
			$connection->rollBack();
			throw $e;
		}
	}

    public function sendMailOnReopenIssue($issueId)
    {
        $issue          = InstanceProvider::getRepository()->getTask()->find($issueId);
        $projectId      = $issue->getProject()->getId();
        $issueName      = $issue->getName();
        /** @var Entities\Resource $reporter */
        $reporter       = DevOpsUtilities::getResourceNameForSendingMail($issue, Entities\Allocation::TYPE_REPORTER);
        $reporterId      = DevOpsUtilities::getResourceIdByType($issue, Entities\Allocation::TYPE_REPORTER);
        $sponsor        = DevOpsUtilities::getResourceNameForSendingMail($issue, Entities\Allocation::TYPE_SPONSOR);
        $organizationName = InstanceProvider::getRepository()->getOrganization()->find($issue->getOrganization())->getName();
        $managers   	= InstanceProvider::getRepository()->getProject()->getManagersByProject($projectId);
        $managersEmails = array();

        $resourceRepository = InstanceProvider::getRepository()->getResource();
        foreach($managers as $manager){
            array_push($managersEmails,
                $resourceRepository->find($manager["RESOURCE_C"])->getEmail());
        }
        $commenterName  = $this->environment->getCurrentUser()->getName();
        $issueTitle     = $issueId." - ".$issueName;
        $date           = $this->environment->getCurrentTime()->format(Utilities::FORMAT_DATE_DEFAULT_NO_SECS);
        $productName    = $issue->getProject()->getName();
        $lastComment    = InstanceProvider::getRepository()->getComments()->getIssueLastCommentByCreator($issueId, $reporterId, $issue->getOrganization());
        foreach($managersEmails as $email){
            if($email){
                InstanceProvider::getMailIssueMail()->sendReopentMail($email, $commenterName, $date, $issueTitle, $issueName, $reporter,
                    $sponsor, $issueId, $productName, $lastComment, $organizationName);
            }
        }
    }

    public function sendMailOnAvailableIssue($issueId, $description = null){
        $requestHTMLBuild = true;
        $taskRepository = InstanceProvider::getRepository()->getTask();
        $resourceId    = $this->environment->getCurrentUserId();
        $emails         = $taskRepository->getRelatedResourcesEmailsByIssueId($issueId, $resourceId);
        $issue          = $taskRepository->find($issueId);
        $issueName      = $issue->getName();
        $reporter       = DevOpsUtilities::getResourceNameForSendingMail($issueId, Entities\Allocation::TYPE_REPORTER);
        $sponsor        = DevOpsUtilities::getResourceNameForSendingMail($issueId, Entities\Allocation::TYPE_SPONSOR);
        $organizationName = InstanceProvider::getRepository()->getOrganization()->find($issue->getOrganization())->getName();
        $commenterName  = $this->environment->getCurrentUser()->getName();
        $issueTitle     = $issueId." - ".$issueName;
        $taskIds        = $taskRepository->getRelationedTasks($issueId);
        $date           = $this->environment->getCurrentTime()->format(Utilities::FORMAT_DATE_DEFAULT_NO_SECS);
        $productName    = $issue->getProject()->getName();
        $mailMessage = '';
        $additionalMessage = '(sem necessidade de implementação)';

        foreach ($taskIds as $taskId) {
            /** @var Entities\Task $task */
            $task = $taskRepository->find($taskId['ID']);
            if($task){
                $mailMessage .= InstanceProvider::getMailIssueMail()->buildBodyMessage( $task);
            }
        }
        if(!empty($emails)){
            foreach($emails as $email){
                InstanceProvider::getMailIssueMail()->sendAvailableMail($email['EMAIL'],$issueTitle , $commenterName, $date, $mailMessage,
                    '.', $reporter, $sponsor, $issueName, $issueId, $productName, $description, $organizationName, $requestHTMLBuild, null, false, false, null, $additionalMessage);
            }
        }
    }

    protected function applyOrderToIssuesByTaskSingleRelatedTask(Entities\Task $task){
        $onlyTask = false;
        if($this->getLengthOfRelatedTasks($task) == 1) {
            $this->applyOrderToIssuesByTask($task, true, false, true);
            $onlyTask =  true;
        }
        return $onlyTask;
    }

    public function findTaskTypesForFraming(Entities\Task $issue){
        return InstanceProvider::getService()->getIssue()->findTaskTypesForFraming($issue);
    }


    public function applyOrderToIssuesByTask(Entities\Task $task, $isFlush = true, $insertTaskInIssue = false, $delete = false)
    {
        $taskRelationships = InstanceProvider::getRepository()->getTaskRelationship()->findBy(array('taskRelationship' => $task, 'type' => TaskPrecursor::RELATIONSHIP_ISSUE_TYPE));
        InstanceProvider::getService()->getAppropriation()->setStatusIssues($insertTaskInIssue, $taskRelationships,$isFlush, $delete);

    }

    private function getRelatedTasksByTasks(Entities\Task $task)
    {
        $originTask = $task;
        $taskToUse = false;
        $taskRelationship = InstanceProvider::getRepository()->getTaskRelationship()->findOneBy(array('taskRelationship' => $task, 'type' => TaskPrecursor::RELATIONSHIP_ISSUE_TYPE));
        /** @var Entities\TaskRelationship[] $relatedTasks */
        if(!empty($taskRelationship)) {
            $relatedTasks = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'TaskRelationship')
                ->findBy(array('task' => $taskRelationship->getTask(), 'type' => TaskPrecursor::RELATIONSHIP_ISSUE_TYPE));
            if (count($relatedTasks) > 1) {
                $task = $this->getFirstTaskRelationship($relatedTasks, $originTask);
            }
            $taskToUse = $task->getTaskRelationship();
        }
        return $taskToUse;
    }

    private function getLengthOfRelatedTasks(Entities\Task $task)
    {
        /** @var Entities\TaskRelationship[] $taskRelationships */
        $taskRelationshipRepository = InstanceProvider::getRepository()->getTaskRelationship();
        $taskRelationship = $taskRelationshipRepository->findOneBy(array('taskRelationship' => $task, 'type' => TaskPrecursor::RELATIONSHIP_ISSUE_TYPE));
        $relatedTasks = $taskRelationshipRepository->findBy(array('task' => $taskRelationship->getTask(), 'type' => TaskPrecursor::RELATIONSHIP_ISSUE_TYPE));
        return count($relatedTasks);
    }

    public function findTaskStatusByIssueStatus($issueStatusId){
        return  InstanceProvider::getRepository()->getTaskStatus()->findBy(array('taskType' => $issueStatusId));
    }

    public function findIssueOrderToFrame($issueStatusId){
        return  InstanceProvider::getRepository()->getTaskType() ->findIssueOrderToFrame($issueStatusId);
    }

    public function applyOrderToIssueById($issueId)
    {   /** @var Entities\Task $issue */
        $issue = $this->entityManager->find(Utilities::ENTITY_DIR . 'Task', $issueId);
        $taskTypeForFraming =  $this->findTaskTypesForFraming($issue);
        $appropriationService = InstanceProvider::getService()->getAppropriation();
        $projectTypeForFraming =  $appropriationService->findProjectTypesForFraming($issue);
        $notHomologated = !$appropriationService->validateIssueHomologated($issue);
        if($taskTypeForFraming && $notHomologated){
            $appropriationService->setStatusFraming($issue, $taskTypeForFraming);
            $this->entityManager->persist($issue);
            $this->entityManager->flush();
        }else if($projectTypeForFraming && $notHomologated) {
            $appropriationService->setStatusFraming($issue, $projectTypeForFraming);
            $this->entityManager->persist($issue);
            $this->entityManager->flush();
        }else if($appropriationService->validateFraming($issue) && $taskTypeForFraming) {
            /** @var Entities\TaskStatus $taskStatus */
            if($issue->getTaskStatus()
                && $issue->getTaskStatus()->getTaskType()
                && $issue->getTaskStatus()->getTaskType()->getId() == $taskTypeForFraming['ID']){
                $taskStatus = null;
            }else if($taskTypeForFraming['INITIAL_TASK_STATUS']){
                $taskStatus = InstanceProvider::getRepository()->getTaskStatus()->find($taskTypeForFraming['INITIAL_TASK_STATUS']);
            }else{
                $taskStatus = InstanceProvider::getRepository()->getTaskStatus()->findOneBy(array('taskType' => $taskTypeForFraming['ID']));
            }
            if($taskStatus){
                $issue->setTaskStatus($taskStatus);
                $this->entityManager->persist($issue);
                $this->entityManager->flush();
            }
        }else{
            $frameOrder = InstanceProvider::getRepository()->getTask()->getIssueFrameOrderBasedOnRelatedTasks($issue->getId());
            if ($frameOrder != null) {
                /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskType $taskType */
                $taskType = InstanceProvider::getRepository()->getTaskType() ->findOneBy(array('orderToFrame' => $frameOrder,
                    'organization' => $this->environment->getCurrentOrganizationId()));

                /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskStatus $newTaskStatus */
                $newTaskStatus = InstanceProvider::getRepository()->getTaskStatus()->findOneBy(array('taskType' => $taskType->getId(),
                    'isIssue' => 'Y',
                    'organization' => $this->environment->getCurrentOrganizationId()));

                if($issue->getTaskStatus() != $newTaskStatus){
                    $issue->setCurrentTaskStatusIn((new \DateTime()));
                }

                $issue->setTaskStatus($newTaskStatus);


                $this->entityManager->persist($issue);
                $this->entityManager->flush();
            }
        }
    }

    public function findBoardIssues($boardId, $conditions, $structureIds, $projectStatusIds) {
	    $filterComments = '';
	    $filterRelevance = '';
        $filterProductVersion = '';
        $cditemtrabFilter = '';
	    foreach ($conditions as $key =>$filter){
	         if($filter['columnName'] == 'CREATED_AT'){
		        $value = $filter['value'];

		        if($value){
                    $filterRelevance .= " AND CREATION_DATE BETWEEN '" . $value[0] . "'AND '" . $value[1] . "'";
                }
                 unset($conditions[$key]);
		    }

            if ($filter['columnName'] == 'RELEVANCE'){
	            $value = $filter['value'];
	            if($value){
	                $filterRelevance .= " AND (I.RELEVANCE >=  $value) ";
	            }
	            unset($conditions[$key]);
	        }else if ($filter['columnName'] == 'LAST_DATE_VIEW' && $filter['value'] == '%Y%' ){
	            $filterComments .= ' AND (I.CREATED_AT > I.LAST_POST_VIEW  OR I.LAST_POST_VIEW IS NULL) ';
	            unset($conditions[$key]);
	        }else if($filter['columnName'] == 'LAST_DATE_VIEW' && $filter['value'] == '%N%' ){
	            unset($conditions[$key]);
	        } else if ($filter['columnName'] == 'LABEL_CHECK' && $filter['value'] == '%Y%' ){
                $MeetingProjectIds = $this->getImInMeetingProjectIds();
                if($boardId && is_array($boardId)) {
                    $MeetingProjectIds = array_merge($boardId,$MeetingProjectIds);
                } else {
                    $MeetingProjectIds[] = $boardId;
                }
                $projectIdString = DevOpsUtilities::getStringForArray($MeetingProjectIds);
                $filterComments .= " AND (I.ID NOT IN (SELECT TL.TASK FROM LABEL LA JOIN TASK_LABEL TL ON TL.LABEL = LA.ID WHERE (RESOURCE_T = :userId OR PROJECT IN ( $projectIdString ))))";
                unset($conditions[$key]);
            } else if($filter['columnName'] == 'LABEL_CHECK' && $filter['value'] == '%N%' ){
	            unset($conditions[$key]);
	        } else if($filter['columnName'] == 'VERSION'){
                $filterProductVersion = "AND UPPER({$filter['columnName']}) LIKE '{$filter['value']}'";
            } else if ($filter['columnName'] == 'VERSION_DEV'){
                $filterProductVersion .= " AND UPPER({$filter['columnName']}) LIKE '{$filter['value']}'";
            } else if ($filter['columnName'] == 'CDITEMTRAB') {
                $cditemtrabString = DevOpsUtilities::getStringForArray($filter['value']);
                $cditemtrabString = "'" . str_replace(",", "', '", $cditemtrabString) . "'";
                $cditemtrabFilter = "AND TW.CDITEMTRAB IN ($cditemtrabString)";
            } else if ($filter['columnName'] == 'PRIORITY'){
                $priorityIdString = $filter['value'];
                $priorityIdString = DevOpsUtilities::getStringForArray($priorityIdString);
                $filterProductVersion .= " AND I.PRIORITY IN ($priorityIdString)";
            }
	    }
        Utilities::getValueInFilterDataSourceByName($conditions, 'VERSION', true);
        Utilities::getValueInFilterDataSourceByName($conditions, 'VERSION_DEV', true);
        Utilities::getValueInFilterDataSourceByName($conditions, 'CDITEMTRAB', true);

        $filterStructureId = empty($structureIds) ? '' : $this->getQueryFilterStructureId($structureIds);
        $filterProjectStatusIds = empty($projectStatusIds) ? '' : $this->getQueryFilterProjectStatusIds($projectStatusIds);

	    $issues = InstanceProvider::getRepository()->getTask()->findBoardIssues($this->environment->getCurrentOrganizationId(),
	        $this->environment->getCurrentUserId(), $boardId, $conditions,$filterComments, $filterRelevance, $filterProductVersion,
            $filterStructureId, $filterProjectStatusIds, $cditemtrabFilter);
	    foreach ($issues as &$issue) {
	        $issue['ID'] = intval($issue['ID']);
	    }

	    return $issues;
	}

	public function findKanbanItems($boardId, $conditions, $structureIds, $projectStatusIds) {
		$filterComments = '';
        $filterRelevance = '';
        $filterTaskStatus = '';
        $filterProductVersion = '';
        $cditemtrabFilter = '';
		foreach ($conditions as $key =>$filter){

		    if($filter['columnName'] == 'CREATED_AT'){
		        $value = $filter['value'];

		        if($value){
                    $filterRelevance .= " AND CREATED_AT BETWEEN '" . $value[0] . "'AND '" . $value[1] . "'";
                }
                 unset($conditions[$key]);
		    }
            if ($filter['columnName'] == 'RELEVANCE'){
                $value = $filter['value'];
                if($value){
                    $filterRelevance .= " AND (I.RELEVANCE >=  $value) ";
                }
                unset($conditions[$key]);
            } else if ($filter['columnName'] == 'FRAMING'){
                $value = implode(',', $filter['value']);
                if($value){
                    $filterTaskStatus = " AND (FRAMING IN  ($value)) ";
                }
                unset($conditions[$key]);
            } else if ($filter['columnName'] == 'LAST_DATE_VIEW' && $filter['value'] == '%Y%' ){
				$filterComments .= ' AND (COM.CREATED_AT > ALO.LAST_POST_VIEW  OR ALO.LAST_POST_VIEW IS NULL) ';
				unset($conditions[$key]);
			} else if ($filter['columnName'] == 'LAST_DATE_VIEW' && $filter['value'] == '%N%' ){
				unset($conditions[$key]);
			} else if ($filter['columnName'] == 'LABEL_CHECK' && $filter['value'] == '%Y%' ){
                $MeetingProjectIds = $this->getImInMeetingProjectIds();
                if($boardId && is_array($boardId)) {
                    $MeetingProjectIds = array_merge($boardId,$MeetingProjectIds);
                } else {
                    $MeetingProjectIds[] = $boardId;
                }
                $projectIdString = DevOpsUtilities::getStringForArray($MeetingProjectIds);
				$filterComments .= " AND (I.ID NOT IN (SELECT TL.TASK FROM LABEL LA JOIN TASK_LABEL TL ON TL.LABEL = LA.ID WHERE (RESOURCE_T = :userId OR PROJECT IN ( $projectIdString ))))";
				unset($conditions[$key]);
			} else if($filter['columnName'] == 'LABEL_CHECK' && $filter['value'] == '%N%' ){
				unset($conditions[$key]);
			} else if($filter['columnName'] == 'VERSION'){
                $filterProductVersion = "AND UPPER({$filter['columnName']}) LIKE '{$filter['value']}'";
            } else if ($filter['columnName'] == 'VERSION_DEV'){
                $filterProductVersion .= " AND UPPER({$filter['columnName']}) LIKE '{$filter['value']}'";
            } else if ($filter['columnName'] == 'CDITEMTRAB') {
                $cditemtrabString = DevOpsUtilities::getStringForArray($filter['value']);
                $cditemtrabString = "'" . str_replace(",", "', '", $cditemtrabString) . "'";
                $cditemtrabFilter = "AND TW.CDITEMTRAB IN ($cditemtrabString)";
            }
		}
        $filterStructureId = empty($structureIds) ? '' : $this->getQueryFilterStructureId($structureIds);
        $filterProjectStatusIds = empty($projectStatusIds) ? '' : $this->getQueryFilterProjectStatusIds($projectStatusIds);

        Utilities::getValueInFilterDataSourceByName($conditions, 'VERSION', true);
        Utilities::getValueInFilterDataSourceByName($conditions, 'VERSION_DEV', true);
        Utilities::getValueInFilterDataSourceByName($conditions, 'CDITEMTRAB', true);

        return InstanceProvider::getRepository()->getTask()
                                                ->findKanbanIssues($this->environment->getCurrentOrganizationId(),
                                                                   $this->environment->getCurrentUserId(), $boardId,
                                                                   $conditions, $filterComments, $filterRelevance,
                                                                   $filterTaskStatus, $filterProductVersion,
                                                                    $filterStructureId, $filterProjectStatusIds, $cditemtrabFilter);
	}

	private function getQueryFilterStructureId($structureIds)
    {
        $structureIdsString = DevOpsUtilities::getStringForArray($structureIds);
        $structures = InstanceProvider::getRepository()->getStructure()->findAllStructuresChildrenById($structureIdsString);
        $structuresArray = DevOpsUtilities::getArrayFromArrayAssociative($structures);
        $structureIdsString = DevOpsUtilities::getStringForArray($structuresArray);
        return "AND PRO.STRUCTURE IN ($structureIdsString)";
    }

    private function getQueryFilterProjectStatusIds($projectStatusIds)
    {
        $structureIdsString = DevOpsUtilities::getStringForArray($projectStatusIds);
        return "AND PRO.PROJECT_STATUS IN ($structureIdsString)";
    }

    public function findAllIssues($conditions) {
        $filterComments = '';
        $filterRelevance = '';
        $filterTaskStatus = '';
        $filterProductVersion = '';
        $cditemtrabFilter = '';
        foreach ($conditions as $key =>&$filter){

            if ($filter['columnName'] == 'RELEVANCE'){
                $value = $filter['value'];
                if($value){
                    $filterRelevance .= " AND (I.RELEVANCE >=  $value) ";
                }
                unset($conditions[$key]);
            } else if ($filter['columnName'] == 'LAST_DATE_VIEW' && $filter['value'] == '%Y%' ){
                $filterComments .= ' AND (COM.CREATED_AT > ALO.LAST_POST_VIEW  OR ALO.LAST_POST_VIEW IS NULL) ';
                unset($conditions[$key]);
            } else if ($filter['columnName'] == 'LAST_DATE_VIEW' && $filter['value'] == '%N%' ){
                unset($conditions[$key]);
            } else if ($filter['columnName'] == 'LABEL_CHECK' && $filter['value'] == '%Y%' ){
                $MeetingProjectIds = $this->getImInMeetingProjectIds();
                $projectIdString = \Zeedhi\DevOps\Util\Utilities::getStringForArray($MeetingProjectIds);
                $filterComments .= "AND (I.ID NOT IN (SELECT TL.TASK FROM LABEL LA JOIN TASK_LABEL TL ON TL.LABEL = LA.ID WHERE (RESOURCE_T = :userId OR PROJECT IN ( $projectIdString ))))";
                unset($conditions[$key]);
            } else if($filter['columnName'] == 'LABEL_CHECK' && $filter['value'] == '%N%' ){
                unset($conditions[$key]);
            } else if($filter['columnName'] == 'FRAMING'){
            	$value = implode(', ', $filter['value']);
        		$filterTaskStatus .= " AND (FRAMING IN ($value)) ";
                unset($conditions[$key]);
            }else if($filter['columnName'] == 'VERSION'){
                $filterProductVersion = "AND UPPER({$filter['columnName']}) LIKE '{$filter['value']}'";
            } else if ($filter['columnName'] == 'VERSION_DEV'){
                $filterProductVersion .= " AND UPPER({$filter['columnName']}) LIKE '{$filter['value']}'";
            } else if ($filter['columnName'] == 'CDITEMTRAB') {
                $cditemtrabString = DevOpsUtilities::getStringForArray($filter['value']);
                $cditemtrabString = "'" . str_replace(",", "', '", $cditemtrabString) . "'";
                $cditemtrabFilter = "AND TW.CDITEMTRAB IN ($cditemtrabString)";
            }
            $filter['columnName'] = str_replace('--','.',$filter['columnName']);
        }
        Utilities::getValueInFilterDataSourceByName($conditions, 'VERSION', true);
        Utilities::getValueInFilterDataSourceByName($conditions, 'VERSION_DEV', true);
        Utilities::getValueInFilterDataSourceByName($conditions, 'CDITEMTRAB', true);
        $conditionsSQL = $this->environment->createQueryFiltered($conditions);

        $issues = InstanceProvider::getRepository()->getTask()->getAllIssues($this->environment->getCurrentOrganizationId(),
            $this->environment->getCurrentUserId(), $conditionsSQL, $filterComments, $filterRelevance, $filterTaskStatus, $filterProductVersion, $cditemtrabFilter);
        if(count($issues) > 1000){
        	throw new \Exception(Messages::NOT_POSSIBLE_LOAD_RECORDS);
        } else {
        	return $issues;
        }
	}


    public function findAllIssuesGrid($conditions) {
        $filterComments = '';
        $filterRelevance = '';
        $filterProductVersion = '';
        $cditemtrabFilter = '';
        foreach ($conditions as $key =>&$filter){

            if ($filter['columnName'] == 'RELEVANCE'){
                $value = $filter['value'];
                if($value){
                    $filterRelevance .= " AND (I.RELEVANCE >=  $value) ";
                }
                unset($conditions[$key]);
            }

            if ($filter['columnName'] == 'LAST_DATE_VIEW' && $filter['value'] == '%Y%' ){
                $filterComments .= ' AND (COM.CREATED_AT > ALO.LAST_POST_VIEW  OR ALO.LAST_POST_VIEW IS NULL) ';
                unset($conditions[$key]);
            }else if($filter['columnName'] == 'LAST_DATE_VIEW' && $filter['value'] == '%N%' ){
                unset($conditions[$key]);
            }
            if ($filter['columnName'] == 'LABEL_CHECK' && $filter['value'] == '%Y%' ){
                $MeetingProjectIds = $this->getImInMeetingProjectIds();
                $projectIdString = \Zeedhi\DevOps\Util\Utilities::getStringForArray($MeetingProjectIds);
                $filterComments .= "AND (I.ID NOT IN (SELECT TL.TASK FROM LABEL LA JOIN TASK_LABEL TL ON TL.LABEL = LA.ID WHERE (RESOURCE_T = :userId OR PROJECT IN ( $projectIdString ))))";
                unset($conditions[$key]);
            }else if($filter['columnName'] == 'LABEL_CHECK' && $filter['value'] == '%N%' ){
                unset($conditions[$key]);
            } else if($filter['columnName'] == 'VERSION'){
                $filterProductVersion = "AND UPPER({$filter['columnName']}) LIKE '{$filter['value']}'";
            } else if ($filter['columnName'] == 'VERSION_DEV'){
                $filterProductVersion .= " AND UPPER({$filter['columnName']}) LIKE '{$filter['value']}'";
            } else if ($filter['columnName'] == 'CDITEMTRAB') {
                $cditemtrabString = DevOpsUtilities::getStringForArray($filter['value']);
                $cditemtrabString = "'" . str_replace(",", "', '", $cditemtrabString) . "'";
                $cditemtrabFilter = "AND TW.CDITEMTRAB IN ($cditemtrabString)";
            }
            $filter['columnName'] = str_replace('--','.',$filter['columnName']);
        }
        Utilities::getValueInFilterDataSourceByName($conditions, 'VERSION', true);
        Utilities::getValueInFilterDataSourceByName($conditions, 'VERSION_DEV', true);
        Utilities::getValueInFilterDataSourceByName($conditions, 'CDITEMTRAB', true);
        $conditionsSQL = $this->environment->createQueryFiltered($conditions);
        $issues = InstanceProvider::getRepository()->getTask()->getAllIssuesGrid($this->environment->getCurrentOrganizationId(),
            $this->environment->getCurrentUserId(),  $conditionsSQL, $filterComments, $filterRelevance, $filterProductVersion, $cditemtrabFilter);
        return $issues;
	}
    public function findAllIssuesFromFilter($conditions) {
        $conditionsSQL = $this->environment->createQueryFiltered($conditions);
		return InstanceProvider::getRepository()->getTask()->getAllIssuesFromFilter($this->environment->getCurrentOrganizationId(), $conditionsSQL);
	}

	public function findMyKanbanItems( $filters = array()) {
		$filterComments = '';
        $cditemtrabFilter = '';
		foreach ($filters as $key =>$filter){
			if ($filter['columnName'] == 'LAST_DATE_VIEW' && $filter['value'] == '%Y%' ){
				$filterComments = ' AND (COM.CREATED_AT > ALO.LAST_POST_VIEW  OR ALO.LAST_POST_VIEW IS NULL) ';
				unset($filters[$key]);
			}else if($filter['columnName'] == 'LAST_DATE_VIEW' && $filter['value'] == '%N%' ){
				unset($filters[$key]);
			}
            if ($filter['columnName'] == 'LABEL_CHECK' && $filter['value'] == '%Y%' ){
                $MeetingProjectIds = $this->getImInMeetingProjectIds();
                $projectIdString = \Zeedhi\DevOps\Util\Utilities::getStringForArray($MeetingProjectIds);
                $filterComments .= "AND (I.ID NOT IN (SELECT TL.TASK FROM LABEL LA JOIN TASK_LABEL TL ON TL.LABEL = LA.ID WHERE (RESOURCE_T = :userId OR PROJECT IN ( $projectIdString ))))";
                unset($filters[$key]);
            } else if($filter['columnName'] == 'LABEL_CHECK' && $filter['value'] == '%N%' ){
                unset($filters[$key]);
            } else if ($filter['columnName'] == 'CDITEMTRAB') {
                $cditemtrabString = DevOpsUtilities::getStringForArray($filter['value']);
                $cditemtrabString = "'" . str_replace(",", "', '", $cditemtrabString) . "'";
                $cditemtrabFilter = "AND TW.CDITEMTRAB IN ($cditemtrabString)";
            }
		}
        Utilities::getValueInFilterDataSourceByName($filters, 'CDITEMTRAB', true);
		$conditions = Utilities::createQueryFiltered($filters);
        return InstanceProvider::getRepository()->getTask()->findMyKanbanIssues($this->environment->getCurrentOrganizationId(), $this->environment->getCurrentUserId(), $filterComments, $conditions, $cditemtrabFilter);
	}

	public function findMyProductsIssuesUnread($conditions, $page, $pageSize) {
		$limit = $page * $pageSize;
        $offset = ($page - 1) * $pageSize;
        return InstanceProvider::getRepository()->getTask()->findMyProductsIssuesUnread($this->environment->getCurrentOrganizationId(),
            $this->environment->getCurrentUserId(), $conditions, $limit, $offset);
	}

	public function findMyIssuesUnread($filter, $page, $pageSize){
		$orgId = $this->environment->getCurrentOrganizationId();
		$userId = $this->environment->getCurrentUserId();
		$limit = $page * $pageSize;
        $offset = ($page - 1) * $pageSize;
        return InstanceProvider::getRepository()->getTask()->findMyIssuesUnread($orgId, $userId, $filter, $limit, $offset);
	}

	public function totalCommentsUnread($filter) {
		return InstanceProvider::getRepository()->getTask()->totalCommentsUnread($this->environment->getCurrentOrganizationId(), $this->environment->getCurrentUserId());
	}

	public function findProductsByManager($filter) {
		return InstanceProvider::getRepository()->getTask()->findProductsByManager($this->environment->getCurrentOrganizationId(), $this->environment->getCurrentUserId());
	}

	public function findFollowedKanbanItems($filters = array()) {
		$filterComments = '';
        $cditemtrabFilter = '';
		foreach ($filters as $key =>$filter){
			if ($filter['columnName'] == 'LAST_DATE_VIEW' && $filter['value'] == '%Y%' ){
				$filterComments = ' AND (COM.CREATED_AT > ALO.LAST_POST_VIEW  OR ALO.LAST_POST_VIEW IS NULL) ';
				unset($filters[$key]);
			}else if($filter['columnName'] == 'LAST_DATE_VIEW' && $filter['value'] == '%N%' ){
				unset($filters[$key]);
			}
            if ($filter['columnName'] == 'LABEL_CHECK' && $filter['value'] == '%Y%' ){
                $MeetingProjectIds = $this->getImInMeetingProjectIds();
                $projectIdString = \Zeedhi\DevOps\Util\Utilities::getStringForArray($MeetingProjectIds);
                $filterComments .= "AND (I.ID NOT IN (SELECT TL.TASK FROM LABEL LA JOIN TASK_LABEL TL ON TL.LABEL = LA.ID WHERE (RESOURCE_T = :userId OR PROJECT IN ( $projectIdString ))))";
                unset($filters[$key]);
            }else if($filter['columnName'] == 'LABEL_CHECK' && $filter['value'] == '%N%' ){
                unset($filters[$key]);
            } else if ($filter['columnName'] == 'CDITEMTRAB') {
                $cditemtrabString = DevOpsUtilities::getStringForArray($filter['value']);
                $cditemtrabString = "'" . str_replace(",", "', '", $cditemtrabString) . "'";
                $cditemtrabFilter = "AND TW.CDITEMTRAB IN ($cditemtrabString)";
            }
		}
        Utilities::getValueInFilterDataSourceByName($filters, 'CDITEMTRAB', true);
		$conditions = Utilities::createQueryFiltered($filters);
        $return = InstanceProvider::getRepository()->getTask()->findFollowedKanbanIssues($this->environment->getCurrentOrganizationId(), $this->environment->getCurrentUserId(),$filterComments, $conditions, $cditemtrabFilter);
        if(count($return) > 1000){
        	throw new \Exception(Messages::NOT_POSSIBLE_LOAD_ISSUES);
        } else {
        	return $return;
        }
	}

    public function findAllTasksByProject($filters = array()) {
        $project = false;
        $statusResource = false;
        $addFilter = array();
        $filterLabel = '';
        $filterIssue = '';
        $late = Utilities::getValueInFilterDataSourceByName($filters, 'LATE', true);
        foreach ($filters as $key => &$filter){
            if($filter['columnName'] == 'TASK_NAME') {
                $filter['value'] = str_replace("'", "''", $filter['value']);
            }
            if($filter['columnName'] == 'PROJECT_ID'){
                $project = false;
            }else if($filter['columnName'] == 'STRUCTURE'){
                $filter['value'] = DevOpsUtilities::getArrayFromArrayAssociative(InstanceProvider::getService()->getStructure()->getAllStructuresChildrens($filter['value']));
            }else if($filter['columnName'] == 'PERIOD'){
                $periodFilter = $filter;
                $periodFilter['columnName'] = 'PERIOD_END';
                array_push($addFilter, $periodFilter);
            }else if($filter['columnName'] == 'RESPONSIBLE'){

                $structures = DevOpsUtilities::getArrayFromArrayAssociative(InstanceProvider::getService()->getStructure()->getAllStructuresChildrensByManager($filter['value']));
                $structureFilter = array(
                    'columnName'=>'STRUCTURE',
                    'value'=>$structures,
                    'operator'=>'IN',
                );
                array_push($filters,$structureFilter);
                unSet($filters[$key]);
            }else if($filter['columnName'] == 'LABEL') {
                $values = $filter['value'];
                $filterLabel = implode(",", $values);
                unSet($filters[$key]);
            }else if($filter['columnName'] == 'ISSUE') {
                $values = $filter['value'];
                $filterIssue = implode(",", $values);
                unSet($filters[$key]);
            }else if($filter['columnName'] == 'STATUS_RESOURCE') {
                $statusResource = $filter['value'];
                unSet($filters[$key]);
            }
        }
        $filters = array_merge($filters,$addFilter);
        $filterString = DevOpsUtilities::createQueryFiltered($filters);
        if(!empty($late)){
            $lateTasks = InstanceProvider::getRepository()->getTask()->getTotalDelayedTasksByCurrentUser($this->environment->getCurrentUserId(),true);
            if(!empty($lateTasks)) {
                $tasks = implode(",", Utilities::getArrayFromAssociativeArray($lateTasks));
                $filterString .= $late == 'Y' ? " AND TASK_ID IN ($tasks)" : "AND TASK_ID NOT IN ($tasks)";
            }
        }

        $tasks = InstanceProvider::getRepository()->getTask()->getAllTasksByProjectOnDemand($this->environment->getCurrentOrganizationId(),
        $filterString, $filterLabel, $filterIssue, $project,$statusResource);
        foreach($tasks as &$task){
            if(empty($task['_hasChild'])){
                if(empty($task['TOTAL_APPROPRIATION']  )) {
                    $task['TOTAL_APPROPRIATION'] = 0;
                }
            }
        }
        return $tasks;
	}

	public function getDefaultTaskTypes() {
		return InstanceProvider::getRepository()->getTask()->getStandardTaskTypes($this->environment->getCurrentOrganizationId());
	}

    public function changeTimePlanned($rows) {
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            $taskRepository = InstanceProvider::getRepository()->getTask();
            $taskRelationshipRepository = InstanceProvider::getRepository()->getTaskRelationship();

            foreach($rows as $row){
                $task = $taskRepository->find($row['ID']);
                $currentStatus = $task->getStatus();
                $taskRelationship = $taskRelationshipRepository->findRelatedTaskIds($row['ID'], $currentStatus);
                if(!empty($taskRelationship)) {
                    $arrayTaskRelationship = explode(',',$taskRelationship['RELATED_TASKS']);
                    array_push($arrayTaskRelationship, $taskRelationship['MAIN_TASK']);
                    foreach ($arrayTaskRelationship as $taskId) {
                        $relatedTask = $taskRepository->find($taskId);
                        $relatedTask->setTimePlanned($row['TIME_PLANNED']);
                        $this->entityManager->persist($relatedTask);
                    }
                } else {
                    $task->setTimePlanned($row['TIME_PLANNED']);
                    $this->entityManager->persist($task);  
                }
                             
            }
            $this->entityManager->flush();
            $connection->commit();
        } catch(\Exception $e) {
			$connection->rollBack();
			throw $e;
		}        
	}

    public function getTaskTypes($filters = array(), $taskStatusIds = '')
    {
    	$conditions = Utilities::createQueryFiltered($filters);
        $types =  InstanceProvider::getRepository()->getTask()->getAllTypeTask($this->environment->getCurrentOrganizationId(), $conditions);

        $taskTypes = array();
        foreach($types as $type) {
            if(empty($type['ORDER_TO_VIEW'])&& $type['ORDER_TO_VIEW'] != 0){
                $column = $type['ID']+100;
            }else{
                $column = $type['ORDER_TO_VIEW'];
            }
            $taskTypes[] = array(
                "column" => $column,
                "columnLabel" => $type['NAME'],
                "ID" => $type['ID'],
                "ORDER_TO_FRAME" => $type['ORDER_TO_FRAME'],
                "CONTROL" => $type['CONTROL']
            );
        }

        return $taskTypes;
    }

	public function getDataSetForEditTask($idTask)
    {
        $taskRepository = InstanceProvider::getRepository()->getTask();
        $editTasks = $taskRepository->getDataSetForEditTask($idTask, $this->environment->getCurrentOrganizationId());
        $taskRelationshipRepository = InstanceProvider::getRepository()->getTaskRelationship();
        foreach ($editTasks as &$editTask) {
            $taskId = $editTask['ID'];
            $editTask['RELATED_TASK'] = $taskRelationshipRepository->findRelatedTaskIds($taskId);
            $meetingProjectIds = $this->getImInMeetingProjectIds();
            $editTask['TASK_LABEL'] = $taskRepository->getLabelRelatedUserByTask($taskId,$this->environment->getCurrentUserId(),$meetingProjectIds);
            $editTask['TASK_LABEL'] = DevOpsUtilities::getArrayFromArrayAssociative($editTask['TASK_LABEL']);
        }
        return $editTasks;
    }


    public function getInfoTaskInTaskTracking($idTask,$resource = null) {
        $taskRepository = InstanceProvider::getRepository()->getTask();
		$infoTask = $taskRepository->getInfoTaskInTaskTracking($idTask, $this->environment->getCurrentOrganizationId(),$resource);
        $taskRelationshipRepository = InstanceProvider::getRepository()->getTaskRelationship();
        foreach($infoTask as &$task) {
            $task['IS_RELATED_TASK'] = false;
            $relatedTasks = $taskRelationshipRepository->findRelatedTaskIds($task['ID']);
            if(!empty($relatedTasks)) {
                $mainProject = $taskRepository->find($relatedTasks['MAIN_TASK'])->getProject()->getId();
                $task['MAIN_TASK'] = $relatedTasks['MAIN_TASK'];
                $task['MAIN_PROJECT'] = $mainProject;
                $task['IS_RELATED_TASK'] = true;
            }
        }
        return $infoTask;
	}

	public function getDataSetForEditProjectTask($project, $filter = '') {
		$tasks = InstanceProvider::getRepository()->getTask()->getDataSetForEditProjectTask($project, $this->environment->getCurrentOrganizationId(), $filter);
		return $tasks;
	}

    /***********************************************************************
     * Saves a task.
     **********************************************************************
     * @param $task
     * @param $project
     * @param $latency
     * @param $name
     * @param $estimatedBeginDate
     * @param $estimatedEndDate
     * @param $status
     * @param $description
     * @param $percentCompleted
     * @param $duration
     * @param array $resource
     * @param null $precursor
     * @param null $parent
     * @param null $precursorType
     * @param null $taskStatus
     * @param string $dateExceptionType
     * @param null $dateException
     * @param null $isParent
     * @param null $urlToTest
     * @param null $classification
     * @param null $response
     * @param null $generatorTaskId
     * @param null $generatorTaskType
     * @param null $taskFeatures
     * @param null $timePlanned
     * @param null|array $labels
     * @param null $taskResponse
     * @throws \Doctrine\DBAL\ConnectionException
     * @throws \Exception
     * @return Entities\Task
     */
	public function saveProjectTask($task, $project, $latency, $name, $estimatedBeginDate, $estimatedEndDate, $status, $description, $percentCompleted,
									$duration, $resource = array(), $precursor = null, $parent = null, $precursorType = null, $taskStatus = null,
									$dateExceptionType = 'MB', $dateException = null, $isParent = null, $urlToTest = null, $classification = null,
                                    $response = null, $generatorTaskId = null, $generatorTaskType = null, $taskFeatures = null, $timePlanned = null, $labels = null, $taskResponse = null, $versionDevId = null)
    {

        $connection = $this->entityManager->getConnection();
        $taskReposiry = InstanceProvider::getRepository()->getTask();
        try {
            $connection->beginTransaction();
            $name = substr($name, 0, 200);
            /** @var Entities\Project $project */
            $project = InstanceProvider::getRepository()->getProject()->find($project);
            /** @var \Service\Project $projectService */
            $projectService = InstanceProvider::getService()->getProject();

            if ($classification) {
                /** @var Entities\Classification $classification */
                $classification = InstanceProvider::getRepository()->getClassification()->find($classification);
            }
            $timeScore = $timePlanned;
            if($taskFeatures){
                /** @var Entities\TaskFeatures $taskFeatures */
                $taskFeatures = InstanceProvider::getRepository()->getTaskFeatures()->find($taskFeatures);
                $timeScore = $taskFeatures->getTimeBase();
            }

            if ($taskStatus) {
                $taskStatus = InstanceProvider::getRepository()->getTaskStatus()->find($taskStatus);
            }
            if (!empty($task)) {
                $isNew = false;
                $task = $this->saveEditTaskProject($task, $project, 0, $name, $estimatedBeginDate, $estimatedEndDate, $status, $description,
                    $percentCompleted, $duration, $resource, $precursor, $parent, $precursorType, $taskStatus,
                    $dateExceptionType, $dateException, $isParent, $urlToTest, $classification, $response, $taskFeatures, $versionDevId);
            } else {
                $isNew = true;
                $task = $this->createNewTaskProject($project, 0, $name, $estimatedBeginDate, $estimatedEndDate, $status, $description,
                    $percentCompleted, $duration, $resource, $precursor, $parent, $precursorType, $taskStatus,
                    $dateExceptionType, $dateException, $isParent, $urlToTest, $classification, $taskFeatures, null, $versionDevId);
            }
            $task->setResponse($taskResponse);
            if (!$task->isParent()) {
            	if(!$timeScore)
            	{
            		$timeScore = $task->getDuration();
            	}
                $task->setTimeScore($timeScore);
                $task->setTimePlanned($timePlanned);
            }
            if ($latency) {
                $task->setLatency(InstanceProvider::getService()->getCalendarTakCalendar()->getDurationInMilliseconds($latency, $task));
            } else {
                $task->setLatency(0);
            }

            if(is_array($labels)){
                $this->associateTaskAndLabels($task, $labels, null, true);
            }
            if($task->getTaskStatus() && $task->getTaskStatus()->getControl() == Entities\TaskStatus::TASK_STATUS_CONTROL_CONCLUDED
                && !$task->getEndDate()){
                $task->setEndDate($this->environment->getCurrentTime());
            }
            $this->replicationRelationshipRelated($task,$resource);

            $this->entityManager->persist($task);
            $this->entityManager->flush();
            if($generatorTaskId) {
                /** @var Entities\Task $generatorTask */
                $generatorTask = InstanceProvider::getRepository()->getTask()->find($generatorTaskId);
                $relationship = InstanceProvider::getFactory()->getTaskRelationship()->createTaskRelationship($generatorTask, $task, $generatorTaskType);
                $this->entityManager->persist($relationship);
            }

            $this->applyOrderToIssuesByTask($task);
            $this->setTaskTimePlannedIfNull($task, $resource,$timePlanned);
            $this->setTaskPrecursorInTaskParent($task);

            /** @var Entities\Project $baselineProject */
            $baselineProject = $projectService->getBaselineProjectRelatedOrNull($project->getId());

            if($baselineProject && $isNew && (($task->getClassification() && !$task->getClassification()->getControl()) || !$task->getClassification())){
                $baselineTask = $this->createBaselineTaskByTaskCreated($task, $baselineProject, $status);
                $this->createBaselineTypeTaskRelationship($task, $baselineTask);
            }

            if($baselineProject){
                $baselineTask = $taskReposiry->findBaselineTaskOfTaskCurrentProject($task->getId(), $baselineProject->getId());
                if(!$isNew && !empty($baselineTask)) {
                    if($status == Entities\Task::NOT_ACTIVE){
                        $this->setBaselineOccurrenceInBaselineTask($baselineTask['ID'], Entities\Task::BASELINE_OCCURRENCE_INACTIVATED);
                    } else {
                        $this->setBaselineOccurrenceInBaselineTask($baselineTask['ID'], null);
                    }
                }
            }

            $this->entityManager->flush();
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
        return $task;

    }

    /**
     * @param Entities\Task $task
     * @param $baselineProject
     * @return Entities\Task
     * @throws \Doctrine\ORM\OptimisticLockException
     */
    public function createBaselineTaskByTaskCreated(Entities\Task $task, Entities\Project $baselineProject, $status = null)
    {
        $taskRepository = InstanceProvider::getRepository()->getTask();
        $baselineTask = clone $task;
        if($baselineTask->getParent()){
            $baselineTaskParent = $taskRepository->findBaselineTaskOfTaskCurrentProject($baselineTask->getParent()->getId(), $baselineProject->getId());
            if(!empty($baselineTaskParent)){
                /** @var Entities\task $baselineTaskParent */
                $baselineTaskParent = $taskRepository->find($baselineTaskParent['ID']);
                $baselineTask->setParent($baselineTaskParent);
            }
        }
        $baselineTask->setProject($baselineProject);
        $baselineTask->setIsOriginalBaseline(Entities\Task::IS_NOT_ORIGINAL_BASELINE);

        if($status == Entities\Task::NOT_ACTIVE){
            $baselineTask->setBaselineOccurrence(Entities\Task::BASELINE_OCCURRENCE_INACTIVATED);
        }

        $this->entityManager->persist($baselineTask);
        $this->entityManager->flush();
        return $baselineTask;
    }

    /**
     * @param Entities\Task $task
     * @param Entities\Task $baselineTask
     * @return Entities\TaskRelationship
     * @throws \Doctrine\ORM\OptimisticLockException
     */
    public function createBaselineTypeTaskRelationship(Entities\Task $task, Entities\Task $baselineTask)
    {
        $taskRelationship = new Entities\TaskRelationship();
        $taskRelationship->setTask($task);
        $taskRelationship->setTaskRelationship($baselineTask);
        $taskRelationship->setType(Entities\TaskRelationship::BASELINE_TYPE);

        $this->entityManager->persist($taskRelationship);
        $this->entityManager->flush();
        return $taskRelationship;
    }

    /**
     * @param  array $taskIds
     * @throws \Doctrine\ORM\OptimisticLockException
     */
    public function createBaselineTasksByTaskIds($taskIds)
    {
        foreach ($taskIds as $id){

            $projectRepository = InstanceProvider::getRepository()->getProject();

            /** @var Entities\Task $task */
            $task = InstanceProvider::getRepository()->getTask()->find($id);
            /** @var Entities\Project $project */
            $project = $projectRepository->find($task->getProject());
            $taskRelationshipFilter = array('project' => $project->getId(), 'type' => Entities\ProjectRelationship::TYPE_TIME_LINE);
            $projectRelationship = InstanceProvider::getRepository()->getProjectRelationship()->findOneBy($taskRelationshipFilter);

            if(!empty($projectRelationship) && ($task->getClassification() && !$task->getClassification()->getControl())){
                /** @var Entities\Project $baselineProject */
                $baselineProject = $projectRepository->find($projectRelationship->getProjectRelationship());

                $baselineTask = $this->createBaselineTaskByTaskCreated($task, $baselineProject);
                $this->createBaselineTypeTaskRelationship($task, $baselineTask);
            }
        }
    }

    /**
     * @param $taskIdsForBaselineRegister
     * @param $originProjectId
     * @throws \Exception
     */
    public function setBaselineTaskMoved($taskIdsForBaselineRegister, $originProjectId)
    {
        $taskRepository = InstanceProvider::getRepository()->getTask();

        foreach ($taskIdsForBaselineRegister as $taskId){
            $baselineTaskOriginProject = $taskRepository->getBaselineTaskFromOriginBaselineProject($taskId, $originProjectId);
            if(!empty($baselineTaskOriginProject)){
                /** @var Entities\Task $baselineTaskOriginProject */
                $baselineTaskOriginProject = $taskRepository->find($baselineTaskOriginProject[0]['ID']);

                $baselineTaskOriginProject->setBaselineOccurrence(Entities\Task::BASELINE_OCCURRENCE_MOVED);
                $baselineTaskOriginProject->setBaselineOccurrenceAt(new \DateTime());
                $this->entityManager->persist($baselineTaskOriginProject);
            }
        }
    }

    /**
     * @param $baselineTaskId
     * @param $baselineOccurrence
     * @throws \Exception
     */
    public function setBaselineOccurrenceInBaselineTask($baselineTaskId, $baselineOccurrence)
    {
        /** @var Entities\Task $baselineTask */
        $baselineTask = InstanceProvider::getRepository()->getTask()->find($baselineTaskId);
        $baselineTask->setBaselineOccurrence($baselineOccurrence);
        $baselineTask->setBaselineOccurrenceAt(new \DateTime());
        $this->entityManager->persist($baselineTask);
    }

    public function setTaskPrecursorInTaskParent(Entities\Task $task)
    {
        $taskRepository = InstanceProvider::getRepository()->getTask();
        $taskParent = null;
        if ($task->isContainer()) {
            $taskParent = $task;
        } else if ($task->getParent()) {
            $taskParent = $task->getParent();
        }
        if ($taskParent) {
            $fistTaskId = $taskRepository->getFistTaskIdFromParentTask($taskParent->getId());
            if($fistTaskId){
                $fistTask = $taskRepository->find($fistTaskId);
                /** @var Entities\TaskRelationship $precursorTaskRelationship */
                $precursorTaskRelationship = InstanceProvider::getRepository()->getTaskRelationship()
                    ->findOneBy(array('task' => $fistTask, 'typePrecursor' => Entities\TaskRelationship::END_START_TYPE,
                        'type' => Entities\TaskRelationship::PRECURSOR_TYPE));
                if ($precursorTaskRelationship) {
                    $precursorTask = $precursorTaskRelationship->getTaskRelationship();
                    if (($precursorTask->getParent() && $precursorTask->getParent()->getId() != $taskParent->getId())
                        || is_null($precursorTask->getParent())) {

                        $this->setPrecursorTask($taskParent, $precursorTask, Entities\TaskRelationship::END_START_TYPE);

                    }else{
                        $this->removePrecursor($taskParent);
                    }
                }else{
                    $this->removePrecursor($taskParent);
                }
            }
        }
    }

    public function removePrecursor(Entities\Task $task){
        $precursorRelationshipTask = InstanceProvider::getRepository()->getTaskRelationship()->findOneBy(array("task" => $task->getId(),
            'type' => TaskPrecursor::PRECURSOR_TYPE));
        if($precursorRelationshipTask){
            $this->entityManager->remove($precursorRelationshipTask);
        }
    }

    public function replicationRelationshipRelated(Entities\Task $task,$resourceIds){
        $relatedTasks = InstanceProvider::getRepository()->getTaskRelationship()->findRelatedTaskIds($task->getId());
        if (!empty($relatedTasks)) {
            $taskIds = DevOpsUtilities::setSelectMultiInQuery($relatedTasks['RELATED_TASKS']);
            $taskIds[] = $relatedTasks['MAIN_TASK'];
            $taskStatus = $task->getTaskStatus();

            $taskRepository = InstanceProvider::getRepository()->getTask();
            foreach ($taskIds as $taskId) {
                if($taskId != $task->getId()){
                    /** @var Entities\Task $relatedTask */
                    $relatedTask = $taskRepository->find($taskId);
                    $this->allocateResourcesInProject($resourceIds,$relatedTask->getProject());
                    $this->setAllocationsInTask($relatedTask, $resourceIds, $taskStatus);
                    $relatedTask->setName($task->getName());
                    $relatedTask->setTaskStatus($taskStatus);
                    $relatedTask->setTimePlanned($task->getTimePlanned());
                    $relatedTask->setInCalc('Y');
                    if($relatedTask->getTaskStatus()->getControl() == Entities\TaskStatus::TASK_STATUS_CONTROL_CONCLUDED
                        && $task->getEndDate()
                        && !$relatedTask->getEndDate()){
                        $relatedTask->setEndDate($task->getEndDate());
                    }
                }
            }

        }
    }

    protected function allocateResourcesInProject($resourcesIds, Entities\Project $targetProject)
    {
        $resourceProjectFactory = InstanceProvider::getFactory()->getResourceProject();
        $resourceRepository = InstanceProvider::getRepository()->getResource();
        $resourceProjectRepository = InstanceProvider::getRepository()->getResourceProject();
        foreach ($resourcesIds as $resourceId) {
            $resourceProject = $resourceProjectRepository->findOneBy(array('resourceC' => $resourceId, 'project' => $targetProject));
            if(empty($resourceProject)){
                $resource = $resourceRepository->find($resourceId);
                $resourceProjectEntity = $resourceProjectFactory->createResourceProject($resource, $targetProject);
                $this->entityManager->persist($resourceProjectEntity);
            }
        }
        $this->entityManager->flush();

    }

    public function validateParentProjectTask($projectId, $parentId)
    {
        /** @var Entities\Task $parent */
        $parent = InstanceProvider::getRepository()->getTask()->find($parentId);
        if(!empty($parent)){
            return $projectId == $parent->getProject()->getId();
        }
        return false;
    }

    /***********************************************************************
     * Saves a task that already exists.
     **********************************************************************
     * @param $task
     * @param Entities\Project $project
     * @param $latency
     * @param $name
     * @param \DateTime $estimatedBeginDate
     * @param \DateTime $estimatedEndDate
     * @param $status
     * @param $description
     * @param $percentCompleted
     * @param $duration
     * @param $resource
     * @param $precursor
     * @param null $parent
     * @param $precursorType
     * @param $taskStatus
     * @param $dateExceptionType
     * @param $dateException
     * @param $isParent
     * @param null $urlToTest
     * @param null $classification
     * @param null $response
     * @throws \Exception
     * @return Entities\Task
     */
	public function saveEditTaskProject($task, \Zeedhi\DevOps\Model\Projectmanagement\Entities\Project $project, $latency,
                                        $name, \DateTime $estimatedBeginDate,\DateTime $estimatedEndDate, $status,
                                        $description, $percentCompleted, $duration, $resource, $precursor, $parent = null,
                                        $precursorType, $taskStatus,$dateExceptionType, $dateException, $isParent,
                                        $urlToTest = null, $classification = null, $response = null, $taskFeatures = null, $versionDevId = null){

		try{
            $taskRepository = InstanceProvider::getRepository()->getTask();
            /** @var Entities\Task $task */
			$task = $taskRepository->find($task);
			if (!empty($parent)) {
				if($task->getId() != $parent){
					$parent = $taskRepository->find($parent);
				}
			}
            $this->setAllocationsInTask($task, $resource, $taskStatus);
            $this->setPrecursorTask($task, $precursor, $precursorType);
			if($dateExceptionType == RestrictionFactory::DO_NOT_START_BEFORE && empty($dateException)){
				$dateException = clone $estimatedBeginDate;
			}
			$task = $this->updateTask($task, $project, $latency, $name, $estimatedBeginDate, $estimatedEndDate,
				$status, $description, $percentCompleted, $duration, $parent, $taskStatus,$dateExceptionType,
                $dateException, $isParent, $urlToTest, $classification, $response, $taskFeatures, $versionDevId);
            $circularRecursion = $task->hasCircularRecursion();
            if($circularRecursion){
                throw new \Exception($circularRecursion);
            }
			InstanceProvider::getService()->getCalendarTakCalendar()->setPreviousTimeTask($task);
			$this->setStatusAllSubTask($task,$status);
			if($percentCompleted == 100){
				$this->setAllSubTaskPercent100($task);
			}
            $this->entityManager->persist($task);
            $this->entityManager->flush();
		}catch (\Exception $e ){
			throw $e;
		}
		return $task;
	}

	public function updateStatusRelatedTasks($taskId, $taskStatus, $currentStatus)
	{
	    $taskRelationshipRepository = InstanceProvider::getRepository()->getTaskRelationship();
	    $taskRepository = InstanceProvider::getRepository()->getTask();
	    $taskRelationship = $taskRelationshipRepository->findRelatedTaskIds($taskId, $currentStatus);
	    if($taskRelationship){
	        $arrayTaskRelationship = explode(',',$taskRelationship['RELATED_TASKS']);
    	    array_push($arrayTaskRelationship, $taskRelationship['MAIN_TASK']);
    	    foreach ($arrayTaskRelationship as $taskId) {
    	        $task = $taskRepository->find($taskId);
    	        $task->setStatus($taskStatus);
    	        $this->entityManager->persist($task);
    	    }
    	    $this->entityManager->flush();
	    }

	}

	public function deleteAllResourcesAllocated ($tasks) {
		$connection = $this->entityManager->getConnection();
		$connection->beginTransaction();
		try{
            $allocationRepository = InstanceProvider::getRepository()->getAllocation();
			foreach ($tasks as $task) {
				$allocations = $allocationRepository->findBy(array('task' => $task['ID']));
				foreach($allocations as $allocation){
					$this->entityManager->remove($allocation);
				}
			}
			$this->entityManager->flush();
			$connection->commit();
		}
		catch(\Exception $e){
			$connection->rollBack();
			throw $e;
		}
	}

	public function createNewTaskProject(\Zeedhi\DevOps\Model\Projectmanagement\Entities\Project $project, $latency, $name, \DateTime $estimatedBeginDate,
										 \DateTime $estimatedEndDate, $status, $description, $percentCompleted, $duration,
										 $resource, $precursor, $parent = null, $precursorType, $taskStatus, $dateExceptionType, $dateException, $isParent,
                                         $urlToTest = null, $classification = null, $taskFeatures = null, $errorClassification = null, $versionDevId = null) {
        if($project->getProjectStatus()
            && $project->getProjectStatus()->getControl() == Entities\TaskStatus::TASK_STATUS_CONTROL_IN_HOMOLOGATION) {
            if (!$this->validationToTaskInHomologation($project, $classification)) {
                throw new \Exception(str_replace('%1',$project->getProjectStatus()->getLabel(),Messages::ERR_CLASSIFICATION_IN_HOMOLOGATION), 2);
            }
        }
		if ($parent) {
			$parent = InstanceProvider::getRepository()->getTask()->find($parent);
		}

		$task = $this->createTask(
        		    $project, $latency, $name, $estimatedBeginDate, $estimatedEndDate, $status,
        			$description, $percentCompleted, $duration, $parent,$taskStatus, $dateExceptionType, $dateException,
        			$isParent, $urlToTest, $classification, $taskFeatures, $errorClassification, null, $versionDevId
    			);

		if ($resource) {
            $resourceRepository = InstanceProvider::getRepository()->getResource();
			foreach ($resource as $eachResource) {
				$resourceProject = $resourceRepository->find($eachResource);
				InstanceProvider::getService()->getAllocation()->createAllocation($task, $resourceProject,null,$taskStatus);
			}
		}

		if ($precursor) {
			$taskPrecursor = InstanceProvider::getRepository()->getTask()->findOneBy(array('id' => $precursor));
			if($taskPrecursor){
			    InstanceProvider::getService()->getTaskPrecursor()->createTaskPrecursor($task, $taskPrecursor, TaskPrecursor::PRECURSOR_TYPE, $precursorType);
			}
		}
        $calendarTakCalendar = InstanceProvider::getService()->getCalendarTakCalendar();
		$task->setDuration($calendarTakCalendar->getDurationInMilliseconds($duration,$task));
		$calendarTakCalendar->setPreviousTimeTask($task);
		$this->entityManager->persist($task);
		$this->entityManager->flush();

		return $task;
	}


	protected function setStatusAllSubTask(\Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $task , $status){

		/** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Task[] $allSubTask */
		$allSubTask = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Task')->findBy(array("parent" => $task));
		if (!empty($allSubTask)) {
			foreach ($allSubTask as $subTask) {
				$subTask->setStatus($status);
				$this->setStatusAllSubTask($subTask,$status);
			}
		}
	}

	protected function setAllSubTaskPercent100(\Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $task){
		/** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Task[] $allSubTasks */
		$allSubTasks = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Task')->findBy(array("parent" => $task));
		foreach($allSubTasks as $subTask){
			$this->setAllSubTaskPercent100($subTask);
			$subTask->setPercentCompleted(100);
			$this->entityManager->persist($subTask);
		}
		$this->entityManager->flush();
	}

	public function getTasksByUserPerPeriod($resourceId, $beginDate, $endDate){

        $taskRepository = InstanceProvider::getRepository()->getTask();
		if(!empty($beginDate) && !empty($endDate)){
			return $taskRepository->getNotCompletedPerPeriod($resourceId, $beginDate, $endDate);
		} else {
			return $taskRepository->getNotCompletedTasksByUser($resourceId);
		}
	}
    public function setAllocationTasks($tasks){

    }

    public function saveTasksAllocation($tasksIds, $resourcesIds, $allocationType){
        $qntityAllocatedResources = 0;

        $taskRepository = InstanceProvider::getRepository()->getTask();
        foreach($tasksIds as $id){
            $allTasksEntities = array();
            /** @var Entities\Task $taskEntity */
            $taskEntity = $taskRepository->find($id);
            $taskChildrenEntities = $taskRepository->getAllChildren($id);
            $allResourceAllocateds = array();
            $resourceAllocateds = $taskEntity->getAllocations();
            foreach ($resourceAllocateds as $resourceAllocated) {
                  array_push($allResourceAllocateds,$resourceAllocated->getResource()->getId());
            }

            array_push($allTasksEntities, $taskEntity);
            $allTasksEntities = array_merge($allTasksEntities, $taskChildrenEntities);

            $connection = $this->entityManager->getConnection();
            $connection->beginTransaction();

            try {
                $allocationService = InstanceProvider::getService()->getAllocation();
                $allocationRepository = InstanceProvider::getRepository()->getAllocation();
                $taskRelationshipRepository = InstanceProvider::getRepository()->getTaskRelationship();
                // Navigates through each task
                foreach ($allTasksEntities as $taskEntity) {
                    $taskMain = $taskRelationshipRepository->findRelatedTaskIds($taskEntity->getId());
                    if (!empty($taskMain)) {
                        throw new\Exception('A tarefa ' . $taskEntity->getName() . Messages::ERR_RELATED_IN_SET_ALLOCATION);

                    }
                    if($allocationType == self::REPLACE_ALLOCATION) {
                        $allocationsEntities = $allocationRepository->findBy(array('task' => $taskEntity->getId()));
                        foreach ($allocationsEntities as $allocationEntity) {
                            $this->entityManager->remove($allocationEntity);
                        }
                    }

                    $resourceRepository = InstanceProvider::getRepository()->getResource();
                    foreach ($resourcesIds as $resourceId) {
                        if (!in_array($resourceId, $allResourceAllocateds) || $allocationType == self::REPLACE_ALLOCATION) {
                            $resourceEntity = $resourceRepository->find($resourceId);
                            $allocationService->createAllocation($taskEntity, $resourceEntity, null, null, true);
                            $qntityAllocatedResources++;
                        }
                    }
                }

                $this->entityManager->flush();
                $connection->commit();

            } catch(\Exception $e){
                $connection->rollBack();
                throw $e;
            }
        }

        return $qntityAllocatedResources;
    }

	public function deleteAllocatedResources ($taskIds) {
		$connection = $this->entityManager->getConnection();
		$connection->beginTransaction();
		try{
            $allocationRepository = InstanceProvider::getRepository()->getAllocation();
			foreach ($taskIds as $task) {
				$allocations =  $allocationRepository->findBy(array('task' => $task));
				foreach($allocations as $allocation){
					$this->entityManager->remove($allocation);
				}
			}
			$this->entityManager->flush();
			$connection->commit();
		}
		catch(\Exception $e){
			$connection->rollBack();
			throw $e;
		}
	}
    public static function validateRecursiveInPersistManyAllocations($allocationType){
        return $allocationType == self::ADD_ALLOCATION;
    }
    public function persistManyAllocations($resourceIds, $taskIds,$allocationType = false)
    {   $count = 0;
        $recursive = self::validateRecursiveInPersistManyAllocations($allocationType);
        foreach($resourceIds as $resourceId){
            $this->persistAllocation($resourceId, $taskIds,$recursive);
            $count++;
        }
        return $count;
    }

	public function persistAllocation($resourceId, $tasks, $recursive= false)
    {
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            $allocationService = InstanceProvider::getService()->getAllocation();
            $taskRepository = InstanceProvider::getRepository()->getTask();
            $resourceRepository = InstanceProvider::getRepository()->getResource();
            $allocationRepository = InstanceProvider::getRepository()->getAllocation();
            foreach ($tasks as $taskId) {
                $allocation = $allocationRepository->findOneBy(array('resource' => $resourceId, 'task' => $taskId));
                if (empty($allocation)) {
                    $resource = $resourceRepository->find($resourceId);
                    if($taskId instanceof Entities\Task ){
                        /** @var Entities\Task $task */
                        $task = $taskId;
                    }else{
                        $task = $taskRepository->find($taskId);
                    }
                    $allocationService->createAllocation($task, $resource, null, $task->getTaskStatus());
                    if($recursive){
                        $childrenTasks = $taskRepository->findBy(array('parent' => $taskId));
                        $this->persistAllocation($resourceId, $childrenTasks, $recursive);
                    }
                }
            }
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    public function getLabelsForEditAndNotEditInManagerIssue(array $labelIds,$board = null, $accompanyingIssue = false, $editable = array(),
                                                             $notEditable = array())
    {
        $meetingProjectIds = $this->getImInMeetingProjectIds();
        if($board && !$accompanyingIssue){
            $meetingProjectIds[] = $board;
        }
        $userId = $this->environment->getCurrentUserId();
        foreach ($labelIds as $labelId) {
            /** @var Entities\Label $label */
            $label = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Label')->find($labelId);
            $relatedResource = $label->getResource();
            if ($relatedResource) {
                if ($relatedResource->getId() == $userId) {
                    $editable[] = $labelId;
                } else {
                    $notEditable[] = $labelId;
                }
            } else if ($label->getProject() && in_array($label->getProject()->getId(), $meetingProjectIds)) {
                $editable[] = $labelId;
            } else {
                $notEditable[] = $labelId;
            }
        }
        return array($editable,$notEditable);
    }

    public function redoBudget($issueId,$description){
        return InstanceProvider::getService()->getBudget()->redoBudget($issueId,$description);
    }

    public function allowRedoBudget($issueId){
        return InstanceProvider::getService()->getBudget()->allowRedoBudget($issueId);
    }

    /**
     * Used in accompanyingIssue and boardIssue
     * @param $issueId
     * @param bool $accompanyingIssue
     * @return array
     */
	public function findIssue($issueId,$accompanyingIssue = false){
		$issues = InstanceProvider::getRepository()->getTask()->getIssue($issueId, $this->environment->getCurrentUserId());
        foreach($issues as &$issue) {
            $issue['FOLLOWERS'] = Utilities::setSelectMultiInQuery($issue['FOLLOWERS'],true);

            $issue['LABELS'] =  Utilities::setSelectMultiInQuery($issue['LABELS']);

            $issue['CLIENT_ACCEPT'] = Entities\TaskStatus::TASK_STATUS_CONTROL_CLIENT_ACCEPT == $issue['CONTROL'];

            $issue['TASK_CLASSIFICATION_IDS'] =  Utilities::setSelectMultiInQuery($issue['TASK_CLASSIFICATION_IDS']);

            $issue['ACTIONS_OF_EXITS'] = $this->getActionsOfExits($issueId);

            list($issue['LABELS'], $issue['LABELS_NOT_EDITABLE']) =
                $this->getLabelsForEditAndNotEditInManagerIssue($issue['LABELS'], $issue['BOARD'],$accompanyingIssue);

            $issue['ACTIONS_REDO_BUDGET'] = $this->allowRedoBudget($issueId);

            $issue['HAS_SPECIFICATION_COMMENT'] = !!$issue['HAS_SPECIFICATION_COMMENT'];
        }
		return $issues;
	}

    public function getActionsOfExits($issueId){
        $actionsOfExits = InstanceProvider::getRepository()->getTask()->getActionsOfExits($issueId);
        foreach ($actionsOfExits as &$actionsOfExit) {
            $actionsOfExit['COMMENT_DESCRIPTION'] = $actionsOfExit['COMMENT_DESCRIPTION']?$actionsOfExit['COMMENT_DESCRIPTION']:'';
        }

        return $actionsOfExits;
    }

	public function convertStringToArraySeparatedByComma($text){
		if($text) return explode(', ', $text);
		return array();
	}

    /**
     * @param $labels
     * @param $labelsNotEditable
     * @param $issue
     */
    protected function setLabelsIssue($labels, $labelsNotEditable, $issue)
    {
        $labels = array_unique(array_merge($labelsNotEditable, $labels));
        $this->setLabels($labels, $issue);
    }


    public function validationToTaskInHomologation(Entities\Project $project,Entities\Classification $classification = null)
    {
        return $project->getProjectStatus()
        && $project->getProjectStatus()->getControl() == Entities\TaskStatus::TASK_STATUS_CONTROL_IN_HOMOLOGATION
        && $classification
        && (in_array($classification->getControl(), Classification::getArrayControlForHomologation()));
    }


    protected function setLabelInSchedule(Entities\Task $scheduleInMeeting,array $labelsSchedule = array())
    {
        if ($scheduleInMeeting->getParent()) {
            $taskLabelFactory = InstanceProvider::getFactory()->getTaskLabel();
            $labelRepository = InstanceProvider::getRepository()->getLabel();
            $taskLabelRepository = InstanceProvider::getRepository()->getTaskLabel();
            foreach ($labelsSchedule as $labelScheduleId) {
                if (!empty($labelScheduleId)) {
                    /** @var Entities\Label $labelSchedule */
                    $labelSchedule = $labelRepository->find($labelScheduleId);
                    /** @var Entities\TaskLabel $taskLabel */
                    $taskLabel = $taskLabelRepository->findOneBy(array("task" => $scheduleInMeeting,
                        "label" => $labelSchedule));
                    if (empty($taskLabel)) {
                        $taskLabel = $taskLabelFactory->createTaskLabel($scheduleInMeeting, $labelSchedule);
                        $this->entityManager->persist($taskLabel);
                    }
                }
            }
            /** @var Entities\TaskLabel[] $taskLabels */
            $taskLabels = $taskLabelRepository->findBy(array("task" => $scheduleInMeeting));
            foreach ($taskLabels as $taskLabel) {
                if (!in_array($taskLabel->getLabel()->getId(), $labelsSchedule)) {
                    $this->entityManager->remove($taskLabel);
                }
            }
            $this->entityManager->flush();
        }
    }


    public function saveTaskTypeClassification($classificationIds, $taskType, $type,$deleteNotIn = true)
    {
        $taskTypeClassifications = array();
        $taskTypeClassificationService = InstanceProvider::getService()->getTaskTypeClassification();
        $classificationRepository = InstanceProvider::getRepository()->getClassification();
        foreach ($classificationIds as $classificationId) {
            $classification = $classificationRepository->findOneBy(array('id' => $classificationId));
            $taskTypeClassifications[] = $taskTypeClassificationService->createTaskTypeClassification($classification,
                $taskType, $type);
            $this->entityManager->flush();
        }
        if($deleteNotIn){
            /** @var Entities\TaskTypeClassification[] $allTaskTypeClassifications */
            $allTaskTypeClassifications = InstanceProvider::getRepository()->getTaskTypeClassification()->findBy(
                array( 'taskType'=>$taskType, 'taskIssueClassif' => $type));
            foreach ($allTaskTypeClassifications as $taskTypeClassification) {
                if (!in_array($taskTypeClassification->getClassification()->getId(), $classificationIds)) {
                    $this->entityManager->remove($taskTypeClassification);
                    $this->entityManager->flush();
                }
            }
        }
        return  $taskTypeClassifications;
    }

    public function saveTaskTypeProjectType($projectTypeIds, $taskType)
    {
        $projectTypeService = InstanceProvider::getService()->getProjectType();
        $projectTypeRepository= InstanceProvider::getRepository()->getProjectType();
        $taskTypeProjectTypes = array();
        foreach ($projectTypeIds as $projectTypeId) {
            $projectType = $projectTypeRepository->find($projectTypeId);
            $taskTypeProjectTypes[] = $projectTypeService->createTaskTypeProjectType($projectType,$taskType);
           $this->entityManager->flush();
        }
        return  $taskTypeProjectTypes;
   }

    private function divideLabelsByRelatedAndNotRelated($labelsIds, &$related, &$notRelated,$accompanyingIssue = false){
        if($accompanyingIssue){
            $meetingProjectIds = $this->getImInMeetingProjectIds();
        }
        foreach($labelsIds as $labelId){
            /** @var Entities\Label $labelEntity */
            $labelEntity = $this->entityManager->getRepository(Utilities::ENTITY_DIR.'Label')->find($labelId);
            $relatedResourceEntity = $labelEntity->getResource();
            if($relatedResourceEntity){
                if($relatedResourceEntity->getId() == $this->environment->getCurrentUserId()){
                    array_push($related, $labelId);
                } else {
                    array_push($notRelated, $labelId);
                }
            } else if(!empty($meetingProjectIds)){
                if($labelEntity->getProject() && in_array($labelEntity->getProject()->getId() ,$meetingProjectIds)){
                    array_push($related, $labelId);
                }
            }else{
                array_push($notRelated, $labelId);
            }
        }
    }

	public function saveNewTaskClassification($label){
		$connection = $this->entityManager->getConnection();
		try{
			$connection->beginTransaction();
			$classification = InstanceProvider::getFactory()->getClassification()->createClassification($label, null, self::IS_TASK);
			$this->entityManager->persist($classification);
			$this->entityManager->flush();
			$connection->commit();
		}
		catch(\Exception $e){
			$connection->rollBack();
			throw $e;
		}
	}

	public function getTaskRelationship($taskId, $classificationId = null, $filter = ' '){
		return InstanceProvider::getRepository()->getTask()->getRelationedTasks($taskId, $classificationId, $filter);
	}
	public function getTaskRelationshipByIssue($taskId, $filter ){
		return InstanceProvider::getRepository()->getTask()->getTaskRelationshipByIssue($taskId,  $filter);
	}

	public function closeIssue($issueIds){
        $user = $this->environment->getCurrentUser();
		/**@var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskProperties $taskProperties*/
		$taskProperties = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskProperties')->findOneBy(array('organization' =>$this->environment->getCurrentOrganizationId()));
		if($taskProperties){
            $taskRepository = InstanceProvider::getRepository()->getTask();
            $allocationRepository = InstanceProvider::getRepository()->getAllocation();
			foreach($issueIds AS $issueId){
				/** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $issue*/
				$issue = $taskRepository->find($issueId);
                /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Allocation $reporter */
                $reporter = $allocationRepository->findOneBy(array('type'=>\Service\Allocation::REPORTER,'resource' => $user,'task'=>$issue));
                if(empty($reporter)){
                    throw new \Exception('Essa issue só pode ser finalizada pelo solicitante!');
                }
				$issue->setPercentCompleted(100);
				$issue->setTaskStatus($taskProperties->getFinalTaskStatus());
				$issue->setCurrentTaskStatusIn((new \DateTime()));
				$this->entityManager->persist($issue);

                $closeComments = InstanceProvider::getFactory()->getComments()->createComments($issue,self::CLOSE_COMMENTS);
                $this->entityManager->persist($closeComments);
                $this->viewCommentsSaveData($issue->getId());
              
                
            }
            $this->entityManager->flush();
        }else{
			throw new \Exception('Padrão de tarefa não cadastrada!');
		}
		return $issue;
	}

    public function deleteTaskRelationshipByType($task, $relatedTask, $type) {
        $taskRelationshipArray = InstanceProvider::getRepository()->getTaskRelationship()->findBy(
            array('task' => $task, 'taskRelationship' => $relatedTask, 'type' => $type));
        foreach($taskRelationshipArray as $taskRelationship){
            $this->entityManager->remove($taskRelationship);
        }
        $this->entityManager->flush();
    }


	/**
	 * @param $task
	 * @param null $relatedTask
	 * @throws \Exception
	 */
	public function deleteTaskRelationship($task, $relatedTask = null) {
		$connection = $this->entityManager->getConnection();
		$connection->beginTransaction();
		try{
            $taskRelationshipRepository = InstanceProvider::getRepository()->getTaskRelationship();
			if(empty($relatedTask)){
				$taskRelationshipArray = array_merge(
					$taskRelationshipRepository->findBy(array('task' => $task)),
					$taskRelationshipRepository->findBy(array('taskRelationship' => $task))
				);
				/** @var Entities\TaskRelationship[] $taskRelationship */
				foreach($taskRelationshipArray as $taskRelationship){
					$this->entityManager->remove($taskRelationship);
				}
			}
			else{
				$taskRelationshipArray = array_merge(
					$taskRelationshipRepository->findBy(array('task' => $task, 'taskRelationship' => $relatedTask)),
					$taskRelationshipRepository->findBy(array('taskRelationship' => $task, 'task' => $relatedTask))
				);
				/** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskRelationship[] $taskRelationship */
				foreach($taskRelationshipArray as $taskRelationship){
					$this->entityManager->remove($taskRelationship);
				}
                $task = InstanceProvider::getRepository()->getTask()->find($task);
                $relatedTask = $this->getRelatedTasksByTasks($task);
                if($relatedTask) {
                    $onlyTask = $this->applyOrderToIssuesByTaskSingleRelatedTask($task);
                    if(!$onlyTask){
                        $this->applyOrderToIssuesByTask($relatedTask);
                    }
                }
			}
			$this->entityManager->flush();
			$connection->commit();
		}
		catch(\Exception $e){
			$connection->rollBack();
			throw $e;
		}
	}

	public function getAllTasksUnderProject($projectId, $taskId, $classificationId = null,$allFieldsField = ''){
        if($allFieldsField){
            $allFieldsField =   Utilities::createQueryFiltered($allFieldsField);
        }
        $projectChildrenIds = DevOpsUtilities::getAllProjectChildrenIdInArray($projectId);

        $tasks = array();
		$alreadyAssociated = array();
		foreach ($taskId as $taskID) {
			foreach ($this->getTaskRelationship($taskID, $classificationId) as $task) {
				if(!in_array($task, $alreadyAssociated)){
					array_push($alreadyAssociated, $task);
				}
			}
		}
        foreach ($alreadyAssociated as $task) {
            if(!in_array($task, $taskId)){
                array_push($taskId, $task['ID']);
            }
        }

        $taskId = empty($taskId) ? '0' : implode(', ', $taskId);
        $taskRepository = InstanceProvider::getRepository()->getTask();
        foreach ($projectChildrenIds as $projectChildrenId) {
            foreach ($taskRepository->getDataSetForEditProjectTaskIssue($projectChildrenId,
                $this->environment->getCurrentOrganizationId(), $taskId, null,$allFieldsField,
                null)
					 as $task) {
                if(!in_array($task, $tasks)){
                    array_push($tasks, $task);
                }
            }
        }
        return $tasks;
	}
    public function onlyCreateTaskRelationships($taskId, $issueIds,$type){

        $taskRepository = InstanceProvider::getRepository()->getTask();
        /** @var Entities\Task[] $issues */
        $issues = $taskRepository->findBy(array('id' => $issueIds));
        /** @var Entities\Task $task */
        $task = $taskRepository->find($taskId);
        $taskRelationshipFactory = InstanceProvider::getFactory()->getTaskRelationship();
        $taskRelationshipRepository = InstanceProvider::getRepository()->getTaskRelationship();
        foreach ($issues as $issue) {
            $taskRelationship = $taskRelationshipRepository
                ->findOneBy(array('task' => $issue,'type' => $type,'taskRelationship'=>$task));
            if(empty($taskRelationship)){
                $taskRelationship = $taskRelationshipFactory->createTaskRelationship($issue, $task,$type);
                $this->entityManager->persist($taskRelationship);
            }
        }
        $this->entityManager->flush();
    }

    public function validateTaskToRelate($tasks) {
        $taskClassificationIds = [];

        $taskClassifications = InstanceProvider::getRepository()->getTaskType()->taskClassificationByIssues($tasks[0]['ISSUE']);
        if(!empty($taskClassifications)) {
            foreach ($taskClassifications as $taskClassification) {
                $taskClassificationIds[] = $taskClassification['ID'];
            }
        }

        foreach ($tasks as $task) {
            $issueClassificationId = $task['ISSUE_CLASSIFICATION_ID'];
            // valida se a categoria das tarefas são iguais a categoria da issue, ou a outras categorias parametrizadas
            if($issueClassificationId !== $task['CLASSIFICATION_ID'] && !in_array($task['CLASSIFICATION_ID'], $taskClassificationIds)){
               throw new \Exception(' a tarefa "' . $task['NAME'] . '" não possui uma categoria compatível com a categoria das issues selecionadas.');
            }
        }
    }

    public function createTaskRelationshipRelatedTask($taskIds)
    {
        $type = Entities\TaskRelationship::RELATED_TASK_TYPE;
        /** @var Entities\Task $task */
        $tasks = InstanceProvider::getRepository()->getTask()->findBy(array('id' => $taskIds));
        $taskRelationshipRepository = InstanceProvider::getRepository()->getTaskRelationship();
        /** @var Entities\TaskRelationship $taskRelationship */
        $taskRelationship = $taskRelationshipRepository->findOneBy(array('task' => $tasks, 'type' => $type));
        $taskMaster = null;
        if ($taskRelationship) {
            $taskMaster = $taskRelationship->getTask();
        }
        $taskRelationshipFactory = InstanceProvider::getFactory()->getTaskRelationship();
        foreach ($tasks as $task) {
            DevOpsUtilities::setPreFix($task);
            /** @var Entities\TaskRelationship $taskRelationship */
            $taskRelationship = $taskRelationshipRepository
                ->findOneBy(array('task' => $taskMaster, 'type' => $type, 'taskRelationship' => $taskRelationship));
            if (empty($taskRelationship)) {
                if(empty($taskMaster)||$taskMaster->getId() == $task->getId()){
                    $taskMaster = $task;
                }else{
                    $taskRelationship = $taskRelationshipFactory->createTaskRelationship($taskMaster, $task, $type);
                    $this->entityManager->persist($taskRelationship);
                }
            }
        }
        $this->entityManager->flush();
    }

    public function validateMergeTaskConclude($taskId){
        $projects = array();
        $taskRepository = InstanceProvider::getRepository()->getTask();
        /** @var Entities\Task $task */
        $task = $taskRepository->find($taskId);
        $project = $task->getProject();
        $usemerge = $project->getProjectType()->getUsemergerequest();
        $isError = $taskRepository->validateIfProjectIsError($taskId);
        $classification = $task->getClassification();
        $controlArray = array('T', 'S');
        $productProject = InstanceProvider::getService()->getProject()->getProductProject($project);
        $productProjectId = $productProject ? $productProject->getId() : null;
        if(($classification && !in_array($classification->getControl(), $controlArray)) && !$isError && ($usemerge == 'N' ||
           ($usemerge == 'Y' && $classification && $classification->getControl() == Entities\Classification::CONTROL_MERGE_REQUEST))) {
            $projects = $taskRepository->validateMergeTaskInSprint($productProjectId, $taskId);
        }
        return $projects;
    }

	public function createTaskRelationship($task, $issues,$type){
		$connection = $this->entityManager->getConnection();
		$connection->beginTransaction();
		try{
            $taskRepository = InstanceProvider::getRepository()->getTask();
            /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $task */
			$task = $taskRepository->find($task);
            $taskRelationshipFactory = InstanceProvider::getFactory()->getTaskRelationship();
			foreach ($issues as $issue) {
				$issue = $taskRepository->find($issue);
				/** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskRelationship $taskRelationshipIssue */
				$taskRelationshipIssue = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskRelationship')
					->findOneBy(array('task' => $issue,'type' => \Service\TaskPrecursor::RELATIONSHIP_ISSUE_TYPE,'taskRelationship'=>$task));
				if(empty($taskRelationshipIssue)){
					$taskRelationship = $taskRelationshipFactory->createTaskRelationship($issue, $task,$type);
					$this->entityManager->persist($taskRelationship);
				}
			}
			$this->entityManager->flush();

            $isIssue = $type == \Service\TaskPrecursor::RELATIONSHIP_ISSUE_TYPE;

            $this->applyOrderToIssuesByTask($task,true,$isIssue);

			$connection->commit();
		}
		catch(\Exception $e){
			$connection->rollBack();
			throw $e;
		}
	}

	/**
	 * Finds the parameters already registered.
	 * @return array $taskOptionParameters
	 */
	public function findTaskOptionsParameters(){
		$organizationId = $this->environment->getCurrentOrganizationId();
		$taskOptionsParameters = InstanceProvider::getRepository()->getTask()->getTaskOptionsParameters($organizationId);
		return $taskOptionsParameters;
	}

	/**
	 * Updates the task option parameter.
	 * @param int $propertyId
	 * @param $initialTaskStatusId
	 * @param $finalTaskStatusId
	 */
	public function updateTaskOptionsParameter($propertyId, $initialTaskStatusId, $finalTaskStatusId){
		$taskProperty 		= $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskProperties')->findBy(array('id' => $propertyId))[0];
		$initialTaskStatus 	= $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskStatus')->findBy(array('id' => $initialTaskStatusId))[0];
		$finalTaskStatus 	= $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskStatus')->findBy(array('id' => $finalTaskStatusId))[0];
		$updatedTaskProperty = InstanceProvider::getFactory()->getTask()->updateTaskOptionsParameter($taskProperty, $initialTaskStatus, $finalTaskStatus);
		$this->entityManager->persist($updatedTaskProperty);
		$this->entityManager->flush();
	}

	/**
	 * Creates a new task option parameter.
	 * @param $initialTaskStatusId
	 * @param $finalTaskStatusId
	 */
	public function createTaskOptionsParameter($initialTaskStatusId, $finalTaskStatusId){
		$propertyInitialTaskStatus = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskStatus')->findBy(array('id' => $initialTaskStatusId))[0];
		$propertyFinalTaskStatus = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskStatus')->findBy(array('id' => $finalTaskStatusId))[0];
		$propertyControl = '';
		$propertyLabel = '';
		$propertyTaskType = '';
		$propertyIsIssue = 'N';

		$createdTaskProperty = InstanceProvider::getFactory()->getTask()->createTaskOptionsParameter($propertyInitialTaskStatus, $propertyFinalTaskStatus, $propertyControl,
			$propertyLabel, $propertyTaskType, $propertyIsIssue);
		$this->entityManager->persist($createdTaskProperty);
		$this->entityManager->flush();
	}

	/**
	 * @param $taskId
	 * @return \Datetime
	 */
	public function viewCommentsSaveData($taskId){
		$user = $this->environment->getCurrentUser();
		$issue = InstanceProvider::getRepository()->getTask()->find($taskId);
        $lastDate = null;
        if($issue){
            /** @var Entities\Allocation $allocation */
            $allocations = InstanceProvider::getRepository()->getAllocation()->findBy(array('task' => $issue ,'resource' => $user));
            if(empty($allocations)){
                $allocations = InstanceProvider::getService()->getIssue()->addFollower($issue, $user);
                $allocations->setLastPostView($this->environment->getCurrentTime());
	            $this->entityManager->persist($allocations);
	            $this->entityManager->flush();
            }else{
            	foreach($allocations as $allocation){
            		$allocation->setLastPostView($this->environment->getCurrentTime());
		            $this->entityManager->persist($allocation);
		            $this->entityManager->flush();
	        		$lastDate = $allocation->getLastPostView();
            	}
            }
        }
		return $lastDate;
	}
	public function viewCommentsSaveDataFist($taskId){
		$lastDate = null;
		$comments = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Comments')->findBy(array( 'task' => $taskId ));

		if( count($comments) <= 1){
			$lastDate = $this->viewCommentsSaveData($taskId);
		}
		return $lastDate;
	}

	protected function dateStringToDiminishedWeekDay($date){
		return strtoupper(substr(\DateTime::createFromFormat('d/m/Y', $date)->format('D'), 0, 2));
	}

	protected function calculateMaxHoursDay($day,  $calendarDate)
    {
        $workedTime = InstanceProvider::getRepository()->getWorkedTime()->getWorkTimeByUserAndDate($this->environment->getCurrentUserId(), $day);
        $dailyHours = $this->setDefaultCalendar($day);
        if (empty($workedTime)) {
            $calendarDay = $this->getCalendarDay($day, $calendarDate);
            if ($calendarDay) {
                if ($calendarDay['TYPE'] == 'O') {
                    $dailyHours = 0;
                } else {
                    $dailyHours = $calendarDay['DAILY_HOURS'];
                }
            } else {

                $dailyHours = $this->setDefaultCalendar($day);
            }
        } else {
            $dailyHours = $workedTime[0]['WORK_TIME'];
        }
        return $dailyHours;

    }

    protected function calculateDailyHours($day, array $calendarDate,$workedTime)
    {
        if (empty($workedTime)) {
            $calendarDay = $this->getCalendarDay($day, $calendarDate);
            if ($calendarDay) {
                if ($calendarDay['TYPE'] == 'O') {
                    $dailyHours = 0;
                } else {
                    $dailyHours = $calendarDay['DAILY_HOURS'];
                }
            } else {

                $dailyHours = $this->setDefaultCalendar($day);
            }
        } else {
            $dailyHours = $workedTime['WORK_TIME'];
        }
        return $dailyHours;

    }

    protected function getCalendarDay($day,$calendarDates){
        foreach($calendarDates as $calendarDate ){
            if($calendarDate['DATA'] == $day){
                $calendarDate['DAILY_HOURS'] = Utilities::timeToSeconds($calendarDate['DAILY_HOURS']);
                return $calendarDate ;
            }
        }
        return null;
    }

	protected function in_tasksDate_array($taskDate, $array){
		$return = false;
		foreach ($array as $object) {
			if($object['ID'] === $taskDate['ID'] && $object['DAY'] === $taskDate['DAY']){
				$return = true;
			}
		}
		return $return;
	}

    /**
     * @param \DateTime $datePrevious
     * @param \DateTime $Date
     * @return array
     */
    public function getDiffAppropriationOfWorkDay(\DateTime $datePrevious,\DateTime $Date)
    {
        $nowString = $Date->format(Utilities::FORMAT_DATE_DEFAULT_ONLY_DATE);
        $datePreviousString = $datePrevious->format(Utilities::FORMAT_DATE_DEFAULT_ONLY_DATE);
        $rangeFilter = array(
            "name" => "range",
            "operator" => "BETWEEN",
            "value" => array( $datePreviousString,$nowString, null)
        );

        $rangeFilter = $this->buildQueryForTimeControl($rangeFilter);
        $range = $rangeFilter['range'];
        $userId = $this->environment->getCurrentUserId();
        $organizationId = $this->environment->getCurrentOrganizationId();
        $allAppropriation = InstanceProvider::getRepository()->getTask()->getTotalAppropriatedForDayInRange($userId,$range,$organizationId);
        $allAppropriation = Utilities::getArrayAssociativeFromArray($allAppropriation,'DAY');
        $calendarDates = InstanceProvider::getService()->getResource()->findExceptionCalendarDatesResource($userId,$datePreviousString,
            $nowString);
        $workedTimeDates = InstanceProvider::getRepository()->getWorkedTime()->getWorkTimeByUserAndBetweenDates($userId,$range);
        $workedTimeDates = Utilities::getArrayAssociativeFromArray($workedTimeDates,'WORK_DATE');
        $isPending = null;
        foreach($calendarDates as &$calendarDate) {
            $day = $calendarDate['DATA']; // format: 01/01/2000
            $calendarDate['DATE'] = $calendarDate['DATA']; // format: 01/01/2000
            $workedTime = empty($workedTimeDates[$day]) ? null : $workedTimeDates[$day];
            $appropriation = empty($allAppropriation[$day]) ? null : $allAppropriation[$day];

            $maxHours = $this->calculateDailyHours($day, $calendarDates, $workedTime); // gets from WORKED_TIME entity
            $differenceHours = abs($maxHours) - $appropriation['HOURS'];
            $differenceHoursLabel = '0';
            if ($differenceHours > 0) {
                $differenceHoursLabel = '- ';
            } else if ($differenceHours < 0) {
                $differenceHoursLabel = '+ ';
            }
            $differenceHoursLabel .= $this->environment->secondsToTime(abs($differenceHours));
            $calendarDate['HOURS_CONTROL'] = $differenceHoursLabel;
            $calendarDate['IS_PENDING'] = $workedTime['IS_PENDING'];

            if($isPending){
                $calendarDate['IS_PENDING'] = $isPending;
            } else if($calendarDate['IS_PENDING'] == Entities\WorkedTime::IS_PENDING){
                $isPending = Entities\WorkedTime::IS_PENDING;
            }

            if($isPending){
                $calendarDate['IS_PENDING'] = $isPending;
            }
        }

        return $calendarDates;
    }


    protected function compareTimeControlDates(){
		return function($a, $b){
			$return = intval(\DateTime::createFromFormat('d/m/Y', trim(substr($a, 2)))
				->diff(\DateTime::createFromFormat('d/m/Y', trim(substr($b, 2))))->format('%r%a'));
			return $return === 0 ? 0 : $return > 0 ? 1 : -1;
		};
	}

    public function removeWeekDayOfString($string){
        $array = explode(' ',$string);
        return end($array);
    }

    public function appropriationValidationOfLast30Day()
    {
        $user = $this->environment->getCurrentUser();

        $structure = InstanceProvider::getService()->getResource()->getStructureForResource($user);
        $appropriationValidation = array("PREVIOUS_DAY_APPROPRIATION_OK" => true,'DATES' => '');
        if($user){
            $appropriationValidation['USER_NAME'] = $user->getName();
            $appropriationValidation['USER_ID'] = $user->getId();
        }

        if ($this->validateAvailableFromToday() && $structure && $structure->getTimeSheetRequired() == 'Y') {
            $now = new \DateTime();
            $yesterday = $now->modify('-1 day');
            $now = new \DateTime();
            $last30Day = $now->modify('-31 day');
            if($user->getAvailableFrom() && $user->getAvailableFrom() > $last30Day ){
                $last30Day = clone $user->getAvailableFrom();
            }
            $allDiffAppropriation = $this->getDiffAppropriationOfWorkDay($last30Day,$yesterday);
            $previousDayAppropriationOk = true;
            foreach ($allDiffAppropriation as $diff) {
                if ($diff['IS_PENDING'] !== 'Y' && $diff['HOURS_CONTROL'] != '0') {
                    $appropriationValidation['DATES'][] = $diff['HOURS_CONTROL'].' no dia '.$diff['DATE'];
                    $previousDayAppropriationOk = false;
                }
            }
            $appropriationValidation['PREVIOUS_DAY_APPROPRIATION_OK'] = $previousDayAppropriationOk;

        }

        return array($appropriationValidation);

    }

    private function validateAvailableFromToday(){
        $user = $this->environment->getCurrentUser();
        $today = new \DateTime();
        return !($user->getAvailableFrom() && Utilities::compareTwoDates($user->getAvailableFrom(), $today));
    }

	/********************************************************************************
	 * Gets all appropriated tasks filtering it by a range.
	 * These appropriations are shown on Time Control screen.
     * Trust me, if you're here for modify this code, you're gonna cry.
	 * @param $rangeFilter
	 * @return array
	 * @throws \Exception
     ********************************************************************************/
	public function getTasksAppropriatedForCurrentUserInRange($rangeFilter, $hideWeekend = null){
		try{

            if($hideWeekend == "Y"){
                $rangeFilter['value'][2] = "N";
            }else{
                $rangeFilter['value'][2] = "Y";
            }
			$rangeFilter = $this->buildQueryForTimeControl($rangeFilter);
			$range = $rangeFilter['range'];
			$begin = $rangeFilter['begin'];
            $hideWeekend = empty($hideWeekend) ? null : $hideWeekend;

            $end = $rangeFilter['end'];
			$weekendQuery = $rangeFilter['weekendQuery'];
            $weekendQueryDate = $rangeFilter['weekendQueryDate'];

            // Get all current user's tasks which have appropriations
			$tasks = InstanceProvider::getRepository()->getTask()->getTasksAppropriatedForUserInRange($this->environment->getCurrentUserId(),
                $this->environment->getCurrentOrganizationId(), $begin, $end, $range, $weekendQuery, $weekendQueryDate);
            // Creates the line which contains total of appropriations in each day
            $dailyTotalHours = array();
			$this->createTotalAppropriationsForEachDayAndTask($tasks, $dailyTotalHours);

            // Creates each column of the crosstable
			$this->createColumnForEachDayAndTask($tasks, $dailyTotalHours, $begin, $end);

			uksort($dailyTotalHours, $this->compareTimeControlDates());

			// Useful only for be used on calculateMaxHoursDay(), but we inserted out of foreach()
            $calendarDate = InstanceProvider::getService()->getResource()->findExceptionCalendarDatesResource($this->environment->getCurrentUserId(),$begin,$end);


            $tasks = $this->createAppropriationDifferenceForEachDayAndTask($tasks,$dailyTotalHours, $calendarDate);

            if($hideWeekend == '%Y%'){
				$tasks = $this->removeWeekendsFromResult($tasks, $calendarDate);
			}

            // Rename week days by their numbers
			$tasks = $this->renameWeekDayNumberToDayName($tasks);

			return $tasks;

		}
		catch(\Exception $e){
			throw $e;
		}

	}


    /********************************************************************************
     * This method is useful for Time Control screen. It creates a fake task named as
     * Diferença de Apropriação, and shows, in the truth, the difference of appropriation
     * in each day in each task (if that day has a negative, positive or right appropriation).
     * @param $tasks
     * @param $dailyTotalHours
     * @param $calendarDate
     * @return mixed
     ********************************************************************************/
    protected function createAppropriationDifferenceForEachDayAndTask(&$tasks,$dailyTotalHours, $calendarDate)
    {
        foreach ($dailyTotalHours as $day => $data) {
            $maxHours = $this->calculateMaxHoursDay(trim(substr($day, 2)), $calendarDate); // gets from WORKED_TIME entity
            $differenceHours = abs($maxHours) - $data['HOURS'];
            $differenceHoursLabel = '0';

            if ($differenceHours > 0) {
                $differenceHoursLabel = '- ';
            } else if ($differenceHours < 0) {
                $differenceHoursLabel = '+ ';
            }

            $differenceHoursLabel .= $this->environment->secondsToTime(abs($differenceHours));

            array_unshift($tasks, array(
                'ID' => '0', // id <= 0 for apply a specified CSS style to the row
                'TASK_NAME' => 'Diferença de Apropriação',
                'HOURS' => $differenceHoursLabel,
                'HOURS_CONTROL' => $differenceHours,
                'DAY' => $day,
                'DAY_LABEL' => $day
            ));
            array_unshift($tasks, array(
                'ID' => $data['ID'],
                'TASK_NAME' => $data['TASK_NAME'],
                'HOURS' => $data['HOURS'] > 0 ? $this->environment->secondsToTime($data['HOURS']) : '0',
                'DAY' => $day,
                'DAY_LABEL' => $day
            ));

        }
        return $tasks;
    }


    /********************************************************************************
     * This method is useful for Time Control screen. It creates a fake task named as
     * Total Apropriado, and shows, in the truth, the total of appropriations for each
     * day in each task.
     ********************************************************************************/
    private function createTotalAppropriationsForEachDayAndTask(&$tasks, &$dailyTotalHours){
        foreach($tasks as $key => $task){
            $day = $task['DAY']; // format: 1 01/01/2000
            if(empty($dailyTotalHours[$day])){
                $dailyTotalHours[$day]['ID'] = '-1';
                $dailyTotalHours[$day]['TASK_NAME'] = 'Total apropriado';
                $dailyTotalHours[$day]['HOURS'] = intval($task['HOURS']);
            } else {
                $dailyTotalHours[$day]['HOURS'] += intval($task['HOURS']);
            }
            $tasks[$key]['HOURS'] = empty($task['HOURS']) ? 0 : $this->environment->secondsToTime($task['HOURS']);
        }
    }

    /********************************************************************************
     * This method is useful for Time Control screen. It creates the columns of the
     * CrossTable. The screen has a date range, and shows appropriations for each
     * date in this range.
     ********************************************************************************/
    private function createColumnForEachDayAndTask(&$tasks, &$dailyTotalHours, $begin, $end){
        $rangeStart = \DateTime::createFromFormat('d/m/Y', $begin);
        $rangeEnd = \DateTime::createFromFormat('d/m/Y', $end);

        // Navigates into each day in the range
        while($rangeEnd >= $rangeStart){
            $day = $rangeStart->format('d/m/Y');
            $dayWeek = intval($rangeStart->format('w'))+1;
            $day = $dayWeek.' '.$day;

            if(empty($dailyTotalHours[$day])){
                $dailyTotalHours[$day]['ID'] = '-1'; // id <= 0 for apply a specified CSS style to the row
                $dailyTotalHours[$day]['TASK_NAME'] = 'Total apropriado';
                $dailyTotalHours[$day]['HOURS'] = 0;
            }

            $rangeStart->modify('+1 day');
        }
    }

    /********************************************************************************
	 * Since week day has, as default, an integer value (0 - 6),
	 * this method renames them to readable names (Monday, Thuesday,
	 * etc.) for the final user.
     ********************************************************************************/
	private function renameWeekDayNumberToDayName($tasks){
		foreach($tasks as $key => &$task){
			$day = $task['DAY'];
            $dayLabel = $task['DAY_LABEL'];
			$dayWeek = intval(substr($day,0,1));

			$dayWeek = \Service\Calendar\TaskCalendar::$ARRAY_WEEK_BR_LABEL[$dayWeek-1];
			$task['DAY'] = $dayWeek.' '.substr($day, 2);
            $task['DAY_LABEL'] = $dayWeek.' '.substr($dayLabel, 2);
		}
		return $tasks;
	}

	/********************************************************************************
	 * Builds a query for gets appropriated tasks for Time Control
	 * screen.
	 * @param $rangeFilter
	 * @return array
     ********************************************************************************/
	private function buildQueryForTimeControl($rangeFilter){
		if(empty($rangeFilter)){
			$begin =  $this->environment->getCurrentTime()->modify('-6 day')->format('d/m/Y');
			$end = $this->environment->getCurrentTime()->format('d/m/Y');
			$showWeekend = 'N';
		} else {
			$begin =  $rangeFilter['value'][0];
			$end = $rangeFilter['value'][1];
			$showWeekend = $rangeFilter['value'][2];
		}
		$range = "BETWEEN TO_DATE('{$begin} 00:00:00', 'DD/MM/YYYY HH24:MI:SS')
						AND TO_DATE('{$end} 23:59:59', 'DD/MM/YYYY HH24:MI:SS')";

		$weekendQuery = $weekendQueryDate = '';
		if($showWeekend == 'N'){
			$weekendQuery = "AND TO_CHAR(A.BEGIN_HOUR, 'D') NOT IN (1,7)";
            $weekendQueryDate = "AND TO_CHAR(D.DATA, 'D') NOT IN (1,7)";
		}

		return array(
			'range' => $range,
			'showWeekend' => $showWeekend,
			'begin' => $begin,
			'end' => $end,
			'weekendQuery' => $weekendQuery,
			'weekendQueryDate' => $weekendQueryDate
		);
	}

    /**
     * User can choose if wants to show weekends or not.
     * This method removes weekends from an array and
     * return another one.
     * @param $tasks
     * @param $calendarDates
     * @throws \Exception
     * @return array
     */
	private function removeWeekendsFromResult($tasks, $calendarDates){
		$filteredTasks = array();
		foreach($tasks as $task){
            $date = substr($task['DAY'], 2);
            if(!Utilities::dateNotUtil($calendarDates,$date)){
                $filteredTasks[] = $task;
            }
		}
		return $filteredTasks;
	}

    /********************************************************************************
     * Get all tasks which have not appropriations made by current user.
     * @param array $rangeFilter
     * @return array
     ********************************************************************************/
	public function getTasksNotAppropriatedForCurrentUser($rangeFilter){
		if(empty($rangeFilter)){
			$begin =  $this->environment->getCurrentTime()->modify('-5 day')->format('d/m/Y');
			$end = $this->environment->getCurrentTime()->format('d/m/Y');
			$range = "BETWEEN TO_DATE( '$begin 00:00:00' ,'DD/MM/YYYY HH24:MI:SS')
                                AND TO_DATE( '$end 23:59:59' ,'DD/MM/YYYY HH24:MI:SS')";
		}else{
            $value = "value";
            if(isset($rangeFilter[0])) {
                $start = $rangeFilter[0];
                $end = $rangeFilter[1];
            }else {
                $start = $rangeFilter[$value][0];
                $end = $rangeFilter[$value][1];
            }
            $range = "BETWEEN TO_DATE( '$start 00:00:00' ,'DD/MM/YYYY HH24:MI:SS')
                                AND TO_DATE( '$end 23:59:59' ,'DD/MM/YYYY HH24:MI:SS')";
        }

        $taskRepository = InstanceProvider::getRepository()->getTask();
		$tasks = $taskRepository->getTasksNotAppropriatedForUser($this->environment->getCurrentUserId(), $range);
		$tasksIds = array();
		foreach ($tasks as $key => $task) {
			if(!in_array($task['ID'], $tasksIds)){
				array_push($tasksIds, $task['ID']);
			}
			if(empty($task['PARENT'])){
				$task['PARENT'] = $task['PROJECT'];
				$tasks[$key]['PARENT'] = $task['PROJECT'];
				$parentRepository = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Project');
			}
			else{
				$parentRepository = $taskRepository;
			}
			if(!in_array($task['PARENT'], $tasksIds)){
				array_push($tasksIds, $task['PARENT']);
				array_push($tasks, array(
					"ID" => $task['PARENT'],
					"TASK_NAME" => $parentRepository->find($task['PARENT'])->getName()
				));
			}
		}
		return $tasks;
	}
	public function getTasksForCurrentUser(){

		$tasks = InstanceProvider::getRepository()->getTask()->getTasksForCurrentUser($this->environment->getCurrentUserId());

		return $tasks;
	}

    public static function getArrayTaskRelationshipForResponseTypes(){
        return array(
            Entities\TaskRelationship::ERROR_MR_TYPE,
            Entities\TaskRelationship::ERROR_TYPE,
        );
    }
    public static function getArrayTaskRelationshipForMergeRequest(){
        return array(
            Entities\TaskRelationship::MERGE_REQUEST_TYPE,
            Entities\TaskRelationship::ORIGIN_TYPE,
        );
    }

    public static function getArrayClassificationForError(){
        return array(
            Entities\Classification::CONTROL_ERROR_MERGE_REQUEST,
            Entities\Classification::CONTROL_ERROR_TEST,
        );
    }

    public static function addTaskResponse(Entities\Task $task,&$takIds){

        $response = "";
        $id = $task->getId();
        if(!in_array($id,$takIds)){
            $takIds[] = $id;
            $response =  "-- Resposta da tarefa $id --  \n ".$task->getResponse()."\n\n";
        }
        return $response;
    }

    public function getRecursiveResponsesInMergeRequest($taskPrime,$taskResponse, $takIds){
        /** @var Entities\TaskRelationship[] $taskRelationshipMerges*/
        $taskRelationshipMerges = InstanceProvider::getRepository()->getTaskRelationship()->findBy(array('taskRelationship' => $taskPrime,
            'type'=> self::getArrayTaskRelationshipForMergeRequest()));
        foreach ($taskRelationshipMerges as $taskRelationshipMerge) {
            $task = $taskRelationshipMerge->getTask();
            $taskResponse[] = self::addTaskResponse($task, $takIds);
            $taskResponse = $this->getRecursiveResponsesForMergeRequest($task,$taskResponse, $takIds);
        }
        return $taskResponse;
    }

    public function getRecursiveResponsesForMergeRequest($taskMerge,$taskResponse = array(), $takIds){
        if($taskMerge){
            /** @var Entities\TaskRelationship[] $taskRelationships*/
            $taskRelationships = InstanceProvider::getRepository()->getTaskRelationship()->findBy(array('taskRelationship' => $taskMerge,
                'type'=> Entities\TaskRelationship::MERGE_REQUEST_TYPE));
            foreach ($taskRelationships as $taskRelationship) {
                $task = $taskRelationship->getTask();
                $taskResponse = $this->getResponsesInTest($task,$taskResponse, $takIds);
            }
        }
        return $taskResponse;
    }

    public function getResponsesInTest($taskPrime,$taskResponse, $takIds){
        /** @var Entities\TaskRelationship[] $taskRelationships*/
        $taskRelationships = InstanceProvider::getRepository()->getTaskRelationship()->findBy(array('taskRelationship' => $taskPrime,
            'type'=> Entities\TaskRelationship::TEST_TYPE));
        foreach ($taskRelationships as $taskRelationship) {
            $task = $taskRelationship->getTask();
            $taskResponse = $this->getRecursiveResponsesInTest($task,$taskResponse, $takIds);
        }
        return $taskResponse;
    }

    public function getRecursiveResponses(Entities\Task $taskPrime, $taskResponse = array(), $takIds){
        $taskResponse[] = self::addTaskResponse($taskPrime, $takIds);
        /** @var Entities\TaskRelationship $taskRelationship */
        $taskRelationships = InstanceProvider::getRepository()->getTaskRelationship()->findBy(array('taskRelationship' => $taskPrime,
            'type'=> self::getArrayTaskRelationshipForResponseTypes()));
        foreach ($taskRelationships as $taskRelationship) {
            $task = $taskRelationship->getTask();
            $taskResponse = $this->getRecursiveResponsesInTest($task,$taskResponse, $takIds);
            $taskResponse = $this->getRecursiveResponsesInMergeRequest($task, $taskResponse, $takIds);
        }
        return $taskResponse;
    }


    protected function getRecursiveResponsesInTest($taskPrime,$taskResponse = array(),$takIds = array())
    {
        /** @var Entities\TaskRelationship $taskRelationship */
        $taskRelationship = InstanceProvider::getRepository()->getTaskRelationship()->findOneBy(array('taskRelationship' => $taskPrime,
            'type' => Entities\TaskRelationship::TEST_TYPE));
        if($taskRelationship){
            $task = $taskRelationship->getTask();
            $taskResponse = $this->getRecursiveResponses($task, $taskResponse,$takIds);
        }
        return $taskResponse;
    }

    public static function responseAdjustment( array $taskResponses){
        $taskResponses = array_reverse($taskResponses);
        $response = '';
        foreach ($taskResponses as $taskResponse) {
            $response .= $taskResponse;
        }
        return $response;

    }

    public function getAllResponsesForTestTask($taskId)
    {
        $task = InstanceProvider::getRepository()->getTask()->findOneBy(array('id' => $taskId));
        $taskResponse = $this->getRecursiveResponsesInTest($task);
        $taskResponse = self::responseAdjustment($taskResponse);
        return $taskResponse;
    }

	public function getTaskForTimeControl($taskId){
        $userId = $this->environment->getCurrentUserId();
        $taskRepository = InstanceProvider::getRepository()->getTask();
		$tasks = $taskRepository->getTaskDataForTimeControl($taskId, $userId);
        $appropriationService = InstanceProvider::getService()->getAppropriation();

		if(!empty($tasks)) {
            $task = $tasks[0];
            /** @var Entities\Task $taskEntity */
            $taskEntity = $taskRepository->find($taskId);
            $task['BRANCH_INFO'] = $taskEntity->getTaskBranchRecursive();
            $task['DURATION'] = $this->environment->secondsToTime($task['DURATION']);
			$task['PATHS'] = $appropriationService->findPaths($task['TASK_ID'], $userId);
			$task['PATHS'] = $this->setBranchMasters($task['PATHS'],$task['PROJECT_EXTERNALTYPE']);
            $task['PATHS_UNCHANGED'] = $task['PATHS'];
            $task['DIRECTORIES_VALIDATION'] = $appropriationService->validateProjectRepositories($task['TASK_ID']);
            $taskMainRelated = \Util\Utilities::getMainRelatedTaskByTaskId($task['TASK_ID']);
            $task['MAIN_RELATED_ID'] = $taskMainRelated ? $taskMainRelated->getId() : null;
            if($task['TYPE_RELATIONSHIP'] == Entities\TaskRelationship::TEST_TYPE) {
                $task['TASK_RESPONSE'] = $this->getAllResponsesForTestTask($task['ID']);
            }
            if(strpos($task['TASK_NAME'], '[REP-MASTER]')) {
                $task['ORIGIN_PRODUCT_VERSION_ID_DEV'] = null;
            }
            $task['VERSION_TAG'] = $this->getVersionTag($task['PRODUCT_VERSION_ID']);
            $tasks = array($task);
		}
		return $tasks;
	}

    public function getVersionTag($productVersionId) {
        $versionTag = null;

        if($productVersionId) {
            /** @var EntitiesDevOps\ProductVersion $productVersion */
            $productVersion = InstanceProvider::getRepository()->getProductVersion()->find($productVersionId);
            $currentBuild   = InstanceProvider::getRepository()->getProductVersion()->findCurrentBuildFromVersion (
                $productVersion->getProduct()->getId(), $this->environment->getCurrentOrganizationId(),
                $productVersion->getVersion(), $productVersion->getRevision()
            )[0];
            $versionTag = 'PROD_' . $productVersion->getProduct()->getId() . '_' . $currentBuild['VERSION'];
        }

        return $versionTag;
    }

    public function getBranchInfoByTaskId($taskId){
        $taskEntity = InstanceProvider::getRepository()->getTask()->find($taskId);
        return $taskEntity->getTaskBranchRecursive();
    }

    public function setBranchMasters($paths,$projectExternalType){
        $repositoryRepository = InstanceProvider::getRepository()->getRepository();
        /** @var Entities\Repository $repository */
        foreach ($paths as &$path) {
            if($projectExternalType == Entities\ProjectType::TYPE_PROJECT_CONTROL_ERROR){
                $repository = $repositoryRepository->find($path['ID'][0]);
                $path['MASTER_BRANCH'] = 'FIX_ERROR_PROJECT_' . $repository->getProject()->getId();
            }
        }
        return $paths;
    }

	public function getTaskAllocationForCurrentUser($taskId){
		return InstanceProvider::getRepository()->getAllocation()->getAllocationByTaskAndResource($taskId, $this->environment->getCurrentUserId());
	}

	/**
	 * Originally, task status are saved using
	 * Zeedhi Crud. But since Crud can not work
	 * properly with foreign keys, this method
	 * takes care about these other values.
	 * @param array $status
	 */
	public function saveTaskStatus($status){
		$entityTaskType = $this->entityManager->getRepository(Utilities::REPOSITORY_DIR.'TaskType')->findBy(array('id' => $status['TASK_TYPE']))[0];
		$entityTask		= $this->entityManager->getRepository(Utilities::REPOSITORY_DIR.'TaskStatus')->findBy(array('id' => $status['ID']))[0];
		$entityTaskType = InstanceProvider::getFactory()->getTaskType()->updateTaskTypeFromTaskStatus($entityTaskType, $entityTask);
		$this->entityManager->persist($entityTaskType);
		$this->entityManager->flush();
	}


	/**
	 * @param array $labels
	 * @param Entities\Task $task
	 */
	public function setLabels(array $labels,Entities\Task $task)
	{

        $taskLabelFactory = InstanceProvider::getFactory()->getTaskLabel();
        $labelRepository = InstanceProvider::getRepository()->getLabel();
        $taskLabelRepository = InstanceProvider::getRepository()->getTaskLabel();
		foreach ($labels as $labelId) {
			if(!empty($labelId)){
				/** @var  Entities\Label $label */
				$label = $labelRepository->find($labelId);
				$taskLabel = $taskLabelRepository	->findBy(array("task" => $task->getId(),
                    "label" => $label->getId()));
				if (empty($taskLabel)) {
					$taskLabel = $taskLabelFactory->createTaskLabel($task, $label);
					$this->entityManager->persist($taskLabel);
				}
			}
		}
        $this->entityManager->flush();
		/** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskLabel[] $taskLabels */
		$taskLabels = $taskLabelRepository->findBy(array("task" => $task->getId()));
		foreach ($taskLabels as $taskLabel) {
			if(!in_array($taskLabel->getLabel()->getId(), $labels)){
				$this->entityManager->remove($taskLabel);
				$this->entityManager->flush();
			}
		}
	}
    public function findLabelByBoardIssue($projectId,$param = array(), $board = null){
        $projectIds = $this->getImInMeetingProjectIds();
        $board = (int)$board;
        $projectId = $board;
        $projectIds[] = $projectId;
        $filters = DevOpsUtilities::createQueryFiltered($param);
        return InstanceProvider::getRepository()->getTask()->findLabelByBoardIssue($projectIds,$this->environment->getCurrentOrganizationId(),$this->environment->getCurrentUserId(),$filters);
    }

    public function findLabelsByProject($projectId,$param = array()){
        $projectIds = $this->getImInMeetingProjectIds();
        if(!empty($projectId)){
            if(is_array($projectId)){
                $projectIds = array_merge($projectId,$projectIds);
            }else{
                $projectIds[] = $projectId;
            }
        }
        $filters = DevOpsUtilities::createQueryFiltered($param);
        return InstanceProvider::getRepository()->getTask()->findLabelsByProject($projectIds,$this->environment->getCurrentOrganizationId(),$this->environment->getCurrentUserId(),$filters);
    }
    public function findLabelFromLikeProjectName($filters){
        $filtersString = $this->environment->createQueryFiltered($filters);

		return InstanceProvider::getRepository()->getTask()->findLabelsByFilter($this->environment->getCurrentOrganizationId(),$this->environment->getCurrentUserId(),$filtersString);
	}

	public function saveChangedFiles($files, $task, $fileType, $repository) {
		$organization = $this->environment->getCurrentOrganizationId();
		$connection = $this->entityManager->getConnection();
		try {
			$connection->beginTransaction();
			$taskEntity = InstanceProvider::getRepository()->getTask()->find($task);
            /** @var Entities\TaskRelationship $taskRelationship */
            $taskRelationship = InstanceProvider::getRepository()->getTaskRelationship()->findOneBy(array('taskRelationship' => $task, 'type' => Entities\TaskRelationship::RELATED_TASK_TYPE));
            if(!empty($taskRelationship)) {
                $taskEntity = $taskRelationship->getTask();
            }
            $changedFilesFactory = InstanceProvider::getFactory()->getChangedFiles();
            $repositoryRepository = InstanceProvider::getRepository()->getRepository();
            $changedFilesRepository = InstanceProvider::getRepository()->getChangedFiles();
			foreach ($files as $file) {
                $repositoriesFile = $repository ? $repository : $file['REPOSITORY_ID'];
                if(!is_array($repositoriesFile)) {
                    $repositoriesFile = array($repositoriesFile);
                }
                foreach($repositoriesFile as $repositoryFile) {
                    $repositoryEntity = $repositoryRepository->find($repositoryFile);
                    $fileName = isset($file['name']) ? '\\' . $file['name'] : $file['NAME'];
                    $fileName = str_replace('/', '\\', $fileName);
                    if ($fileName) {
                        $existingFile = $changedFilesRepository->findOneBy(array("task" => $task, "repository" => $repositoryFile, "changedFile" => $fileName));
                        if (empty($existingFile)) {
                            $changedFile = $changedFilesFactory->saveChangedFile($organization, $fileName, $taskEntity, $repositoryEntity, $fileType);
                            $this->entityManager->persist($changedFile);
                        } else {
                            $existingFile->setType($fileType);
                            $this->entityManager->persist($existingFile);
                        }
                    }
                }
			}
			$this->entityManager->flush();
			$connection->commit();
		} catch (\Exception $e) {
			$connection->rollBack();
			throw $e;
		}
	}

	public function findMasterFiles($files, $repoId, $taskId){
		$masterFiles = [];
		$newFile = null;
		$connection = $this->entityManager->getConnection();
		$connection->beginTransaction();
		try {
            $changedFilesRepository = InstanceProvider::getRepository()->getChangedFiles();
            $taskRelationshipRepository = InstanceProvider::getRepository()->getTaskRelationship();
			foreach ($files as $file) {
				$fileName = ltrim($file['FILE_NAME']);
				$fileName = '\\'.$fileName;
				$fileName = str_replace('/', '\\', $fileName);
				$branchFile = $changedFilesRepository->findOneBy(array("task" => $taskId,
                    "repository" => $repoId, "changedFile"=> $fileName));

                $taskRelationship = $taskRelationshipRepository->findOneBy(array("taskRelationship" => $taskId));
                if($taskRelationship) {
                    $mainBranchFile = $changedFilesRepository->findOneBy(array("task" => $taskRelationship->getTask()->getId(),
                        "repository" => $repoId, "changedFile"=> $fileName));
                } else {
                    $mainBranchFile = null;
                }

				if((empty($branchFile) && empty($mainBranchFile)) ||
                   ($mainBranchFile && $mainBranchFile->getType() == 'S')) {
					$fileName = substr($fileName, 1);
					$newFile['name']   = str_replace('\\', "/", $fileName);
					$newFile['status'] = isset($file['FILE_STATUS']) ? $file['FILE_STATUS'] : 'M';
					$newFile['REPOSITORY_ID'] = $repoId;
					array_push($masterFiles, $newFile);
				}
			}
			return $masterFiles;

		} catch (\Exception $e) {
			$connection->rollBack();
			throw $e;
		}
	}

	public function bringChangedFilesByType($files, $repoId, $taskId, $conflictOnly){
		$branchFiles = [];
		$masterFiles = [];
		$allFiles = array();
		$newFile = null;
        $conflictTypes = self::getConflictTypes();
		try {
            $changedFilesRepository = InstanceProvider::getRepository()->getChangedFiles();
			foreach ($files as $file) {
				$fileName = ltrim($file['FILE_NAME']);
				$fileStatusMin = ltrim($file['FILE_STATUS_MIN']);
				$fileName = '\\'.$fileName;
				$fileName = str_replace('/', '\\', $fileName);
				$branchFile = $changedFilesRepository->findOneBy(array("task" => $taskId, "repository" => $repoId,
					"changedFile"=> $fileName));
				if((in_array($fileStatusMin, $conflictTypes) ||
                   ($conflictOnly &&
                   (!empty($branchFile) && $branchFile->getType() === Entities\ChangedFiles::TYPE_LOCAL_FILE) || empty($branchFile))) ) {
					$file['FILE_NAME'] = trim($file['FILE_NAME']);
					array_push($branchFiles, $file);
				} else {
					$file['FILE_NAME'] = trim($file['FILE_NAME']);
					array_push($masterFiles, $file);
				}
			}
			$allFiles['BRANCH_FILES'] = $branchFiles;
			$allFiles['MASTER_FILES'] = $masterFiles;
			return $allFiles;

		} catch (\Exception $e) {
			throw $e;
		}
	}

	public function isTaskProjectDated($task){
		$task = InstanceProvider::getRepository()->getTask()->find($task);
		return $task->getProject()->getProjectType()->getUsetimeplanning() === 'Y';
	}

	public function getMyIssueLabels($filter = null){
		$conditions = Utilities::createQueryFiltered($filter);
		return InstanceProvider::getRepository()->getLabel()->getLabelsByUserAndProject($this->environment->getCurrentUserId(), $conditions);
	}

	public function findMeetingLabels($filters){
		return InstanceProvider::getRepository()->getLabel()->findMeetingLabels($filters);
	}

	public function findMeetingLabelsForSchedule($projectId){
		return InstanceProvider::getRepository()->getLabel()->findMeetingLabelsForSchedule($projectId);
	}

	public function persistIssueLabel($issueLabel, $description, $project){
		$connection = $this->entityManager->getConnection();
		try {
			$connection->beginTransaction();
			if(empty($project)){
				$resource = $this->environment->getCurrentUser();
			}
			else{
				$project = $this->entityManager->find('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Project', $project);
				$resource = null;
			}
            $labelFactory = InstanceProvider::getFactory()->getLabel();
			if(empty($issueLabel)){
				$issueLabel = $labelFactory->createLabel($description, $project, $resource);
			}
			else{
				$issueLabel = $labelFactory
					->updateLabel(	$this->entityManager->find('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Label', $issueLabel), $description,
						$project, $resource);
			}
			$this->entityManager->persist($issueLabel);
			$this->entityManager->flush();
			$connection->commit();
			return array("ID" => $issueLabel->getId());
		} catch (\Exception $e) {
			$connection->rollBack();
			throw $e;
		}
	}

	public function removeIssueLabel($issueLabel){
		$connection = $this->entityManager->getConnection();
		try {
			$connection->beginTransaction();
			$taskLabels = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskLabel')
				->findBy(array('label' => $issueLabel));
			foreach ($taskLabels as $taskLabel) {
				$this->entityManager->remove($taskLabel);
			}
			$this->entityManager->remove($this->entityManager->find('\Zeedhi\DevOps\Model\Projectmanagement\Entities\Label', $issueLabel));
			$this->entityManager->flush();
			$connection->commit();
		} catch (\Exception $e) {
			$connection->rollBack();
			throw $e;
		}
	}


	/************************************************************
	 * Gets all issues related to a specified task.
	 * @param int	$taskId
	 ************************************************************/
	public function findIssuesRelatedToTask($taskId, $conditions){
		$issues = InstanceProvider::getRepository()->getTask()->findIssuesRelatedToTask($taskId, $this->environment->getCurrentOrganizationId(), $conditions);
		return $issues;
	}


	/************************************************************
	 * Gets all issues related to a specified task.
	 * @param int	$issueId
	 ************************************************************/
	public function findCommentsRelatedToIssue($issueId){
		$comments = InstanceProvider::getRepository()->getTask()->findCommentsRelatedToIssue($issueId, $this->environment->getCurrentOrganizationId());
		foreach($comments as &$comment){
			list($year, $month, $day) = $this->convertDateForComments($comment['CREATED_AT_FORMATTED']);
			$comment['AUTHOR'] = $this->createCommentAuthorTextLine($year, $month, $day, $comment['RESOURCE_NAME']);
		}
		return $comments;
	}

	/************************************************************
	 * Creates the author text line for show in the comment.
	 * @param int		$day
	 * @param int 		$month
	 * @param int 		$year
	 * @param string	$authorName
	 ************************************************************/
	protected function createCommentAuthorTextLine($year, $month, $day, $authorName){
		$month = $this->environment->getMonthNameByMonthNumber($month, 'none');
		return "Criado no dia {$day} de {$month} de {$year} por {$authorName}";
	}


	/************************************************************
	 * Crops a date string into three parts (day, month and year).
	 * @param string	$dateString
	 ************************************************************/
	protected function convertDateForComments($dateString){
		$dateTime = \DateTime::createFromFormat('Y-m-d H:i:s', $dateString);
		$year = $dateTime->format('Y');
		$month = $dateTime->format('m');
		$day = $dateTime->format('d');
		return array($year, $month, $day);
	}

    /**
     * @param $task
     * @return mixed
     */
    public function getStartedAllocationsByTask($task, $resource) {
        return InstanceProvider::getRepository()->getTask()->getStartedAllocationsByTaskId($task, $resource);
    }

    /**
     * @param $task
     * @return array
     */
    public function findTasksGeneratedNotConcludedByTask($task) {
        return InstanceProvider::getRepository()->getTask()->getTasksGeneratedNotConcluded($task);
    }

    public function getConflictTypes() {
        return array(
            "DD", //"both deleted",
            "AU", //"added by us",
            "UD", //"deleted by them",
            "UA", //"added by them",
            "DU", //"deleted by us",
            "AA", //"both added",
            "UU" //"both modified"
        );
    }

    public function getProjectTasksInitialized($projectId){
        $user      = $this->environment->getCurrentUserId();
        $projectId = preg_replace("/[^0-9]/", "", $projectId);
        return InstanceProvider::getRepository()->getTask()->getProjectTasksInitialized($user,$projectId);
    }

    /**
     * @param $day
     * @return int
     */
    protected function setDefaultCalendar($day)
    {
        $weekNumber = \DateTime::createFromFormat(Utilities::FORMAT_DATE_DEFAULT_ONLY_DATE,$day)->format('w');
        if( in_array($weekNumber,array(0,6))){
            $dailyHours = 0;
        }else{
            $dailyHours = 28800;
        }

        return $dailyHours;
    }

    public function sendsMailIrregularities($emails, $message){
        if(!empty($emails)){
            foreach($emails as $email){
                InstanceProvider::getMailIssueMail()->sendIrregularitiesMail($email, $message);
            }
        }
    }

    public function calculationBasedStart($taskId,$begin,$end,$duration,$precursor,$precursorType,$parent,$exception,
                                          $exceptionDate,$projectId,$resourceIds,$latency = 0){


        if($parent){
            $parent = InstanceProvider::getRepository()->getTask()->find($parent);
        }
        $project = InstanceProvider::getRepository()->getProject()->find($projectId);
        if($exceptionDate){
            $exceptionDateTask = \DateTime::createFromFormat(\Util\Utilities::FORMAT_DATE_DEFAULT,$exceptionDate.' 00:00:00');
        }else{
            $exceptionDateTask = null;
        }
        $task = InstanceProvider::getFactory()->getTask()->createTask($project,0, null, $begin,$end, 'A', null, 0, 0, $parent, null,
            $exception, $exceptionDateTask);

        if($precursor){
            $precursor = InstanceProvider::getRepository()->getTask()->find($precursor);
            $taskRelationship = InstanceProvider::getFactory()->getTaskRelationship()->createTaskRelationship($task,$precursor,TaskPrecursor::PRECURSOR_TYPE,
                $precursorType);
            $task->addTaskRelationship($taskRelationship);
        }
        if($resourceIds){
            $this->setAllocations($task,$resourceIds);
        }
        return InstanceProvider::getService()->getCalendarTakCalendar()->calculationBasedStart($task,$begin,$duration,$precursor,$precursorType,$parent,
            $exception,$exceptionDate,$latency);
    }

    public function setAllocations(Entities\Task $task, array $resourceIds){
        $allocationService = InstanceProvider::getService()->getAllocation();
        $resourceRepository = InstanceProvider::getRepository()->getResource();
        foreach ($resourceIds as $resourceId){
            /** @var Entities\Resource $resource */
            $resource = $resourceRepository->find($resourceId);
            if($resource){
                $allocation = $allocationService->allocationFactory->createAllocation($task,$resource,null);
                $task->addAllocation($allocation);
            }
        }
    }

    public function calculationBasedEnd($taskId,$begin,$end,$duration,$precursor,$precursorType,$parent,$exception,
                                        $exceptionDate,$projectId,$resourceIds,$latency){
        if($parent){
            $parent = InstanceProvider::getRepository()->getTask()->find($parent);
        }
        $project = InstanceProvider::getRepository()->getProject()->find($projectId);
        if($exceptionDate){
            $exceptionDateTask = \DateTime::createFromFormat(\Util\Utilities::FORMAT_DATE_DEFAULT,$exceptionDate.' 00:00:00');
        }else{
            $exceptionDateTask = null;
        }
        $task = InstanceProvider::getFactory()->getTask()->createTask($project,0, null, $begin,$end, 'A', null, 0, 0, $parent, null,
            $exception, $exceptionDateTask);

        if($precursor){
            $precursor = InstanceProvider::getRepository()->getTask()->find($precursor);
            $taskRelationship = InstanceProvider::getFactory()->getTaskRelationship()->createTaskRelationship($task,$precursor,TaskPrecursor::PRECURSOR_TYPE,
                $precursorType);
            $task->addTaskRelationship($taskRelationship);
        }
        if($resourceIds){
            $this->setAllocations($task,$resourceIds);
        }
        return InstanceProvider::getService()->getCalendarTakCalendar()->calculationBasedEnd($task,$end,$duration,$precursor,$precursorType,$parent,
            $exception,$exceptionDate,$latency);
    }

    public function getIssueDescription($issueId){
        $result = array(array());
        $result[0]['DESCRIPTION'] = $this->entityManager->find(Utilities::ENTITY_DIR.'Task', $issueId)->getDescription();
        return $result;
    }

    public function createFastTask($name, $projectId, $taskClassification, $parentId, $description,$resource = array(),
                                   $labelIds = array(),$commitmentDate = null, $type = null, $notDuration = false,\DateTime $commitmentDateTask = null,  $taskStatusId = null, $priority = null)
    {
        $totalRating = 0;
        /** @var Entities\Project $project */
        $project = InstanceProvider::getRepository()->getProject()->find($projectId);
        if ($parentId) {
            /** @var Entities\Task $parent */
            $parent = InstanceProvider::getRepository()->getTask()->find($parentId);
            $estimatedBeginDate = $estimatedEndDate = $parent->getEstimatedBeginDate();
        } else {
            $parent = null;
            $estimatedBeginDate = $estimatedEndDate = $project->getEstimatedBeginDate();
        }
        $taskStatus = Utilities::getTaskStatusNotInitialized();
        $dateExceptionType = Calendar\Restriction\Strategy\RestrictionFactory::AS_SOON_AS_POSSIBLE;
        if($notDuration){
            $duration = 0;
        }else{
            $duration = 28800;
        }
        $latency = 0;
        $percentCompleted = 0;
        if($taskClassification){
            $taskClassification = InstanceProvider::getRepository()->getClassification()->find($taskClassification);
        }else{
        	if($type != "P"){
            	$taskClassification = InstanceProvider::getRepository()->getClassification()->findOneBy(array());
        	}
        }

        $task = $this->createTask($project, $latency, $name, $estimatedBeginDate, $estimatedEndDate, Entities\Task::ACTIVE, $description,
            $percentCompleted, $duration, $parent, $taskStatus, $dateExceptionType, null, Entities\Task::NOT_IS_PARENT,
            null, $taskClassification);
        if(isset($commitmentDateTask)){
            $task->setCommitmentDate($commitmentDateTask);
        }
        if(isset($taskStatusId)){
            $taskStatus = $this->getTaskStatusById($taskStatusId);
            $task->setTaskStatus($taskStatus);
        }
        $this->entityManager->flush();
        $allocationService = InstanceProvider::getService()->getAllocation();
        $resourceRepository = InstanceProvider::getRepository()->getResource();
        foreach ($resource as $eachResource) {
            $resourceProject = $resourceRepository->find($eachResource);
            $allocationService->createAllocation($task, $resourceProject,null,$taskStatus);
        }

//        /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Resource[] $resourcesObjects  */
//        $resourcesObjects = $resourceRepository->findBy(array('id' => $resource));
//        if(!$resourcesObjects){
//            $totalRating = 1;
//        }
//        foreach($resourcesObjects as $resourceObjects){
//            if($resourceObjects->getRating()){
//                $totalRating += $resourceObjects->getRating();
//            }else{
//                $totalRating += 1;
//            }
//        }
//        $task->setTimePlanned($task->getDuration()*$totalRating);
        $task->setTimePlanned(0);

        /** @var  Entities\Label[] $labels */
        $labels = InstanceProvider::getRepository()->getLabel()->findBy(array('id' => $labelIds ));
        $taskLabelFactory = InstanceProvider::getFactory()->getTaskLabel();
        foreach ($labels as $label) {
            if(!empty($label)){
                $taskLabel = $taskLabelFactory->createTaskLabel($task, $label);
                $this->entityManager->persist($taskLabel);
            }
        }
        $this->entityManager->flush();
        return $task;
    }

    public function relateLabelToIssue($issue, $labelIds){
        foreach($labelIds as $key => $labelId){
            $labelIds[$key] = $labelId["ID"];
        }

        $labels = InstanceProvider::getRepository()->getLabel()->findBy(array('id' => $labelIds ));
        $taskLabelFactory = InstanceProvider::getFactory()->getTaskLabel();
        $notAllowedLabels = InstanceProvider::getRepository()->getTaskLabel()->findExistentLabels($issue->getId());
        foreach ($labels as $label) {
            if(!in_array($label->getId(), $notAllowedLabels)){
                $taskLabel = $taskLabelFactory->createTaskLabel($issue, $label);
                $this->entityManager->persist($taskLabel);
            }
        }
        $this->entityManager->flush();
    }
    public function copySchedule(Entities\Task $targetTask, Entities\Project $targetProject, Entities\Task $parent = null)
    {

        $estimatedBeginDate = $estimatedEndDate = $targetProject->getEstimatedBeginDate();

        $task = $this->createTask($targetProject, $targetTask->getLatency(), $targetTask->getName(), $estimatedBeginDate, $estimatedEndDate,
            $targetTask->getStatus(), $targetTask->getDescription(), $targetTask->getPercentCompleted(), $targetTask->getDuration(), $parent,
            $targetTask->getTaskStatus(), $targetTask->getDateExceptionType(), null, Entities\Task::NOT_IS_PARENT,null, $targetTask->getClassification());
        $task->setCommitmentDate($targetTask->getCommitmentDate());
        $task->setTaskOrder($targetTask->getTaskOrder());
        $this->entityManager->flush();
        /** @var Entities\Allocation[] $allocations */
        $allocations = InstanceProvider::getRepository()->getAllocation()->findBy(array('task' => $targetTask));
        $allocationService = InstanceProvider::getService()->getAllocation();
        foreach ($allocations as $allocation) {
            $allocationService->createAllocation($task, $allocation->getResource(),null,$targetTask->getTaskStatus(),true);
        }
        $labelsRelated = $this->getLabelIdByTask($targetTask);
        $this->setLabelInSchedule($task,$labelsRelated);
        InstanceProvider::getService()->getTaskPrecursor()->createTaskPrecursor($task,$targetTask,Entities\TaskRelationship::SCHEDULE_TYPE,null,false);
        $this->entityManager->flush();
        return $task;
    }

    public function getLabelIdByTask ($task){
        $taskLabels = InstanceProvider::getRepository()->getTaskLabel()->findBy(array("task" => $task));
        return Utilities::getArrayByArrayObjectWithGet($taskLabels,['getLabel','getId']);
    }

    public function editSchedule(Entities\Task $task, Entities\Project $project,Entities\Classification $classification = null,
                                 Entities\Task $parent = null, $description = null, $name = null, $resources = array(),
                                 $taskStatus = null,$commitmentDate = null,$labelsSchedule = array(), $priority = null)
    {
        if ($parent) {
            $estimatedBeginDate = $estimatedEndDate = $parent->getEstimatedBeginDate();
        } else {
            $estimatedBeginDate = $estimatedEndDate = $project->getEstimatedBeginDate();
        }
        if(!$taskStatus){
            $taskStatus = Utilities::getTaskStatusNotInitialized();
        }
        $dateExceptionType = Calendar\Restriction\Strategy\RestrictionFactory::AS_SOON_AS_POSSIBLE;
        $duration = 28800;
        $latency = 0;
        $percentCompleted = 0;
        $task->setCommitmentDate($commitmentDate);
        $task = $this->editTask($task,$project, $latency, $name, $estimatedBeginDate, $estimatedEndDate, Entities\Task::ACTIVE, $description,
            $percentCompleted, $duration, $parent, $taskStatus, $dateExceptionType, null, Entities\Task::NOT_IS_PARENT,
            'N', $classification, $priority);
        $this->entityManager->flush();
        $allocationService = InstanceProvider::getService()->getAllocation();
        $resourceRepository = InstanceProvider::getRepository()->getResource();
        $allocationRepository = InstanceProvider::getRepository()->getAllocation();
        foreach ($resources as $eachResource) {
            if(!empty($eachResource)){
                $resourceProject = $resourceRepository->find($eachResource);
                /** @var Entities\Allocation $allocation */
                $allocation = $allocationRepository->findOneBy(array("task" => $task->getId(),
                    "resource" => $resourceProject->getId()));
                if (empty($allocation)) {
                    $allocationService->createAllocation($task, $resourceProject,null,null);
                }else{
                    $allocationService->updateAllocation($allocation,$task, $resourceProject);
                }
            }
        }
        /** @var Entities\allocation[] $allocation */
        $allocation = $allocationRepository->findBy(array("task" => $task->getId()));
        foreach ($allocation as $eachAllocation) {
            if(!in_array($eachAllocation->getResource()->getId(), $resources)){
                $this->entityManager->remove($eachAllocation);
                $this->entityManager->flush();
            }
        }
        $this->setLabelInSchedule($task,$labelsSchedule);
        $this->entityManager->flush();
        return $task;
    }

    public function createFastIssue($name, $productId, $sponsorId, $followers, $issueClassification, $reporterId,
                                    $issueDescription,$labelIds = array(),$commitmentDate = null, $priority = null)
    {
        /** @var Entities\Project $project */
        $product = InstanceProvider::getRepository()->getProject()->find($productId);
        $taskStatus = Utilities::getTaskStatusBacklog();
        $dateExceptionType = Calendar\Restriction\Strategy\RestrictionFactory::AS_SOON_AS_POSSIBLE;
        $duration = 28800;
        $latency = 0;
        $percentCompleted = 0;

        $issueClassification = InstanceProvider::getRepository()->getClassification()->find($issueClassification);
        $issue = $this->createIssue($product, $latency, $name, $this->environment->getCurrentTime(),
            $this->environment->getCurrentTime(), Entities\Task::ACTIVE, $issueDescription,$percentCompleted,
            $duration, null, $taskStatus, $dateExceptionType, null, Entities\Task::IS_PARENT,
            null, $issueClassification, $priority);

        $issue->setCommitmentDate($commitmentDate);
        $this->setFollowers($followers, $issue);
        $allocationService = InstanceProvider::getService()->getAllocation();
        $resourceRepository = InstanceProvider::getRepository()->getResource();
        if(!empty($reporterId)){
            $reporter = $resourceRepository->find($reporterId);
            $allocationService->createAllocation($issue, $reporter, Allocation::REPORTER);
        }

        if(!empty($sponsorId)){
            $sponsor = $resourceRepository->find($sponsorId);
            $allocationService->createAllocation($issue, $sponsor, Allocation::SPONSOR);
        }

        /** @var  Entities\Label[] $labels */
        $labels = InstanceProvider::getRepository()->getLabel()->findBy(array('id' => $labelIds ));
        $taskLabelFactory = InstanceProvider::getFactory()->getTaskLabel();
        foreach ($labels as $label) {
            if(!empty($label)){
                $taskLabel = $taskLabelFactory->createTaskLabel($issue, $label);
                $this->entityManager->persist($taskLabel);
            }
        }
        $this->entityManager->flush();
        $fistComments = InstanceProvider::getFactory()->getComments()->createComments($issue,self::FIRST_COMMENTS);
        $issue->setCurrentTaskStatusIn(Utilities::getEnvironment()->getCurrentTime());
        $this->entityManager->persist($issue);
        $this->entityManager->persist($fistComments);
        $this->viewCommentsSaveData($issue->getId());
        $this->entityManager->flush();
        return $issue;
    }

    /**
     * @param $scheduleProjectId
     * @return Entities\Task
     */
    public function getPrimarySchedule($scheduleProjectId){
        return InstanceProvider::getRepository()->getTask()->findOneBy(array('project' => $scheduleProjectId,'parent' => null));
    }

    public function createFastSchedule($name, $scheduleProjectId, $scheduleParentId, $scheduleClassification,
                                       $scheduleDescription, $scheduleResource, $task = null, $type = null, \DateTime $commitmentDateTask = null, $taskStatusId = null)
    {
        $schedule = $this->createFastTask($name, $scheduleProjectId, $scheduleClassification, $scheduleParentId,
            $scheduleDescription,$scheduleResource, array(), null, $type, null, $commitmentDateTask, $taskStatusId);

        /** @var Entities\Project $project */
        $project = InstanceProvider::getRepository()->getProject()->find($scheduleProjectId);
        $parent = $project->getParent();
        if($parent){
            if(!$task){
                $taskParentId = $this->getParentIdForTaskTop($schedule);
                $task = $this->createFastTask($name, $parent->getId(), $scheduleClassification,
                    $taskParentId, $scheduleDescription,$scheduleResource);
            }
            InstanceProvider::getService()->getTaskPrecursor()->createTaskPrecursor($schedule,$task,Entities\TaskRelationship::SCHEDULE_TYPE);
        }
        return $schedule;
    }

    public function createSchedule($name,$scheduleProjectId,$scheduleParentId,$scheduleClassification,
                                       $scheduleDescription,$scheduleResource,$task = null)
    {

        if(!$task && !$scheduleParentId){
            $primarySchedule = $this->getPrimarySchedule($scheduleProjectId);
            $scheduleParentId = $primarySchedule ? $primarySchedule->getId() : null;
        }

        $schedule = $this->createFastSchedule($name,$scheduleProjectId,$scheduleParentId,$scheduleClassification,
            $scheduleDescription,$scheduleResource);
        return $schedule;
    }

    public function createScheduleInAllMeetingOccurrences($name, $scheduleProjectId, $scheduleParentId, $scheduleClassification,
                                       $scheduleDescription,$scheduleResource ,$labelsSchedule, $type,  $commitmentDateTask = null, $taskStatusId = null)
    {
        /** @var Entities\Project $project */
        $project = InstanceProvider::getRepository()->getProject()->find($scheduleProjectId);
        if($project->getParent()){
            $scheduleProjectId = $project->getParent()->getId();
        }
        if(!$scheduleParentId){
            $primarySchedule = $this->getPrimarySchedule($scheduleProjectId);
            $parentId = $primarySchedule ? $primarySchedule->getId() : null;
        }else{
            /** @var Entities\Task $scheduleParent */
            $scheduleParent = InstanceProvider::getRepository()->getTask()->find($scheduleParentId);
            if($scheduleParent->getProject()->getId() != $scheduleProjectId){
                $parentId = $this->getIdForTaskTop($scheduleParent);
            }else{
                $parentId = $scheduleParentId;
            }
        }

        $scheduleInMeeting = $this->createFastSchedule($name, $scheduleProjectId, $parentId, $scheduleClassification,
            $scheduleDescription, $scheduleResource, null, $type,  $commitmentDateTask, $taskStatusId);

        $this->setLabelInSchedule($scheduleInMeeting,$labelsSchedule);

        $this->createScheduleMeetingOccurrences($name, $scheduleProjectId, $scheduleParentId, $scheduleClassification,
            $scheduleDescription, $scheduleResource, $scheduleInMeeting, $type);


        return $scheduleInMeeting;
    }

    public function createPrimarySchedule($name,$scheduleProjectId,$scheduleParentId,$scheduleClassification,
                                          $scheduleDescription)
    {
        $schedule = $this->createFastTask($name, $scheduleProjectId, $scheduleClassification, $scheduleParentId,
            $scheduleDescription);

        return $schedule;
    }

    /**
     * @param Entities\Task[] $tasks
     * @return Entities\Task[]
     */
    public function getTaskForProjectIsProgressOrPlanning(array $tasks)
    {
        $taskStatus  = array(Entities\TaskStatus::TASK_STATUS_CONTROL_PLANNING,
            Entities\TaskStatus::TASK_STATUS_CONTROL_PROGRESS );
        $taskIds = array();
        foreach ($tasks as $task) {
            if (in_array($task->getProject()->getProjectStatus()->getControl(), $taskStatus) ) {
                $taskIds[] = $task;
            }
        }
        return $taskIds;
    }

    public function replicationRemoveRelationshipScheduleWithTask($schedule,$task){
        $scheduleTop = $this->getScheduleTop($schedule);
        $scheduleChildren = $scheduleTop->getTaskInRelationshipInverseTypeSchedule();
        $scheduleChildren = $this->getTaskForProjectIsProgressOrPlanning($scheduleChildren);
        $scheduleIds = Utilities::getArrayByArrayObjectWithGet($scheduleChildren,array('getId'));
        $scheduleIds[] = $scheduleTop->getId();
        $taskRelationshipRepository = InstanceProvider::getRepository()->getTaskRelationship();

        foreach ($scheduleIds as $scheduleId) {
            $taskRelationship = $taskRelationshipRepository->findOneBy(array('task' => $task,
                'taskRelationship' =>$scheduleId,
                'type' => Entities\TaskRelationship::TASK_WITH_SCHEDULE_TYPE));
            $this->entityManager->remove($taskRelationship);
            $this->entityManager->flush();
        }
    }

    public function replicationSetRelationshipScheduleWithTask($task, $schedule = null, $meetingId = null,
                                                               $taskOrder = null){
        if($schedule){
            $scheduleTop = $this->getScheduleTop($schedule);
            $scheduleChildren = $scheduleTop->getTaskInRelationshipInverseTypeSchedule();
            $scheduleChildren = $this->getTaskForProjectIsProgressOrPlanning($scheduleChildren);
            $taskRelationship = $this->setRelationshipScheduleWithTask($task, $scheduleTop, $scheduleTop->getProject()->getId());
            $taskRelationship->setTaskOrder($taskOrder);
            $this->entityManager->persist($taskRelationship);
            foreach ($scheduleChildren as $scheduleChild) {
                $taskRelationship = $this->setRelationshipScheduleWithTask($task, $scheduleChild,
                    $scheduleChild->getProject()->getId());
                if($taskOrder){
                    $taskRelationship->setTaskOrder($taskOrder);
                    $this->entityManager->persist($taskRelationship);

                }
            }
            $this->entityManager->flush();
        }else{
            $meeting = InstanceProvider::getRepository()->getProject()->find($meetingId);
            $schedule = $this->getPrimarySchedule($meeting);
            $scheduleTop = $this->getScheduleTop($schedule);
            $scheduleChildren = $scheduleTop->getTaskInRelationshipInverseTypeSchedule();
            $scheduleChildren = $this->getTaskForProjectIsProgressOrPlanning($scheduleChildren);
            $this->setRelationshipScheduleWithTask($task, null, $scheduleTop->getProject()->getId());
            foreach ($scheduleChildren as $scheduleChild) {
                $this->setRelationshipScheduleWithTask($task, null, $scheduleChild->getProject()->getId());
            }
        }
    }

    /**
     * @param $meetingId
     * @param Entities\Task $task
     * @return Entities\TaskRelationship
     */
    public function getTaskRelationshipByMeeting($meetingId, $task, $scheduleRelationship = null )
    {
        /** @var Entities\TaskRelationship[] $taskRelationships */
        $taskRelationships = InstanceProvider::getRepository()->getTaskRelationship()->findBy(array('task' => $task,
            'taskRelationship' =>$scheduleRelationship,
            'type' => Entities\TaskRelationship::TASK_WITH_SCHEDULE_TYPE));

        foreach ($taskRelationships as $taskRelationship) {
            if ($taskRelationship->getTaskRelationship()->getProject()->getId() == $meetingId) {
                return  $scheduleRelationship = $taskRelationship;
            }
        }
    }

    public function setRelationshipScheduleWithTask(Entities\Task $task,Entities\Task $scheduleRelationship = null,
                                                    $meetingId = null,$taskOrder = null)
    {

        $taskRelationship = $this->getTaskRelationshipByMeeting($meetingId, $task, $scheduleRelationship);

        if ($scheduleRelationship) {
            if (!$taskRelationship) {
                $taskRelationship = $this->createTaskRelationshipMeeting($task, $scheduleRelationship,
                    Entities\TaskRelationship::TASK_WITH_SCHEDULE_TYPE);
            }
            $taskRelationship->setTaskRelationship($scheduleRelationship);
            $this->entityManager->persist($taskRelationship);
            $this->entityManager->flush();
        }else if($taskRelationship){
            $this->entityManager->remove($taskRelationship);
            $this->entityManager->flush();
        }
        return $taskRelationship;
    }
    public function setInProgressMeetingWhenBelatedByScheduleId($scheduleId){
        $reload = false;
        if($scheduleId){
            /** @var Entities\Task $schedule */
            $schedule =  InstanceProvider::getRepository()->getTask()->find($scheduleId);
            $reload = $this->setInProgressMeetingWhenBelated($schedule->getProject());
        }
        return $reload;
    }

    public function setInProgressMeetingWhenBelated(Entities\Project $meeting){
        $reload = false;
        $today = new \DateTime();
        $controls = array(Entities\TaskStatus::TASK_STATUS_CONTROL_CONCLUDED,
            Entities\TaskStatus::TASK_STATUS_CONTROL_CANCELED,
            Entities\TaskStatus::TASK_STATUS_CONTROL_PROGRESS);
        if($meeting->getBeginDate() < $today && !in_array($meeting->getProjectStatus()->getControl(), $controls) ){
            $projectStatus = Utilities::getProjectStatusProgress();
            $meeting->setProjectStatus($projectStatus);
            $this->entityManager->persist($meeting);
            $this->entityManager->flush();
            $reload = true;
        }
        return $reload;

    }

    public function createFastTaskAndIssue($type, $name, $projectId, $taskClassification, $resource, $parentId,
                                           $taskDescription, $productId, $sponsorId, $followers, $issueClassification,
                                           $reporter, $issueDescription, $scheduleProjectId, $scheduleParentId,
                                           $scheduleClassification, $scheduleDescription,$scheduleResource, $labelsTask,
                                           $labelsIssue, $scheduleRelationshipIssueIds, $scheduleRelationshipTaskIds,
                                           $commitmentDateIssue, $commitmentDateTask, $orderIssue, $orderTask,
                                           $labelsSchedule, $taskStatusId = null, $priority = null)
    {
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            $reload = false;
            $taskRepository = InstanceProvider::getRepository()->getTask();
            if($priority){
                $priority = $this->entityManager->find(Utilities::ENTITY_DIR . 'Priority', $priority);
            }
            if ($type == 'T') {
                $task = $this->createFastTask($name, $projectId, $taskClassification, $parentId, $taskDescription,
                    $resource, $labelsTask, $commitmentDateTask, $type);
                foreach($scheduleRelationshipTaskIds as $scheduleRelationshipTaskId) {
                    /** @var Entities\Task $scheduleRelationshipTask */
                    $scheduleRelationshipTask = $taskRepository->find($scheduleRelationshipTaskId);
                    $this->replicationSetRelationshipScheduleWithTask($task, $scheduleRelationshipTask, null, $orderTask);
                    $reload = $this->setInProgressMeetingWhenBelated($scheduleRelationshipTask->getProject());
                }
            } else if ($type == 'I') {
                $issue = $this->createFastIssue($name, $productId, $sponsorId, $followers, $issueClassification, $reporter,
                    $issueDescription, $labelsIssue, $commitmentDateIssue, $priority);
                foreach($scheduleRelationshipIssueIds as $scheduleRelationshipIssueId) {
                    /** @var Entities\Task $scheduleRelationshipIssue */
                    $scheduleRelationshipIssue = $taskRepository->find($scheduleRelationshipIssueId);
                    $this->replicationSetRelationshipScheduleWithTask($issue, $scheduleRelationshipIssue, null, $orderIssue);
                    $reload = $this->setInProgressMeetingWhenBelated($scheduleRelationshipIssue->getProject());
                }
                $this->sendOpenIssueMail($issue);
            } else {
                $this->createScheduleInAllMeetingOccurrences($name, $scheduleProjectId, $scheduleParentId, $scheduleClassification,
                    $scheduleDescription, $scheduleResource,$labelsSchedule, $type, $commitmentDateTask, $taskStatusId);
                /** @var Entities\Project $scheduleProject */
                $scheduleProject = InstanceProvider::getRepository()->getProject()->find($scheduleProjectId);
                $reload = $this->setInProgressMeetingWhenBelated($scheduleProject);
            }
            $connection->commit();
            return $reload;
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }

    }

    /**
     * @param $issueIds
     * @return Entities\Task
     * @throws \Exception
     */
    public function reopenIssue($issueIds){
        $user = $this->environment->getCurrentUser();
        /**@var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskType $backlogTaskType*/
        $backlogTaskType = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskType')->findOneBy(array('control' => 'B'));
        /**@var \Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskStatus $backlogTaskStatus*/
        $backlogTaskStatus = $this->entityManager->getRepository('\Zeedhi\DevOps\Model\Projectmanagement\Entities\TaskStatus')->findOneBy(array('taskType' => $backlogTaskType->getId()));

        $taskRepository = InstanceProvider::getRepository()->getTask();
        $allocationRepository = InstanceProvider::getRepository()->getAllocation();
        foreach($issueIds AS $issueId){
            /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Task $issue*/
            $issue = $taskRepository->find($issueId);
            /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Allocation $reporter */
            $reporter = $allocationRepository->findOneBy(array('type'=>\Service\Allocation::REPORTER,'resource' => $user,'task'=>$issue));
            if(empty($reporter)){
                throw new \Exception('Essa issue só pode ser recusada pelo solicitante!');
            }
            $issue->setPercentCompleted(0);
            $issue->setTaskStatus($backlogTaskStatus);
            $issue->setCurrentTaskStatusIn((new \DateTime()));
            /** @var Entities\TaskRelationship[] $taskRelationships */
            $taskRelationships = InstanceProvider::getRepository()->getTaskRelationship()->findBy(array('task' => $issue, 'type' => TaskPrecursor::RELATIONSHIP_ISSUE_TYPE));
            foreach ($taskRelationships as $taskRelationship) {
                $task = $taskRelationship->getTaskRelationship();
                $task->setUsedInFraming(Entities\Task::NOT_USED_IN_FRAMING);
                $this->entityManager->persist($task);
            }
            $this->entityManager->persist($issue);
            $this->sendMailOnReopenIssue($issueId);
        }
        $this->entityManager->flush();
    }

    /***********************************************************************
     * Gets the specified issue logs.
     * @param mixed $issueId
     **********************************************************************
     * @return array
     */
    public function findIssueLogByIssueId($issueId, $filter = ' ')
    {
        $organizationId = $this->environment->getCurrentOrganizationId();
        return InstanceProvider::getRepository()->getTask()->findIssueLogByIssueId($issueId, $organizationId, $filter);
    }

    /**
     * @param $taskIds
     * @param $insertTaskInIssue
     * @throws \Doctrine\DBAL\ConnectionException
     * @throws \Exception
     */
    public function applyOrderToIssuesCreateTasks($taskIds, $insertTaskInIssue) {
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            foreach ($taskIds as $taskId) {
                /** @var Entities\Task $task */
                $task = $this->entityManager->getRepository(Utilities::ENTITY_DIR . 'Task')->find($taskId);
                $this->applyOrderToIssuesByTask($task, false, $insertTaskInIssue);
            }

            $this->entityManager->flush();
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }

    /**
     * @param $orderToView
     * @param Entities\TaskType $taskType
     * @param $control
     * @throws \Exception
     * @return bool
     */
    protected function uniqueValidationsTaskType($orderToView, $taskType, $control)
    {
        $return = false;
        $taskTypeRepository = InstanceProvider::getRepository()->getTaskType();
        $uniqueOrderToView = $taskTypeRepository->findOneBy(array(
            'organization' => $this->environment->getCurrentOrganizationId(),
            'orderToView' => $orderToView));

        $uniqueControl = $taskTypeRepository->findOneBy(array(
            'organization' => $this->environment->getCurrentOrganizationId(),
            'control' => $control));
       if ($uniqueOrderToView && $taskType->getId() != $uniqueOrderToView->getId()) {

            throw new \Exception(str_replace('%1',$uniqueOrderToView->getLabel(),Messages::NOT_POSSIBLE_TO_SAVE_TYPE_ORDER_TO_VIEW_ALREADY_USED));

        }else if ($control && $uniqueControl && $taskType->getId() != $uniqueControl->getId()) {

            throw new \Exception(str_replace('%1',$uniqueControl->getLabel(),Messages::NOT_POSSIBLE_TO_SAVE_TYPE_CONTROL_ALREADY_USED));

        } else {
            $return = true;
        }

        return $return;
    }

    /**
     * @param $taskId
     * @param $newTaskId
     * @param $typeRelationship
     * @throws \Doctrine\DBAL\ConnectionException
     * @throws \Exception
     */
    public function duplicateChangedFilesForNewTask($taskId, $newTaskId, $typeRelationship){
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            $taskRelationshipRepository = InstanceProvider::getRepository()->getTaskRelationship();
            $relatedTasks = $taskRelationshipRepository->findRelatedTaskIds($taskId);
            $newRelatedTasks = $taskRelationshipRepository->findRelatedTaskIds($newTaskId);
            if(!empty($relatedTasks)) {
                $taskId = $relatedTasks['MAIN_TASK'];
            }
            if(!empty($newRelatedTasks)) {
                $newTaskId = $newRelatedTasks['MAIN_TASK'];
            }

            $taskRepository = InstanceProvider::getRepository()->getTask();
            /** @var Entities\Task $task */
            $task = $taskRepository->find($taskId);
            if($typeRelationship == TaskPrecursor::MERGE_REQUEST_TYPE || $typeRelationship == TaskPrecursor::TEST_TYPE) {
                /** @var Entities\TaskRelationship $relationship */
                $relationship = $taskRelationshipRepository->findOneBy(array('taskRelationship' => $task,
                                                                                   'type' => array(
                                                                                       TaskPrecursor::MERGE_REQUEST_TYPE,
                                                                                       TaskPrecursor::TEST_TYPE,
                                                                                   )));
                $task = $relationship->getTask();
                $mrRelatedTasks = $taskRelationshipRepository->findRelatedTaskIds($task->getId());
                if(!empty($mrRelatedTasks)) {
                    $task = $taskRepository->find($mrRelatedTasks['MAIN_TASK']);
                }
            }
            /** @var Entities\Task $newTask */
            $newTask = $taskRepository->find($newTaskId);
            $changedFiles = InstanceProvider::getRepository()->getChangedFiles()->findBy(array('task' => $task));

            foreach($changedFiles as $file) {
                /** @var Entities\ChangedFiles $newChangeFile */
                $newChangeFile = clone $file;
                $newChangeFile->setTask($newTask);
                $this->entityManager->persist($newChangeFile);
            }

            $this->entityManager->flush();
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }


    public function associateTaskAndLabels($taskId, $labelIds, $type = null,$meetingProjectIds = null){
        /** @var Entities\Task $task */
        $task = InstanceProvider::getRepository()->getTask()->find($taskId);

        $labelRepository = InstanceProvider::getRepository()->getLabel();
        $labelsIsCurrentUser = $labelRepository->findBy(array('resource' => $this->environment->getCurrentUserId()));
        if($meetingProjectIds) {
            $meetingProjectIds = $this->getImInMeetingProjectIds();
            $labelsIsCurrentUser = array_merge($labelsIsCurrentUser, $labelRepository->findBy(array('project' => $meetingProjectIds)));
        }
        $taskLabelRepository = InstanceProvider::getRepository()->getTaskLabel();
        foreach ($labelIds as $labelId) {
            if(!empty($labelId)){
                /** @var  Entities\Label $label */
                $label = $labelRepository->find($labelId);
                $taskLabel = $taskLabelRepository->findBy(array("task" =>$taskId, "label" => $label->getId(),"type" => $type));

                $taskLabelFactory = InstanceProvider::getFactory()->getTaskLabel();
                if (empty($taskLabel)) {
                    $taskLabel = $taskLabelFactory->createTaskLabel($task, $label, $type);
                    $this->entityManager->persist($taskLabel);
                }
            }
        }
        $this->entityManager->flush();
        /** @var Entities\TaskLabel[] $taskLabels */
        $taskLabels = $taskLabelRepository->findBy(array('label'=>$labelsIsCurrentUser,"task" => $task->getId(),"type" => $type));
        foreach ($taskLabels as $taskLabel) {
            if(!in_array($taskLabel->getLabel()->getId(), $labelIds)){
                $this->entityManager->remove($taskLabel);
                $this->entityManager->flush();
            }
        }
    }

    public function saveTaskReponse($taskId, $taskResponse){
        $connection = $this->entityManager->getConnection();
        $connection->beginTransaction();
        try{
            /** @var Entities\Task $task */
            $task = InstanceProvider::getRepository()->getTask()->find($taskId);
            $task->setResponse($taskResponse);
            $this->entityManager->persist($task);
            $this->entityManager->flush();
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }


    }

    public function associateTaskLabelsMeeting($taskId, $labelIds, $type = null){
        /** @var Entities\Task $task */
        $task = InstanceProvider::getRepository()->getTask()->find($taskId);
        $taskLabelFactory = InstanceProvider::getFactory()->getTaskLabel();
        $labelRepository = InstanceProvider::getRepository()->getLabel();
        $taskLabelRepository = InstanceProvider::getRepository()->getTaskLabel();
        foreach ($labelIds as $labelId) {
            if(!empty($labelId)){
                /** @var  Entities\Label $label */
                $label = $labelRepository->find($labelId);
                $taskLabel = $taskLabelRepository->findBy(array("task" =>$taskId, "label" => $label->getId(),"type" => $type));
                if (empty($taskLabel)) {
                    $taskLabel = $taskLabelFactory->createTaskLabel($task, $label,$type);
                    $this->entityManager->persist($taskLabel);
                }
            }
        }
        $this->entityManager->flush();
        /** @var Entities\TaskLabel[] $taskLabels */
        $taskLabels = $taskLabelRepository->findBy(array("task" => $task->getId(),"type" => $type));
        foreach ($taskLabels as $taskLabel) {
            if(!in_array($taskLabel->getLabel()->getId(), $labelIds)){
                $this->entityManager->remove($taskLabel);
                $this->entityManager->flush();
            }
        }
    }

    public function findCommentsOnIssueExceptCommentedBy($issueId) {
        return InstanceProvider::getRepository()->getComments()->getCommentsOfTaskExceptCommentedBy($issueId);
    }

    /**
     * @param $followers
     * @param $issue
     */
    protected function setFollowers(array $followers, $issue)
    {
        $allocationService = InstanceProvider::getService()->getAllocation();
        $resourceRepository = InstanceProvider::getRepository()->getResource();
        $allocationRepository = InstanceProvider::getRepository()->getAllocation();
        foreach ($followers as $eachResource) {
            if (!empty($eachResource)) {
                $resourceProject = $resourceRepository->find($eachResource);
                $allocation = $allocationRepository->findBy(array("task" => $issue->getId(), "resource" => $resourceProject->getId(),
                    'type' => Allocation::FOLLOWER));
                if (empty($allocation)) {
                    $allocationService->createAllocation($issue, $resourceProject, Allocation::FOLLOWER);
                }
            }
        }
        $this->entityManager->flush();
    }

    /**
     * @param Entities\Task $schedule
     * @return null
     */
    protected function getParentIdForTaskTop(Entities\Task $schedule)
    {
        $scheduleParentId = null;
        if ($schedule->getParent()) {
            $TaskTops = $schedule->getParent()->getTaskInRelationshipTypeSchedule();
            if (!empty($TaskTops)) {
                $scheduleParentId = $TaskTops[0]->getId();
            }
        }
        return $scheduleParentId;
    }
    /**
     * @param Entities\Task $schedule
     * @return null
     */
    public function getIdForTaskTop(Entities\Task $schedule)
    {
        $scheduleParentId = $schedule->getId();
        $TaskTops = $schedule->getTaskInRelationshipTypeSchedule();
        if (!empty($TaskTops)) {
            $scheduleParentId = $TaskTops[0]->getId();
        }
        return $scheduleParentId;
    }

    /**
     * @param Entities\Task $schedule
     * @return Entities\Task
     */
    public function getScheduleTop(Entities\Task $schedule)
    {
        $scheduleTop = $schedule;
        $TaskTops = $schedule->getTaskInRelationshipTypeSchedule();
        if (!empty($TaskTops)) {
            $scheduleTop = $TaskTops[0];
        }
        return $scheduleTop;
    }

    protected function getParentIdForTaskInRelationshipTypeSchedule(Entities\Task $scheduleInMeeting,
                                                                    Entities\Project $projectMeetingOccurrence)
    {
        $scheduleParentId = null;
        if ($scheduleInMeeting->getParent()) {
            $taskDowns = $scheduleInMeeting->getParent()->getTaskInRelationshipInverseTypeSchedule();
            foreach ($taskDowns as $taskDown) {
                if($taskDown->getProject()->getId() == $projectMeetingOccurrence->getId()){
                    $scheduleParentId = $taskDown->getId();
                }
            }
        }
        return $scheduleParentId;
    }

    /**
     * @return array
     */
    public function getAllMeetingProjectIds()
    {
        return InstanceProvider::getService()->getIssue()->getAllMeetingProjectIds();
    }
    /**
     * @return array
     */
    public function getMyMeetingProjectIds()
    {
        return InstanceProvider::getService()->getIssue()->getMyMeetingProjectIds();
    }
    /**
     * @return array
     */
    public function getImInMeetingProjectIds()
    {
        return InstanceProvider::getService()->getIssue()->getImInMeetingProjectIds();
    }

    public function createTaskRelationshipMeeting($task, $taskRelationship, $type = Entities\TaskRelationship::TASK_WITH_SCHEDULE_TYPE)
    {
        return  InstanceProvider::getService()->getTaskPrecursor()->createTaskPrecursor($task, $taskRelationship,$type,null,false);
    }

    /**
     * @param $name
     * @param $scheduleProjectId
     * @param $scheduleParentId
     * @param $scheduleClassification
     * @param $scheduleDescription
     * @param $scheduleResource
     * @param $scheduleInMeeting
     */
    protected function createScheduleMeetingOccurrences($name, $scheduleProjectId, $scheduleParentId, $scheduleClassification, $scheduleDescription, $scheduleResource, $scheduleInMeeting, $type)
    {
        /** @var Entities\Project[] $projectMeetingOccurrences */
        $projectMeetingOccurrences = InstanceProvider::getRepository()->getProject()->findBy(array('parent' => $scheduleProjectId));
        $taskStatus  = array(Entities\TaskStatus::TASK_STATUS_CONTROL_PLANNING,
                             Entities\TaskStatus::TASK_STATUS_CONTROL_PROGRESS );
        foreach ($projectMeetingOccurrences as $projectMeetingOccurrence) {
            if(in_array($projectMeetingOccurrence->getProjectStatus()->getControl(), $taskStatus)){
                if(!$scheduleParentId){
                    $primarySchedule = $this->getPrimarySchedule($projectMeetingOccurrence->getId());
                    $taskParentId = $primarySchedule ? $primarySchedule->getId() : null;
                }else{
                    $taskParentId = $this->getParentIdForTaskInRelationshipTypeSchedule($scheduleInMeeting,$projectMeetingOccurrence);
                }
                $scheduleInMeetingOccurrences = $this->createFastTask($name, $projectMeetingOccurrence->getId(),  $scheduleClassification,
                    $taskParentId, $scheduleDescription, $scheduleResource, array(), null, $type);
                InstanceProvider::getService()->getTaskPrecursor()->createTaskPrecursor($scheduleInMeetingOccurrences, $scheduleInMeeting,
                    Entities\TaskRelationship::SCHEDULE_TYPE);
            }
        }
    }

    /**
     * @param Entities\TaskRelationship[] $relatedTasks
     * @param  Entities\Task $originTask
     * @return Entities\Task
     */
    private function getFirstTaskRelationship($relatedTasks, $originTask)
    {
        $task = null;
        foreach ($relatedTasks as $relatedTask) {
            if ($originTask != $relatedTask->getTaskRelationship()) {
                $task = $relatedTask;
                break;
            }
        }
        return $task;
    }

    /**
     * @param $taskStatusControl
     * @return null|object
     * @throws \Exception
     */
    private function getTaskStatusByControl($taskStatusControl)
    {
        $taskStatus = InstanceProvider::getRepository()->getTaskStatus()->findOneBy(array('control' => $taskStatusControl, 'isTask' => 'Y'));

        if (empty($taskStatus)) {
            throw new \Exception(Messages::ERR_NOT_STATUS_TASK_BY_CONTROL);
        }
        return $taskStatus;
    }

    /**
     * @param $taskId
     * @return array
     */
    private function getTaskIdsToUpdate($taskId)
    {
        $taskIds = array($taskId);
        $relatedTasks = InstanceProvider::getRepository()->getTaskRelationship()->findRelatedTaskIds($taskId);
        if (!empty($relatedTasks)) {
            $relatedTaskIds = $relatedTasks['MAIN_TASK'] . ',' . $relatedTasks['RELATED_TASKS'];
            $taskIds = explode(',', $relatedTaskIds);
        }
        return $taskIds;
    }

    public function getFirstTimePlanned(Entities\task $task){

        $timePlanned = $task->getTimePlanned();
        if(!$timePlanned){
            $timePlanned = 28800;
        }
        /** @var Entities\TaskRelationship $firstTaskRelationship */
        $firstTaskRelationship = InstanceProvider::getRepository()->getTaskRelationship()->findOneBy(array('task' => $task,
            'type' => Entities\TaskRelationship::FIRST_TYPE));
        if($firstTaskRelationship){
            $timePlanned = $firstTaskRelationship->getTaskRelationship()->getTimePlanned();
        }
        return $timePlanned;
    }

    /**
     * @param $tasks
     * @param $finishedTaskId
     * @param bool $isRelatedTask
     * @param array $document
     * @param bool $notSetTotalTime
     * @return array
     * @throws \Exception
     */
    public function saveMultipleProjectTask($tasks, $finishedTaskId, $isRelatedTask = true, $document = array(), $notSetTotalTime = false)
    {
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            if(is_array($finishedTaskId)) {
                $relatedTaskIds = $finishedTaskId;
            } else {
                $relatedTaskIds = InstanceProvider::getRepository()->getTaskRelationship()->findRelatedTaskIds($finishedTaskId);
                if (!empty($relatedTaskIds)) {
                    $relatedTaskIds = explode(',', $relatedTaskIds['RELATED_TASKS'] . ',' . $relatedTaskIds['MAIN_TASK']);
                } else {
                    $relatedTaskIds = array($finishedTaskId);
                }
            }
            $createdTaskIds = array();
            $taskPrecursorService = InstanceProvider::getService()->getTaskPrecursor();
            $projectRepository = InstanceProvider::getRepository()->getProject();
            foreach($tasks as $task) {
                $totalRating = 0;
                list($project, $duration, $name, $estimatedBeginDate, $estimatedEndDate,
                    $status, $description, $percentCompleted, $resources, $dateExceptionType, $dateException,
                    $taskStatus, $parent, $classification, $typeRelationship, $urlToTest, $errorClassificationId, $productVersionId
                ) = $this->getParamsSaveMultipleTasks($task);

                /** @var Entities\Project $project */
                $project = $projectRepository->find($project);
                /** @var Entities\Classification $classification */
                $classification = InstanceProvider::getRepository()->getClassification()->findOneBy(array('control'=> $typeRelationship));
                $taskStatus = DevOpsUtilities::getTaskStatusNotInitialized();
                /** @var \Zeedhi\DevOps\Model\Projectmanagement\Entities\Resource[] $resourcesObjects  */
                $resourcesObjects = InstanceProvider::getRepository()->getResource()->findBy(array('id' => $resources));
                if(!$resourcesObjects){
                    $totalRating = 1;
                }
                /** @var Repositories\ErrorClassification $errorClassificationRepository */
                $errorClassificationRepository = InstanceProvider::getRepository()->getErrorClassification();
                /** @var Entities\ErrorClassification $errorClassification */
                if($errorClassificationId){
                    $errorClassification = $errorClassificationRepository->find($errorClassificationId);
                } else {
                    $errorClassification = null;
                }
                foreach($resourcesObjects as $resourceObjects){
                    if($resourceObjects->getRating()){
                        $totalRating += $resourceObjects->getRating();
                    }else{
                        $totalRating += 1;
                    }
                }
                $taskCreated = $this->createNewTaskProject($project, 0, $name, $estimatedBeginDate, $estimatedEndDate,
                    $status, $description, $percentCompleted, $duration, $resources, null, $parent, null,
                    $taskStatus, $dateExceptionType, $dateException, 'N', $urlToTest, $classification, null, $errorClassification);

                $taskRepository = InstanceProvider::getRepository()->getTask();
                /** @var Entities\task $finishedTask */
                $finishedTask = $taskRepository->find($relatedTaskIds[0]);

                if($taskCreated->getClassification()->getControl() !== Entities\Classification::CONTROL_REPLICATION) {
                    /** @var Entities\TaskRelationship[] $tasksRelationship */
                    $tasksRelationship = InstanceProvider::getRepository()->getTaskRelationship()->findBy(
                        array('taskRelationship' => $finishedTask->getId(), 'type' => 'I')
                    );

                    foreach($tasksRelationship as $taskRelationship){
                        $issue = $taskRepository->find($taskRelationship->getTask());
                        $taskPrecursorService->createTaskPrecursor(
                            $issue, $taskCreated, TaskPrecursor::RELATIONSHIP_ISSUE_TYPE, null, false
                        );
                    }
                }

                if(!$notSetTotalTime) {

                    $finishedTaskTimePlanned = $this->getFirstTimePlanned($finishedTask);
                    $timeProportion = $classification->getEstimatedTimeProportion();
                    if(!$timeProportion){
                        $timeProportion = 100;
                    }
                    $estimatedTimeProportion = $timeProportion / 100;
                    $timePlanned = round($finishedTaskTimePlanned * $estimatedTimeProportion);
                    $taskCreated->setTimePlanned(1);
                    $taskCreated->setDuration($timePlanned / $totalRating);
                }else{
                    $taskCreated->setTimePlanned(1);
                }
                $this->setNotInFraming($taskCreated, $typeRelationship,$relatedTaskIds);
                $this->entityManager->persist($taskCreated);
                $this->entityManager->flush();
                $arrayControls = Project::arrayStatusChildrenReplicate();
                $arrayControls[] = Entities\TaskStatus::TASK_STATUS_CONTROL_PROGRESS;

                if(!in_array($project->getProjectStatus()->getControl(), $arrayControls)) {
                    $projectService = Utilities::getProjectService();
                    $projectService->setsDevOps($project);
                    $project->setProjectStatus(Utilities::getProjectStatusProgress());
                    $this->entityManager->persist($project);
                    $this->entityManager->flush();
                }
                if($productVersionId) {
                    /** @var EntitiesDevOps\ProductVersion $productVersion */
                    $productVersion = InstanceProvider::getRepository()->getProductVersion()->find($productVersionId);
                    $taskCreated->setProductVersionDev($productVersion);
                }
                $this->applyOrderToIssuesByTask($taskCreated);
                $this->createRelationship($taskCreated, $typeRelationship, $relatedTaskIds);
                $createdTaskIds[] = $taskCreated->getId();

            }
            if($isRelatedTask) {
                $this->createRelatedRelationship($createdTaskIds);
                $this->removeRelatedMergeRequestRelationship($createdTaskIds);
            }

            if(!empty($document)) {
                $attachmentService = InstanceProvider::getService()->getAttachment();
            	foreach($createdTaskIds as $createdTaskId){
            		$attachmentService->addDocumentsTask($createdTaskId, $document);
            	}
            }

            $this->entityManager->flush();
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
        return $createdTaskIds;
    }

    /**
     * @param $task
     * @return array
     */
    private function getParamsSaveMultipleTasks($task)
    {
        $project = $task['PROJECT'];
        $duration = $task['TOTAL_TIME'];
        $name = $task['NAME'];
        $estimatedBeginDate = $this->environment->createDateProject($task['START']);
        $estimatedEndDate = $this->environment->createDateProject($task['END']);
        $status = empty($task['STATUS']) ? 'A' : $task['STATUS'];
        $description = empty($task['DESCRIPTION']) ? null : $task['DESCRIPTION'];
        $percentCompleted = empty($task['PERCENT_COMPLETED']) ? 0 : (($task['PERCENT_COMPLETED'] < 100) ? $task['PERCENT_COMPLETED'] : 100);
        $resource = empty($task['RESOURCE']) ? array() : $task['RESOURCE'];
        $dateExceptionType = empty($task['DATE_EXCEPTION_TYPE']) ? 'MB' : $task['DATE_EXCEPTION_TYPE'];
        $dateException = empty($task['DATE_EXCEPTION']) ? null : \DateTime::createFromFormat('d/m/Y', $task['DATE_EXCEPTION']);
        $taskStatus = empty($task['TASK_STATUS']) ? null : $task['TASK_STATUS'];
        $parent = empty($task['PARENT']) ? null : $task['PARENT'];
        $classification = empty($task['TASK_CLASSIFICATION']) ? null : $task['TASK_CLASSIFICATION'];
        $typeRelationship = empty($task['TYPE_RELATIONSHIP']) ? null : $task['TYPE_RELATIONSHIP'];
        $urlToTest = empty($task['URL_TO_TEST']) ? null : $task['URL_TO_TEST'];
        $errorClassificationId = empty($task['ERROR_CLASSIFICATION_ID']) ? null : $task['ERROR_CLASSIFICATION_ID'];
        $productVersionId = empty($task['PRODUCT_VERSION_ID']) ? null : $task['PRODUCT_VERSION_ID'];

        return array($project, $duration, $name, $estimatedBeginDate, $estimatedEndDate, $status,
            $description, $percentCompleted, $resource, $dateExceptionType, $dateException, $taskStatus,
            $parent, $classification, $typeRelationship, $urlToTest, $errorClassificationId, $productVersionId);
    }

    /**
     * @param Entities\Task $task
     * @param $typeRelationship
     * @param $relatedTaskIds
     */
    private function createRelationship($task, $typeRelationship, &$relatedTaskIds)
    {
        $generatorTask = $this->getGeneratorTask($task, $typeRelationship, $relatedTaskIds);
        if ($generatorTask) {
            $relationship = InstanceProvider::getFactory()->getTaskRelationship()->createTaskRelationship($generatorTask, $task, $typeRelationship);
            $this->entityManager->persist($relationship);
            $this->entityManager->flush();
            $taskPrecursorService = InstanceProvider::getService()->getTaskPrecursor();
            $taskPrecursorService->saveFirstRelationship($generatorTask,$task,$typeRelationship);
            if($typeRelationship == Entities\TaskRelationship::MERGE_REQUEST_TYPE || $typeRelationship == Entities\TaskRelationship::TEST_TYPE) {
                 $taskPrecursorService->setDescriptionByTaskRelationship($task->getId(), $typeRelationship, $generatorTask->getId());
            }
        }
    }

    /**
     * @param $createdTaskIds
     */
    private function createRelatedRelationship($createdTaskIds)
    {
        $taskRepository = InstanceProvider::getRepository()->getTask();
        $tasks = $taskRepository->getTasksWithSameResource($createdTaskIds);
        foreach ($tasks as $task) {
            $taskIds = DevOpsUtilities::setSelectMultiInQuery($task['TASKS']);
            if (count($taskIds) > 1) {
                $this->createTaskRelationshipRelatedTask($taskIds);
            } else {
                $task = $taskRepository->findOneBy(array('id' => $taskIds));
                if (strpos($task->getName(), 'RLC')) {
                    $name = DevOpsUtilities::removeRLCTag($task->getName());
                    $task->setName($name);
                    $this->entityManager->persist($task);
                    $this->entityManager->flush();
                }
            }
        }

    }

    /**
     * @param $createdTaskIds
     */
    private function removeRelatedMergeRequestRelationship($createdTaskIds)
    {
        $relationships = InstanceProvider::getRepository()->getTask()->getRelatedMergeRelationshipIdByGeneratedTaskIds($createdTaskIds);
        $taskRelationshipRepository = InstanceProvider::getRepository()->getTaskRelationship();
        foreach ($relationships as $relationship) {
            /** @var Entities\TaskRelationship $relatedRelationship */
            $relatedRelationship = $taskRelationshipRepository->find($relationship['ID']);
            $taskStatus = $taskStatus = Utilities::getTaskStatusPaused();
            $this->updateTaskAndAllocationStatus($relatedRelationship->getTask(), $taskStatus);
            $this->updateTaskAndAllocationStatus($relatedRelationship->getTaskRelationship(), $taskStatus);
            $this->entityManager->remove($relatedRelationship);
        }
    }

    /**
     * @param Entities\Task $task
     * @param Entities\TaskStatus $taskStatus
     */
    public function updateTaskAndAllocationStatus(Entities\Task $task, Entities\TaskStatus $taskStatus)
    {
        $task->setTaskStatus($taskStatus);
        $this->entityManager->persist($task);
        $allocations = InstanceProvider::getRepository()->getAllocation()->findBy(array('task' => $task));
        /** @var Entities\Allocation $allocation */
        foreach($allocations as $allocation) {
            $allocation->setTaskStatus($taskStatus);
            $this->entityManager->persist($allocation);
        }
    }

    /**
     * @param $taskStatus
     * @return null|object
     */
    private function getTaskStatusById($taskStatus)
    {
        if ($taskStatus) {
            return InstanceProvider::getRepository()->getTaskStatus()->find($taskStatus);
        }
        return $taskStatus;
    }

    /**
     * @param $classification
     * @return Entities\Classification
     */
    private function getClassificationById($classification)
    {
        if ($classification) {
            return InstanceProvider::getRepository()->getClassification()->find($classification);
        }
        return $classification;
    }

    /**
     * @param Entities\Task $task
     * @param $resource
     */
    private function setTaskTimePlanned(Entities\Task $task, $resource)
    {
        if (!$task->isParent()) {
            $factor = 1;
            if (!empty($resource)) {
                $factor = count($resource);
            }
            $task->setTimePlanned($factor * $task->getDuration());
        }
    }

    /**
     * @param $task
     * @param $resource
     * @param $taskStatus
     */
    private function setAllocationsInTask(Entities\Task $task, $resource, $taskStatus)
    {
        $oldTaskStatus = $task->getTaskStatus();
        $allocationService = InstanceProvider::getService()->getAllocation();
        $resourceRepository = InstanceProvider::getRepository()->getResource();
        $allocationRepository = InstanceProvider::getRepository()->getAllocation();
        foreach ($resource as $eachResource) {
            if (!empty($eachResource)) {
                $resourceProject = $resourceRepository->find($eachResource);
                /** @var Entities\Allocation $allocation */
                $allocation = $allocationRepository->findOneBy(array("task" => $task->getId(),
                    "resource" => $resourceProject->getId()));
                if (empty($allocation)) {
                    $entity = $allocationService->createAllocation($task, $resourceProject, null, null);
                    return $entity;
                } else {
                    if (empty($oldTaskStatus) || ($oldTaskStatus && $taskStatus && ($oldTaskStatus->getId() != $taskStatus->getId())))
                        $allocationService->updateAllocation($allocation, $task, $resourceProject, null, $taskStatus);
                }
            }
        }
        /** @var Entities\allocation[] $allocation */
        $allocation = $allocationRepository->findBy(array("task" => $task->getId()));
        foreach ($allocation as $eachAllocation) {
            if (!in_array($eachAllocation->getResource()->getId(), $resource)) {
                $this->entityManager->remove($eachAllocation);
                $this->entityManager->flush();
            }
        }
    }

    /**
     * @param $task
     * @param $precursor
     * @param $precursorType
     */
    private function setPrecursorTask($task, $precursor, $precursorType)
    {
        $precursorLink = InstanceProvider::getRepository()->getTaskRelationship()->findOneBy(array("task" => $task->getId(),
            'type' => TaskPrecursor::PRECURSOR_TYPE));
        if ($precursor) {
            $taskPrecursor = InstanceProvider::getRepository()->getTask()->findOneBy(array('id' => $precursor));
            if($taskPrecursor){
                $taskPrecursorService = InstanceProvider::getService()->getTaskPrecursor();
                if (empty($precursorLink)) {
                    $taskPrecursorService->createTaskPrecursor($task, $taskPrecursor, TaskPrecursor::PRECURSOR_TYPE,
                        $precursorType);
                } else {
                    $taskPrecursorService->updateTaskPrecursor($precursorLink, $task, $taskPrecursor,
                        TaskPrecursor::PRECURSOR_TYPE, $precursorType);
                }
            }
        } else if ($precursorLink) {
            $this->entityManager->remove($precursorLink);
            $this->entityManager->flush();
        }
    }

    /**
     * @param $task
     * @throws \Exception
     */
    private function deletingTask($task)
    {
        $onlyTask = true;
        $relatedTask = $this->getRelatedTasksByTasks($task);
        if ($relatedTask) {
            $onlyTask = $this->applyOrderToIssuesByTaskSingleRelatedTask($task);
        }
        $this->deleteTaskRelationship($task);
        $this->recursiveDeleteTask($task);
        $this->entityManager->flush();
        if ($relatedTask && !$onlyTask) {
            $this->applyOrderToIssuesByTask($relatedTask);
        }
    }

    /**
     * @param Entities\Task $task
     * @param $typeRelationship
     * @param  array $relatedTaskIds
     * @return Entities\Task | null
     */
    private function getGeneratorTask($task, $typeRelationship, &$relatedTaskIds)
    {
        /** @var Entities\Task $generatorTask */
        $generatorTask = null;

        $taskRepository = InstanceProvider::getRepository()->getTask();
        if (in_array($typeRelationship, Entities\TaskRelationship::getAllTypesInSameProject())) {
            /** @var Entities\Task[] $relatedTasks */
            $relatedTasks = $taskRepository->findBy(array('id' => $relatedTaskIds));
            foreach ($relatedTasks as $relatedTask) {
                if ($relatedTask->getProject()->getId() == $task->getProject()->getId()) {
                    $generatorTask = $relatedTask;
                    break;
                }
            }
        } else {
            $generatorTaskId = array_shift($relatedTaskIds);
            $generatorTask = $taskRepository->findOneBy(array('id' => $generatorTaskId));
        }
        return $generatorTask;
    }

    public function findRelatedTasks($taskIds){

        $taskRepository = InstanceProvider::getRepository()->getTask();
        $relatedTasks = InstanceProvider::getRepository()->getTaskRelationship()->findRelatedTaskByTaskIds($taskIds);
        foreach ( $relatedTasks as &$relatedTask) {
            $relatedTask['RELATED_TASKS'] = $taskRepository->getTaskDataToGenerateTask(explode(',',$relatedTask['RELATED_TASKS']));
        }
        return $relatedTasks;
    }

    public function saveMultipleProjectTaskFromProjectTask($tasks, $versionDevId = null)
    {
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            $createdTaskIds = array();
            $allRelatedTaskIds = array();
            $taskPrecursorService = InstanceProvider::getService()->getTaskPrecursor();
            $taskRepository = InstanceProvider::getRepository()->getTask();
            $resourceRepository = InstanceProvider::getRepository()->getResource();
            foreach($tasks as $task) {
                $totalRating = 0;

                list($project, $timePlanned, $name, $status, $description, $percentCompleted, $resources,
                    $dateExceptionType, $dateException, $taskStatus, $parent, $classification, $typeRelationship,
                    $urlToTest,$taskId) = $this->getParamsSaveMultipleProjectTaskFromProjectTask($task);

                /** @var Entities\Task $task */
                $task = $taskRepository->find($taskId);
                /** @var Entities\Project $project */
                if($project){
                    $project = InstanceProvider::getRepository()->getProject()->find($project);
                }else{
                    $project = $task->getProject();
                }
                /** @var Entities\Classification $classification */
                $classification = InstanceProvider::getRepository()->getClassification()->findOneBy(array('control'=> $typeRelationship));

                $taskStatus = DevOpsUtilities::getTaskStatusNotInitialized();
                /** @var Entities\Resource[] $resourcesObjects  */
                $resourcesObjects = $resourceRepository->findBy(array('id' => $resources));
                if(!$resourcesObjects){
                    $totalRating = 1;
                }
                foreach($resourcesObjects as $resourceObjects){
                    if($resourceObjects->getRating()){
                        $totalRating += $resourceObjects->getRating();
                    }else{
                        $totalRating += 1;
                    }
                }
                $duration = $timePlanned / $totalRating;
                $duration = DevOpsUtilities::secondsToTime($duration);
                $estimatedBeginDate = clone $task->getEstimatedEndDate();
                $estimatedEndDate = clone $task->getEstimatedEndDate();
                $taskCreated = $this->createNewTaskProject($project, 0, $name, $estimatedBeginDate, $estimatedEndDate,
                    $status, $description, $percentCompleted, $duration, $resources, null, $parent, null,
                    $taskStatus, $dateExceptionType, $dateException, 'N', $urlToTest, $classification, null, null, $versionDevId);
                $taskCreated->setTimePlanned($timePlanned);
                /** @var Entities\TaskRelationship[] $tasksRelationship */
                $tasksRelationship = InstanceProvider::getRepository()->getTaskRelationship()->findBy(
                    array('taskRelationship' => $taskId, 'type' =>  TaskPrecursor::RELATIONSHIP_ISSUE_TYPE)
                );

                foreach($tasksRelationship as $taskRelationship){
                    $issue = $taskRepository->find($taskRelationship->getTask());
                    $taskPrecursorService->createTaskPrecursor(
                        $issue, $taskCreated, TaskPrecursor::RELATIONSHIP_ISSUE_TYPE, null, false
                    );
                }

                $this->entityManager->persist($taskCreated);
                $this->entityManager->flush();

                $taskIds = array($taskId);
                $this->createRelationship($taskCreated, $typeRelationship, $taskIds);
                $createdTaskIds[] = $taskCreated->getId();
                if($typeRelationship <> Entities\TaskRelationship::TEST_TYPE){
                    $relatedTaskIds = InstanceProvider::getRepository()->getTaskRelationship()->findRelatedTaskIds($taskId);
                    if (!empty($relatedTaskIds)) {
                        $allRelatedTaskIds[$relatedTaskIds['MAIN_TASK']][] = $taskCreated->getId();
                    }
                }
                if ($taskCreated->getProject()->getId() == $task->getProject()->getId()
                    && $taskCreated->getProject()->getProjectType()->getUsetimeplanning() == Entities\ProjectType::IS_USE_TIME_PLANNING
                ) {
                    $taskPrecursorService->createTaskPrecursor(
                        $taskCreated, $task, TaskPrecursor::PRECURSOR_TYPE, PrecursorFactory::END_TO_START);
                }

                InstanceProvider::getService()->getCalendarTakCalendar()->setPreviousTimeTask($taskCreated);
                $this->entityManager->persist($taskCreated);
                $this->entityManager->persist($task);
                $this->entityManager->flush();

            }

            foreach ($allRelatedTaskIds as $createdRelatedTaskIds) {
                $this->createRelatedRelationship($createdRelatedTaskIds);
                $this->removeRelatedMergeRequestRelationship($createdRelatedTaskIds);
            }

            $this->entityManager->flush();


            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
        return $createdTaskIds;
    }

    private function getParamsSaveMultipleProjectTaskFromProjectTask($task)
    {
        $project = empty($task['PROJECT']) ? null : $task['PROJECT'];
        $timePlanned = empty($task['TIME_PLANNED']) ? null : $task['TIME_PLANNED'];
        $name = $task['NAME'];
        $status = empty($task['STATUS']) ? 'A' : $task['STATUS'];
        $description = empty($task['DESCRIPTION']) ? null : $task['DESCRIPTION'];
        $percentCompleted = 0;
        $resource = empty($task['RESOURCE']) ? array() : $task['RESOURCE'];
        $dateExceptionType = empty($task['DATE_EXCEPTION_TYPE']) ? 'MB' : $task['DATE_EXCEPTION_TYPE'];
        $dateException = empty($task['DATE_EXCEPTION']) ? null : \DateTime::createFromFormat('d/m/Y', $task['DATE_EXCEPTION']);
        $taskStatus = empty($task['TASK_STATUS']) ? null : $task['TASK_STATUS'];
        $parent = empty($task['PARENT']) ? null : $task['PARENT'];
        $classification = empty($task['TASK_CLASSIFICATION']) ? null : $task['TASK_CLASSIFICATION'];
        $typeRelationship = empty($task['TYPE_RELATIONSHIP']) ? null : $task['TYPE_RELATIONSHIP'];
        $urlToTest = empty($task['URL_TO_TEST']) ? null : $task['URL_TO_TEST'];
        $taskId = empty($task['TASK_ID']) ? null : $task['TASK_ID'];

        return array($project, $timePlanned, $name, $status,
            $description, $percentCompleted, $resource, $dateExceptionType, $dateException, $taskStatus,
            $parent, $classification, $typeRelationship, $urlToTest,$taskId);
    }

    public function validateTaskToBeStarted($taskId)
    {
        try {
            $dataset = array('status' => '', 'msg' =>  '');
            $errors = InstanceProvider::getRepository()->getTask()->validateTaskToBeStarted($taskId);
            if($errors){
                $errorMessage = Messages::PREFIX_ERR_VALIDATE_TASK_TO_BE_STARTED;
                $errorMessages = Messages::getMessageValidateTaskToBeStarted();
                foreach($errors as $error){
                    $errorMessage .= $errorMessages[$error['ERROR']];
                }
                $dataset['status'] = 'error';
                $dataset['msg'] = $errorMessage;
            } else {
                $dataset['status'] = 'ok';
                $dataset['msg'] = '';
            }
            return $dataset;
        } catch (\Exception $e) {
            throw $e;
        }
    }

    /**
     * @param $conditions
     * @return array
     */
    public function getProjectsInfoCreateTasksByIssue($conditions)
    {
        $projectService = InstanceProvider::getService()->getProject();
        $productVersionRepository = InstanceProvider::getRepository()->getProductVersion();
        $projectRepository = InstanceProvider::getRepository()->getProject();
        $projectTypeRepository = InstanceProvider::getRepository()->getProjectType();

        $projectIds = DevOpsUtilities::getValueInFilterDataSourceByName($conditions, 'PROJECT_IDS');
        $conditionsValues = array(array('IS_PARENT', '=', 'N'));
        $conditionsNotIsParent = DevOpsUtilities::createArrayConditions($conditionsValues);

        $projectInfoToValidation = array();

        foreach ($projectIds as $id) {
            $isTasksNotIsParent = $projectService->projectHasTask($id, $conditionsNotIsParent);
            $project = $projectRepository->find($id);
            $projectName = $project->getName();
            $projectStatus = $project->getProjectStatus()->getControl();

            $productVersionLabel = null;
            if($project->getProductVersion()){
                $productVersion = $productVersionRepository->find($project->getProductVersion()->getId());
                $productVersionLabel = $productVersion->getVersion() . '.' . $productVersion->getRevision();
            }

            $productVersionId = $project->getProductVersion() ? $project->getProductVersion()->getId() : null;
            $projectProductId = $projectService->getProductProject($project)->getId();
            $externalType = $projectTypeRepository->find($project->getProjectType())->getExternaltype();

            $projectInfoToValidation[$id] = array(
                'projectId'             => $id,
                'projectName'           => $projectName,
                'isTasksNotIsParent'    => $isTasksNotIsParent,
                'projectStatusControl'  => $projectStatus,
                'productVersionId'      => $productVersionId,
                'productVersionLabel'   => $productVersionLabel,
                'projectProductId'      => $projectProductId,
                'externalType'          => $externalType
            );
        };
        return $projectInfoToValidation;
    }

    public function taskHasBranch($tasks, $targetProjectId, $projectId)
    {
        $tasksWithBranch = array();
        $targetProject = InstanceProvider::getRepository()->getProject()->find($targetProjectId);
        $project = InstanceProvider::getRepository()->getProject()->find($projectId);
        if($targetProject->getProjectType()->getExternaltype() != $project->getProjectType()->getExternaltype()){
            foreach ($tasks as $task) {
                $branch = InstanceProvider::getRepository()->getTaskBranchs()->findBy(array('task' => $task['ID']));
                if($branch){
                    $tasksWithBranch[] = $task;
                }
            }
        }
        return $tasksWithBranch;
    }

    public function findAutomaticFraming($tasks_type)
    {
        $automaticFraming = InstanceProvider::getRepository()->getAutomaticFraming()->getAutomaticFraming($tasks_type);
        
        return $automaticFraming;
    }

    public function saveAutomaticFraming($id, $time, $actualStatus, $destinyStatus, $taskTypeId, $description)
    {
        $connection = $this->entityManager->getConnection();
        if($actualStatus == $destinyStatus){
            throw new \Exception(Messages::ERR_SAME_STATS_BY_STATS_CONTROL);
        }
        $entity = null;
        try {
            $connection->beginTransaction();
            if($id){
                $entity = InstanceProvider::getRepository()->getAutomaticFraming()->find($id);
            }
            $isNotValid = InstanceProvider::getRepository()->getAutomaticFraming()->validateFraming($taskTypeId, $actualStatus, $id);
            if($isNotValid){
                throw new \Exception(Messages::ERR_ALREADY_HAS_FRAMING);
            }
            $automaticFraming = InstanceProvider::getFactory()->getAutomaticFraming()->createAutomaticFraming($entity, $time, $actualStatus, $destinyStatus, $taskTypeId, $description);    
            $this->entityManager->persist($automaticFraming);
            $this->entityManager->flush();
            $connection->commit();
            return $automaticFraming->getId();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }


    public function removeAutomaticFraming($id)
    {
        $connection = $this->entityManager->getConnection();
        try {
            $connection->beginTransaction();
            $automaticFraming = InstanceProvider::getRepository()->getAutomaticFraming()->find($id);
            $this->entityManager->remove($automaticFraming);
            $this->entityManager->flush();
            $connection->commit();
        } catch (\Exception $e) {
            $connection->rollBack();
            throw $e;
        }
    }
}
